! file src/gmsh_module.f90
!
! Copyright 2009-2011 Dalton Harvie (daltonh@unimelb.edu.au)
! 
! This file is part of arb finite volume solver, referred to as `arb'.
! 
! arb is a software package designed to solve arbitrary partial
! differential equations on unstructured meshes using the finite volume
! method.  Primarily it consists of fortran source code, perl source
! code and shell scripts.  arb replies on certain third party software
! to run, most notably the computer algebra system maxima
! <http://maxima.sourceforge.net/> which is released under the GNU GPL.
! 
! The copyright of arb is held by Dalton Harvie.
! 
! arb is released under the GNU GPL.  arb is free software: you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License (version 3) as published by the Free Software Foundation.
! You should have received a copy of the GNU General Public License
! along with arb (see file license/gpl.txt after unpacking).  If not,
! see <http://www.gnu.org/licenses/>.
! 
! arb is distributed in the hope that it will be useful, but WITHOUT
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
! FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
! for more details.
! 
! For full details of arb's license see the license directory.
! 
! The current homepage for the arb finite volume solver project is
! <http://www.chemeng.unimelb.edu.au/people/staff/daltonh/downloads/arb>.
!
!-------------------------------------------------------------------------
module gmsh_module

! statements specifying different data types and parameters
implicit none

type node_list_type
  integer, dimension(:), allocatable :: node
end type node_list_type
  
! this data structure holds info about how elements, faces and nodes are defined in gmsh
type gmsh_element_type
  character(len=100) :: name
  integer :: dimensions ! 0 for point, 1 for line, 2 for plane, 3 for volume
  integer :: nnodes ! number of surrounding nodes
  integer :: nfaces ! number of surrounding faces
  type(node_list_type), dimension(:), allocatable :: face_nodes ! lists the node indicies of each of the nfaces faces
  logical :: supported ! suitable for arb?
end type gmsh_element_type

type(gmsh_element_type), dimension(:), allocatable :: gmsh_element

integer, save :: n_gmsh_elements ! number of elements that need to be written to gmsh file including repetition for gmsh regions

!-----------------------------------------------------------------
contains

!-----------------------------------------------------------------

subroutine setup_gmsh_element

! here we allocate data to the reference gmsh_element array

integer :: n
logical, parameter :: debug = .false.

if (debug) write(*,'(80(1h+)/a)') 'subroutine setup_gmsh_element'

allocate(gmsh_element(31))

! default is that element is unsupported
do n=1,ubound(gmsh_element,1)
  gmsh_element(n)%name = "unknown"
  gmsh_element(n)%supported = .false.
end do

! now set supported elements
! faces are definined in the context of the element having the same dimensions as the problem
! ie, as a cell
! if the element has a dimension that is less than the problem dimension, it is assumed to be a
! face (dimension-1)
! surface nodes are directed using the right-hand rule with a positive outward pointing normal

n = 15
gmsh_element(n)%name = "1-node point"
gmsh_element(n)%dimensions = 0
gmsh_element(n)%nnodes = 1
gmsh_element(n)%nfaces = 0
gmsh_element(n)%supported = .true.

n = 1
gmsh_element(n)%name = "2-node line"
gmsh_element(n)%dimensions = 1
gmsh_element(n)%nnodes = 2
gmsh_element(n)%nfaces = 2
allocate(gmsh_element(n)%face_nodes(gmsh_element(n)%nfaces))
allocate(gmsh_element(n)%face_nodes(1)%node(1))
gmsh_element(n)%face_nodes(1)%node = [ 1 ]
allocate(gmsh_element(n)%face_nodes(2)%node(1))
gmsh_element(n)%face_nodes(2)%node = [ 2 ]
gmsh_element(n)%supported = .true.

n = 2
gmsh_element(n)%name = "3-node triangle"
gmsh_element(n)%dimensions = 2
gmsh_element(n)%nnodes = 3
gmsh_element(n)%nfaces = 3
allocate(gmsh_element(n)%face_nodes(gmsh_element(n)%nfaces))
allocate(gmsh_element(n)%face_nodes(1)%node(2))
gmsh_element(n)%face_nodes(1)%node = [ 1, 2 ]
allocate(gmsh_element(n)%face_nodes(2)%node(2))
gmsh_element(n)%face_nodes(2)%node = [ 2, 3 ]
allocate(gmsh_element(n)%face_nodes(3)%node(2))
gmsh_element(n)%face_nodes(3)%node = [ 3, 1 ]
gmsh_element(n)%supported = .true.

n = 3
gmsh_element(n)%name = "4-node quadrangle"
gmsh_element(n)%dimensions = 2
gmsh_element(n)%nnodes = 4
gmsh_element(n)%nfaces = 4
allocate(gmsh_element(n)%face_nodes(gmsh_element(n)%nfaces))
allocate(gmsh_element(n)%face_nodes(1)%node(2))
gmsh_element(n)%face_nodes(1)%node = [ 1, 2 ]
allocate(gmsh_element(n)%face_nodes(2)%node(2))
gmsh_element(n)%face_nodes(2)%node = [ 2, 3 ]
allocate(gmsh_element(n)%face_nodes(3)%node(2))
gmsh_element(n)%face_nodes(3)%node = [ 3, 4 ]
allocate(gmsh_element(n)%face_nodes(4)%node(2))
gmsh_element(n)%face_nodes(4)%node = [ 4, 1 ]
gmsh_element(n)%supported = .true.

n = 4
gmsh_element(n)%name = "4-node tetrahedron"
gmsh_element(n)%dimensions = 3
gmsh_element(n)%nnodes = 4
gmsh_element(n)%nfaces = 4
allocate(gmsh_element(n)%face_nodes(gmsh_element(n)%nfaces))
allocate(gmsh_element(n)%face_nodes(1)%node(3))
gmsh_element(n)%face_nodes(1)%node = [ 1, 4, 3 ]
allocate(gmsh_element(n)%face_nodes(2)%node(3))
gmsh_element(n)%face_nodes(2)%node = [ 2, 3, 4 ]
allocate(gmsh_element(n)%face_nodes(3)%node(3))
gmsh_element(n)%face_nodes(3)%node = [ 1, 2, 4 ]
allocate(gmsh_element(n)%face_nodes(4)%node(3))
gmsh_element(n)%face_nodes(4)%node = [ 1, 3, 2 ]
gmsh_element(n)%supported = .true.

n = 5
gmsh_element(n)%name = "8-node hexahedron"
gmsh_element(n)%dimensions = 3
gmsh_element(n)%nnodes = 8
gmsh_element(n)%nfaces = 6
allocate(gmsh_element(n)%face_nodes(gmsh_element(n)%nfaces))
allocate(gmsh_element(n)%face_nodes(1)%node(4))
gmsh_element(n)%face_nodes(1)%node = [ 1, 5, 8, 4 ]
allocate(gmsh_element(n)%face_nodes(2)%node(4))
gmsh_element(n)%face_nodes(2)%node = [ 5, 6, 7, 8 ]
allocate(gmsh_element(n)%face_nodes(3)%node(4))
gmsh_element(n)%face_nodes(3)%node = [ 2, 3, 7, 6 ]
allocate(gmsh_element(n)%face_nodes(4)%node(4))
gmsh_element(n)%face_nodes(4)%node = [ 1, 4, 3, 2 ]
allocate(gmsh_element(n)%face_nodes(5)%node(4))
gmsh_element(n)%face_nodes(5)%node = [ 1, 2, 6, 5 ]
allocate(gmsh_element(n)%face_nodes(6)%node(4))
gmsh_element(n)%face_nodes(6)%node = [ 3, 4, 8, 7 ]
gmsh_element(n)%supported = .true.

n = 6
gmsh_element(n)%name = "6-node prism"
gmsh_element(n)%dimensions = 3
gmsh_element(n)%nnodes = 6
gmsh_element(n)%nfaces = 5
allocate(gmsh_element(n)%face_nodes(gmsh_element(n)%nfaces))
allocate(gmsh_element(n)%face_nodes(1)%node(4))
gmsh_element(n)%face_nodes(1)%node = [ 1, 2, 5, 4 ]
allocate(gmsh_element(n)%face_nodes(2)%node(4))
gmsh_element(n)%face_nodes(2)%node = [ 2, 3, 6, 5 ]
allocate(gmsh_element(n)%face_nodes(3)%node(4))
gmsh_element(n)%face_nodes(3)%node = [ 1, 4, 6, 3 ]
allocate(gmsh_element(n)%face_nodes(4)%node(3))
gmsh_element(n)%face_nodes(4)%node = [ 1, 3, 2 ]
allocate(gmsh_element(n)%face_nodes(5)%node(3))
gmsh_element(n)%face_nodes(5)%node = [ 4, 5, 6 ]
gmsh_element(n)%supported = .true.

n = 7
gmsh_element(n)%name = "5-node pyramid"
gmsh_element(n)%dimensions = 3
gmsh_element(n)%nnodes = 5
gmsh_element(n)%nfaces = 5
allocate(gmsh_element(n)%face_nodes(gmsh_element(n)%nfaces))
allocate(gmsh_element(n)%face_nodes(1)%node(3))
gmsh_element(n)%face_nodes(1)%node = [ 1, 2, 5 ]
allocate(gmsh_element(n)%face_nodes(2)%node(3))
gmsh_element(n)%face_nodes(2)%node = [ 2, 3, 5 ]
allocate(gmsh_element(n)%face_nodes(3)%node(3))
gmsh_element(n)%face_nodes(3)%node = [ 3, 4, 5 ]
allocate(gmsh_element(n)%face_nodes(4)%node(3))
gmsh_element(n)%face_nodes(4)%node = [ 1, 5, 4 ]
allocate(gmsh_element(n)%face_nodes(5)%node(4))
gmsh_element(n)%face_nodes(5)%node = [ 1, 4, 3, 2 ]
gmsh_element(n)%supported = .true.

if (debug) write(*,'(a/80(1h+))') 'subroutine setup_gmsh_element'

end subroutine setup_gmsh_element

!-----------------------------------------------------------------

subroutine read_gmsh_mesh

! this subroutine reads in a gmsh mesh file
! specifically it:
! - defines any regions not already defined (name, centring and location)
! - defines nodes (x, jface) and sets ktotal
! - defines cells (dimensions, jface, knode, gmsh_element), adds to relevant region list and sets itotal
! - defines faces (dimensions, icell, knode, gmsh_element), adds to relevant region list and sets jtotal

use general_module
integer :: gmsh_bin, ktmp, nelements, k, n, ngmsh, gmsh_type, nphys, phys_dimensions, &
  region_number, error, ntags, nnode, nface, nfaces, j, i, ntag, nn, jj, m, nelements_tenth
integer, dimension(:), allocatable :: tags, local_element_nodes
type phys_to_region_type
  integer :: gmsh, arb
end type phys_to_region_type
type(phys_to_region_type), dimension(:), allocatable :: phys_to_region
real :: gmsh_version
integer, parameter :: fgmsh = 2
type(cell_type) :: default_cell
type(face_type) :: default_face
character(len=1000) :: formatline
character(len=1000) :: textline, region_name, tmp
character(len=1000) :: location
character(len=4) :: centring
logical :: existing
logical, parameter :: debug = .false.

if (debug) write(*,'(80(1h+)/a)') 'subroutine read_gmsh_mesh'

open(unit=fgmsh,file=trim(gmsh_file),status='old',iostat=error)
if (error /= 0) stop 'ERROR: problem opening mesh file'

!---------------------------------------------
! check version and format of file
do 
  read(fgmsh,'(a)',iostat=error) textline
  if (error /= 0) stop 'ERROR: gmsh file version not found'
  if (trim(textline) == "$MeshFormat") exit
end do
read(fgmsh,*,iostat=error) gmsh_version, gmsh_bin
if (error /= 0) stop 'ERROR: problem reading in gmsh file version'
if (gmsh_version < 2.1) then
  write(*,'(a,f4.1,a)') 'ERROR: the gmsh file version (',gmsh_version,') is too old and is not supported (update to v2.1)'
  stop
else if (gmsh_version > 2.1) then
  write(*,'(a,f4.1,a)') 'WARNING: the gmsh mesh file version (',gmsh_version,') is not fully suported yet (but may work)'
end if
if (gmsh_bin /= 0) stop 'ERROR: gmsh binary file type not supported'
if (debug) write(*,*) 'gmsh_version = ',gmsh_version
    
!---------------------------------------------
! read in physical names and possibly create any required regions
rewind(fgmsh)
do 
  read(fgmsh,'(a)',iostat=error) textline
  if (error /= 0) stop 'ERROR: list of physical names not found in gmsh file'
  if (trim(textline) == "$PhysicalNames") exit
end do
read(fgmsh,*,iostat=error) nphys
if (error /= 0) stop 'ERROR: problem reading in number of physical names in gmsh file'
if (debug) write(*,*) 'number of physical names: nphys = ',nphys
allocate(phys_to_region(nphys)) ! this is a mapping from the gmsh physical region number (%gmsh) to the arb region number (%arb)
do n=1,nphys
  read(fgmsh,*,iostat=error) phys_dimensions,phys_to_region(n)%gmsh,region_name
  if (error /= 0) stop 'ERROR: problem reading physical name in gmsh file'
  if (phys_dimensions < dimensions-1) then
    write(*,'(a,a,a)') 'WARNING: physical name ',trim(region_name),' ignored in gmsh file'
    cycle
  end if

! create new region or find number of existing one and check consistency
  write(tmp,'(a,i3)') "GMSH physical entity number = ",phys_to_region(n)%gmsh
  location = trim(tmp)
  if (phys_dimensions == dimensions) then
    centring = "cell"
  else
    centring = "face"
  end if
  region_number = region_number_from_name(name=region_name,location=location,centring=centring,existing=existing,creatable=.true.)
  if (existing) then
!   write(*,*) "ERROR: region "//trim(region_name)//" in gmsh file conflicts with one that has already been defined"
!   stop
    write(*,'(a)') "NOTE: region "//trim(region_name)//" defined in the gmsh file may conflict with a user definition later"
  end if
  if (region_number == 0) then
    write(*,*) "ERROR: problem allocating region "//trim(region_name)// &
      " from gmsh file:- most probably centring conflicts with previously set region of the same name"
    stop
  end if
  phys_to_region(n)%arb = region_number
  region(region_number)%gmsh_region = .true. ! flag to indicate that this region originated from the gmsh file

end do
if (debug) write(*,*) 'physical names read in'

!---------------------------------------------
! read in nodes
rewind(fgmsh)
do 
  read(fgmsh,'(a)',iostat=error) textline
  if (error /= 0) stop 'ERROR: list of nodes not found in gmsh file'
  if (trim(textline) == "$Nodes") exit
end do
read(fgmsh,*,iostat=error) ktotal
if (error /= 0) stop 'ERROR: problem reading in number of nodes in gmsh file'
if (debug) write(*,*) 'number of nodes: ktotal = ',ktotal
allocate(node(ktotal))
do k=1,ktotal
  read(fgmsh,*,iostat=error) ktmp,node(k)%x
  if (ktmp /= k .or. error /= 0) stop 'ERROR: problem reading nodes in gmsh file'
end do
if (debug) write(*,*) 'nodes read in'
    
!---------------------------------------------
! read in elements, creating cells and faces as we go
rewind(fgmsh)

itotal = 0 ! this is the current number of cells
jtotal = 0 ! this is the current number of faces
allocate(default_face%icell(2)) ! allocate spaces for the two surrounding faces

do 
  read(fgmsh,'(a)',iostat=error) textline
  if (error /= 0) stop 'ERROR: list of elements not found in gmsh file'
  if (trim(textline) == "$Elements") exit
end do
read(fgmsh,*,iostat=error) nelements
if (error /= 0) stop 'ERROR: problem reading in number of elements in gmsh file'
if (debug) write(*,*) 'number of elements: nelements = ',nelements

nelements_tenth = max(int(dble(nelements)/10.d0),1)
if (mod(nelements,nelements_tenth) /= 0) nelements_tenth = nelements_tenth + 1

do n = 1, nelements

  if (mod(n,nelements_tenth) == 0 .or. n == nelements) write(*,'(a,i3,a)') &
    'INFO: reading in mesh:',min(int(dble(n)*100./float(nelements)),100),'%'

  read(fgmsh,'(a)',iostat=error) textline
  if (error /= 0) stop 'ERROR: problem reading elements in gmsh file'
  read(textline,*,iostat=error) ngmsh, gmsh_type, ntags
  if (error /= 0) stop 'ERROR: problem reading initial element data in gmsh file'
  if (gmsh_type > ubound(gmsh_element,1) .or. .not.gmsh_element(gmsh_type)%supported) stop &
    'ERROR: element type not supported in gmsh file'

! get tags and local_element_nodes arrays the correct size ready for reading
  call resize_integer_array(array=tags,new_size=ntags)
  call resize_integer_array(array=local_element_nodes,new_size=gmsh_element(gmsh_type)%nnodes)

! re-read line now knowing number of tags and number of nodes
  read(textline,*,iostat=error) ngmsh, gmsh_type, ntags, tags, local_element_nodes
  if (error /= 0) stop 'ERROR: problem reading element tags/nodes in gmsh file'

  if (debug) write(81,*) 'ngmsh = ',ngmsh,': gmsh_type = ',gmsh_type,': name = ',trim(gmsh_element(gmsh_type)%name), &
    ': ntags = ',ntags,': tags = ',tags,': local_element_nodes = ',local_element_nodes

!------------------
! import as a cell

  if (gmsh_element(gmsh_type)%dimensions == dimensions) then

!   nfaces = ubound(gmsh_element(gmsh_type)%face_nodes,1)
    nfaces = gmsh_element(gmsh_type)%nfaces
    if (nfaces == 0) stop 'ERROR: a cell has no faces in the gmsh file'
    call reset_cell(default_cell)
    call resize_integer_array(array=default_cell%jface,new_size=nfaces)

! loop through face node lists
    do nface = 1, nfaces

      call reset_face(default_face)
      call resize_integer_array(array=default_face%knode,new_size=ubound(gmsh_element(gmsh_type)%face_nodes(nface)%node,1))

! assemble face_node list
      do nnode = 1, ubound(default_face%knode,1)
        default_face%knode(nnode) = local_element_nodes(gmsh_element(gmsh_type)%face_nodes(nface)%node(nnode))
      end do

      if (debug) write(81,*) '  nface = ',nface,': default_face%knode = ',default_face%knode

! find face number if this face has been previously defined
      j = jface_from_node_list(default_face%knode)

! if face is not found then create new one
      if (j == 0) then
        jtotal = jtotal + 1
        if (ubound(face,1) < jtotal) call resize_face_array(change=int(ktotal/10)+1)
        j = jtotal
        default_face%gmsh_element = 0
! find gmsh_type for face that has correct dimension and number of nodes
        default_face%gmsh_type = 0
        do m = 1, ubound(gmsh_element,1)
          if (gmsh_element(m)%dimensions == gmsh_element(gmsh_type)%dimensions - 1.and. &
            gmsh_element(m)%nnodes == ubound(default_face%knode,1) ) then
            default_face%gmsh_type = m
            exit
          end if
        end do
        if (default_face%gmsh_type == 0) stop 'ERROR: gmsh_type not identified for generated face in read_gmsh_mesh'

! set node jface references for the new face
        call add_jface_to_nodes(j,default_face%knode)
          
        if (debug) write(81,*) 'creating new face: j = ',j,': knode = ',default_face%knode,': gmsh_type = ',default_face%gmsh_type
      else
        default_face%gmsh_element = face(j)%gmsh_element ! save gmsh element number and type
        default_face%gmsh_type = face(j)%gmsh_type
        if (debug) write(81,*) 'overwriting existing face: j = ',j,': knode = ',default_face%knode
      end if

      default_face%icell = [ 0, 0 ] ! reset icell(1) later, scrap icell(2)

! set or overwrite the face data to ensure that this cell is the icell=1 one
      default_face%dimensions = dimensions - 1
      call set_face(face_to_set=face(j),new_value=default_face)

      default_cell%jface(nface) = j

    end do

! now check that cell doesn't exist already from knowledge of nodes
    i = icell_from_node_list(local_element_nodes)

    if (i == 0) then
      default_cell%dimensions = dimensions
      default_cell%gmsh_element = ngmsh
      default_cell%gmsh_type = gmsh_type
      call copy_integer_array(original=local_element_nodes,copy=default_cell%knode)
      itotal = itotal + 1
      if (ubound(cell,1) < itotal) call resize_cell_array(change=int(ktotal/5)+1)
      i = itotal
      call set_cell(cell_to_set=cell(i),new_value=default_cell)

! set node icell references for the new cell
      call add_icell_to_nodes(i,default_cell%knode)

      if (debug) write(81,*) 'creating new cell: i = ',i,': jface = ',default_cell%jface
    end if

! set icell(1) so that the surrounding face normals point in the direction of the right-hand rule for each set of face nodes
    do jj = 1, ubound(cell(i)%jface,1)
      j = cell(i)%jface(jj)
      face(j)%icell(1) = i
    end do

! add cell number to region list if it isn't there already
    if (tags(1) /= 0) then
      do nn = 1, ubound(phys_to_region,1)
        if (phys_to_region(nn)%gmsh == tags(1)) then
          if (debug) write(81,*) 'cell i = ',i,': is being added to region = ',trim(region(phys_to_region(nn)%arb)%name)
          if (region(phys_to_region(nn)%arb)%centring /= 'cell') &
            stop 'ERROR: gmsh cell is assigned to a region which is not cell centred'
          call push_integer_array(array=region(phys_to_region(nn)%arb)%ij,new_element=i)
        end if
      end do
    end if

!------------------
! import as a face

  else if (gmsh_element(gmsh_type)%dimensions == dimensions-1) then

! find face number if this face has been previously defined
    j = jface_from_node_list(local_element_nodes)

! if face is not found then create new one
    if (j == 0) then
      call reset_face(default_face)
      call copy_integer_array(original=local_element_nodes,copy=default_face%knode)
      default_face%dimensions = dimensions - 1
      default_face%gmsh_element = ngmsh
      default_face%gmsh_type = gmsh_type
      default_face%icell = [ 0, 0 ]
      jtotal = jtotal + 1
      if (ubound(face,1) < jtotal) call resize_face_array(change=int(ktotal/5)+1)
      j = jtotal
      call set_face(face_to_set=face(j),new_value=default_face)
      if (debug) write(81,*) 'creating new face: j = ',j,': knode = ',default_face%knode
! set node jface references for the new face
      call add_jface_to_nodes(j,default_face%knode)
    else
      face(j)%gmsh_element = ngmsh
    end if

! add face number to region list if it isn't there already
    if (tags(1) /= 0) then
      do nn = 1, ubound(phys_to_region,1)
        if (phys_to_region(nn)%gmsh == tags(1)) then
          if (debug) write(81,*) 'face j = ',j,': is being added to region = ',trim(region(phys_to_region(nn)%arb)%name)
          if (region(phys_to_region(nn)%arb)%centring /= 'face') &
            stop 'ERROR: gmsh face is assigned to a region which is not face centred'
          call push_integer_array(array=region(phys_to_region(nn)%arb)%ij,new_element=j)
        end if
      end do
    end if

!------------------
  else
    write(*,*) 'WARNING: element of gmsh_type = ',gmsh_type,' ignored in gmsh file'
  end if

end do

! resize cell and face arrays
call resize_cell_array(new_size=itotal)
call resize_face_array(new_size=jtotal)

!---------------------------------------------

close(fgmsh)

if (debug) write(*,'(a/80(1h-))') 'subroutine read_gmsh_mesh'

end subroutine read_gmsh_mesh

!-----------------------------------------------------------------

subroutine output_gmsh(nstep,debug_dump)

! dumps the data to a file fit for printing

use general_module
integer, parameter :: foutput = 3
integer :: error, region_dimensions, nstep, i, j, k, l, m, region_number, kk, nnodes, n, mc, nrank, nregions, ns
double precision, dimension(:), allocatable :: cellvaluel
double precision, dimension(:,:), allocatable :: cellgradl
character(len=1000) :: formatline
character(len=100) :: filename
character(len=100) :: element_data_string
logical :: debug_dump ! outputs components of all variables without any gradients
logical, parameter :: element_data = .false. ! output cell centred data as ElementData rather than ElementNodeData
logical, parameter :: debug = .false.
                  
if (debug) write(*,'(80(1h+)/a)') 'subroutine output_gmsh'

!----------------------------------------------------
! output of field variables
if (debug_dump) then
  filename = "output/output.debug.msh"
else
  filename = "output/output.msh"
end if
open(foutput,file=trim(filename),status='replace',iostat=error)
if (error /= 0) then
  write(*,*) 'ERROR: problem opening file ',filename
  stop
end if

if (element_data) then
  element_data_string = 'ElementData'
else
  element_data_string = 'ElementNodeData'
end if

! intro stuff
write(foutput,'(a/a/a)') '$MeshFormat','2.1 0 8','$EndMeshFormat'

! list regions
formatline = '(a/'//trim(indexformat)//')'
nregions = 0
do m = 1, ubound(region,1)
  if (region(m)%gmsh_region) nregions = nregions + 1
end do
write(foutput,fmt=formatline) '$PhysicalNames', nregions
do m = 1, ubound(region,1)
  if (.not.region(m)%gmsh_region) cycle
  region_dimensions = dimensions
  if (region(m)%centring == 'face') region_dimensions = dimensions - 1
  write(foutput,'(i1,a,i3,a,a,a)') region_dimensions,' ',m,' "',trim(region(m)%name),'"'
end do
write(foutput,'(a)') '$EndPhysicalNames'

! list nodes
formatline = '(a/'//trim(indexformat)//')'
write(foutput,fmt=formatline) '$Nodes', ktotal
formatline = '('//trim(indexformat)//',3(a,'//trim(floatformat)//'))'
do k = 1, ktotal
  write(foutput,fmt=formatline) k,' ',node(k)%x(1),' ',node(k)%x(2),' ',node(k)%x(3)
end do
write(foutput,'(a)') '$EndNodes'

! list domain cells and boundary faces as elements, using repetition to output multiple gmsh regions
formatline = '(a/'//trim(indexformat)//')'
write(foutput,fmt=formatline) '$Elements', n_gmsh_elements
do i = 1, idomain
  cell(i)%gmsh_element = i
  formatline = '('//trim(indexformat)//',5(a,i3,)'//repeat(',a,'//trim(indexformat),ubound(cell(i)%knode,1))//')'
  do n = 1, ubound(cell(i)%gmsh_region_list,1)
    write(foutput,fmt=formatline) i,' ',cell(i)%gmsh_type,' ',3,' ',cell(i)%gmsh_region_list(n),' ', &
      cell(i)%gmsh_region_list(n),' ',0,(' ',cell(i)%knode(l),l=1,ubound(cell(i)%knode,1)) ! seem to need to create new elementary geometry for each new region_number
  end do
end do
do i = idomain+1, itotal
  cell(i)%gmsh_element = idomain + cell(i)%jface(1)
end do
do j = 1, jtotal
  face(j)%gmsh_element = idomain + j
  formatline = '('//trim(indexformat)//',5(a,i3,)'//repeat(',a,'//trim(indexformat),ubound(face(j)%knode,1))//')'
  do n = 1, ubound(face(j)%gmsh_region_list,1)
    write(foutput,fmt=formatline) j+idomain,' ',face(j)%gmsh_type,' ',3,' ',face(j)%gmsh_region_list(n),' ', &
      face(j)%gmsh_region_list(n),' ',0,(' ',face(j)%knode(l),l=1,ubound(face(j)%knode,1))
  end do
end do
write(foutput,'(a)') '$EndElements'

!---------------------------------------------------------------------------------
! output as elementnode data (ie, includes effect of gradient)

! cell centred var components
do n = 1, allocatable_size(var_list(var_list_number(centring="cell",type="all"))%list)
  m = var_list(var_list_number(centring="cell",type="all"))%list(n)

! done as ElementData when debug_dump is on or element_data and output for each variable is on
  if (debug_dump.or.element_data) then

    if (.not.var(m)%output.and..not.debug_dump) cycle
    formatline = '(a/i1/a/i1/i1/i1/i1/'//trim(indexformat)//')'
    write(foutput,fmt=formatline) '$ElementData',1,'"'//trim(var(m)%name)//'"',0,3,0,1,ubound(var(m)%funk,1)
    do ns = 1, ubound(var(m)%funk,1)
      i = ijvar(m,ns)
      if (debug) write(91,*) 'cell components: i = ',i,': ns = ',ns,': m = ',m,': var = ',trim(var(m)%name)
      formatline = '('//trim(indexformat)//',a,'//trim(floatformat)//')'
      write(foutput,fmt=formatline) cell(i)%gmsh_element,' ',real(var(m)%funk(ns)%v)
    end do
    write(foutput,'(a)') '$EndElementData'

! done as ElementNodeData if output is on otherwise
  else

    if (.not.var(m)%output) cycle
    formatline = '(a/i1/a/i1/i1/i1/i1/'//trim(indexformat)//')'
    write(foutput,fmt=formatline) '$ElementNodeData',1,'"'//trim(var(m)%name)//'"',0,3,0,1,ubound(var(m)%funk,1)
    allocate(cellgradl(totaldimensions,1))
    do ns = 1, ubound(var(m)%funk,1)
      i = ijvar(m,ns)
      cellgradl(:,1) = varcgrad(i,m)
      if (debug) write(91,*) 'cell components with cellgrad: i = ',i,': ns = ',ns,': m = ',m,': var = ',trim(var(m)%name), &
        ': cellgrad = ',cellgradl(:,1)
      nnodes = gmsh_element(cell(i)%gmsh_type)%nnodes
      formatline = '('//trim(indexformat)//',a,i3'//repeat(',a,'//trim(floatformat),nnodes)//')'
      write(foutput,fmt=formatline) cell(i)%gmsh_element,' ',nnodes, &
        (' ',real(var(m)%funk(ns)%v+dot_product(cellgradl(:,1),node(cell(i)%knode(kk))%x-cell(i)%x)),kk=1,nnodes)
    end do
    deallocate(cellgradl)
    write(foutput,'(a)') '$EndElementNodeData'

  end if

end do

! cell centred var compounds, done as ElementNodeData unless element_data is on
if (.not.debug_dump) then
  do mc = 1, ubound(compound,1)
    if (compound(mc)%centring.ne."cell".or..not.compound(mc)%output) cycle
    nrank = ubound(compound(mc)%component,1)
    formatline = '(a/i1/a/i1/i1/i1/i1/'//trim(indexformat)//')'
!   write(foutput,fmt=formatline) '$ElementNodeData',1,'"'//trim(compound(mc)%name)//'"',0,3,0,nrank, &
    write(foutput,fmt=formatline) '$'//trim(element_data_string),1,'"'//trim(compound(mc)%name)//'"',0,3,0,nrank, &
      ubound(region(compound(mc)%region_number)%ij,1)
    do ns = 1, ubound(region(compound(mc)%region_number)%ij,1)
      i = region(compound(mc)%region_number)%ij(ns)
      if (debug) write(91,*) 'cell compound: i = ',i,': ns = ',ns,': mc = ',mc,': compound = ',trim(compound(mc)%name)
      allocate(cellvaluel(nrank))
      allocate(cellgradl(totaldimensions,nrank))
      do n = 1, nrank
        m = compound(mc)%component(n)
        if (m == 0) then
          cellvaluel(n) = 0.d0
          cellgradl(:,n) = [ 0.d0, 0.d0, 0.d0 ]
        else if (element_data) then
          cellvaluel(n) = var(m)%funk(ns)%v
          cellgradl(:,n) = [ 0.d0, 0.d0, 0.d0 ]
        else
          cellvaluel(n) = var(m)%funk(ns)%v
          cellgradl(:,n) = varcgrad(i,m)
        end if
  !     if (debug) write(*,*) 'for cell i = ',i,': and var = ',trim(var(m)%name),': cellgrad = ',cellgradl(:,n), &
  !       ': cellvalue = ',cellvaluel(n)
      end do
      if (element_data) then
        formatline = '('//trim(indexformat)//repeat(',a,'//trim(floatformat),nrank)//')'
        write(foutput,fmt=formatline) cell(i)%gmsh_element,(' ',real(cellvaluel(n)),n=1,nrank)
      else
        nnodes = gmsh_element(cell(i)%gmsh_type)%nnodes
        formatline = '('//trim(indexformat)//',a,i3'//repeat(',a,'//trim(floatformat),nnodes*nrank)//')'
        write(foutput,fmt=formatline) cell(i)%gmsh_element,' ',nnodes, &
          (( ' ', real(cellvaluel(n) + dot_product(cellgradl(:,n),node(cell(i)%knode(kk))%x-cell(i)%x )), &
          n=1,nrank),kk=1,nnodes)
      end if
      deallocate(cellvaluel,cellgradl)
    end do
    write(foutput,'(a)') '$End'//trim(element_data_string)
  end do
end if

! face centred var components, done as ElementData
do m = 1, ubound(var,1)
  if (var(m)%centring.ne."face".or.(.not.var(m)%output.and..not.debug_dump)) cycle
  formatline = '(a/i1/a/i1/i1/i1/i1/'//trim(indexformat)//')'
  write(foutput,fmt=formatline) '$ElementData',1,'"'//trim(var(m)%name)//'"',0,3,0,1,ubound(var(m)%funk,1)
  do ns = 1, ubound(var(m)%funk,1)
    j = ijvar(m,ns)
    if (debug) write(91,*) 'face components: j = ',j,': ns = ',ns,': m = ',m,': var = ',trim(var(m)%name)
    formatline = '('//trim(indexformat)//',a,'//trim(floatformat)//')'
    write(foutput,fmt=formatline) face(j)%gmsh_element,' ',real(var(m)%funk(ns)%v)
  end do
  write(foutput,'(a)') '$EndElementData'
end do

! face centred var compounds, done as ElementData
if (.not.debug_dump) then
  do mc = 1, ubound(compound,1)
    if (compound(mc)%centring.ne."face".or..not.compound(mc)%output) cycle
    nrank = ubound(compound(mc)%component,1)
    formatline = '(a/i1/a/i1/i1/i1/i1/'//trim(indexformat)//')'
    write(foutput,fmt=formatline) '$ElementData',1,'"'//trim(compound(mc)%name)//'"',0,3,0,nrank, &
      ubound(region(compound(mc)%region_number)%ij,1)
    do ns = 1, ubound(region(compound(mc)%region_number)%ij,1)
      j = region(compound(mc)%region_number)%ij(ns)
      if (debug) write(91,*) 'face compound: j = ',j,': ns = ',ns,': mc = ',mc,': compound = ',trim(compound(mc)%name)
      allocate(cellvaluel(nrank))
      do n = 1, nrank
        m = compound(mc)%component(n)
        if (m == 0) then
          cellvaluel(n) = 0.d0
        else
          cellvaluel(n) = var(m)%funk(ns)%v
        end if
  !     if (debug) write(*,*) 'for face j = ',j,': and var = ',trim(var(m)%name),': cellvalue = ',cellvaluel(n)
      end do
      formatline = '('//trim(indexformat)//repeat(',a,'//trim(floatformat),nrank)//')'
      write(foutput,fmt=formatline) face(j)%gmsh_element,(' ',real(cellvaluel(n)),n=1,nrank)
      deallocate(cellvaluel)
    end do
    write(foutput,'(a)') '$EndElementData'
  end do
end if

! none centred var components, done as Data
do m = 1, ubound(var,1)
  if (var(m)%centring.ne."none".or.(.not.var(m)%output.and..not.debug_dump)) cycle
  formatline = '(a/i1/a/i1/i1/i1/i1/'//trim(indexformat)//')'
  write(foutput,fmt=formatline) '$Data',1,'"'//trim(var(m)%name)//'"',0,3,0,1,1
  formatline = '('//trim(indexformat)//',a,'//trim(floatformat)//')'
  write(foutput,fmt=formatline) 1,' ',real(var(m)%funk(1)%v)
  write(foutput,'(a)') '$EndData'
  if (debug) write(91,*) 'none components: m = ',m,': var = ',trim(var(m)%name)
end do

! none centred var compounds, done as Data
if (.not.debug_dump) then
  do mc = 1, ubound(compound,1)
    if (compound(mc)%centring.ne."none".or..not.compound(mc)%output) cycle
    nrank = ubound(compound(mc)%component,1)
    formatline = '(a/i1/a/i1/i1/i1/i1/'//trim(indexformat)//')'
    write(foutput,fmt=formatline) '$Data',1,'"'//trim(compound(mc)%name)//'"',0,3,0,nrank,1
    if (debug) write(91,*) 'none compound: mc = ',mc,': compound = ',trim(compound(mc)%name)
    allocate(cellvaluel(nrank))
    do n = 1, nrank
      m = compound(mc)%component(n)
      if (m == 0) then
        cellvaluel(n) = 0.d0
      else
        cellvaluel(n) = var(m)%funk(1)%v
      end if
      if (debug) write(*,*) 'for none centred var = ',trim(var(m)%name),': cellvalue = ',cellvaluel(n)
    end do
    formatline = '('//trim(indexformat)//repeat(',a,'//trim(floatformat),nrank)//')'
    write(foutput,fmt=formatline) 1,(' ',real(cellvaluel(n)),n=1,nrank)
    deallocate(cellvaluel)
    write(foutput,'(a)') '$EndData'
  end do
end if

close(foutput)

if (debug) write(*,'(80(1h+)/a)') 'subroutine output_gmsh'

end subroutine output_gmsh

!-----------------------------------------------------------------

end module gmsh_module
!-----------------------------------------------------------------
