#!/usr/bin/perl -w
# file src_equations/setup_equations.pl
#
# Copyright 2009-2011 Dalton Harvie (daltonh@unimelb.edu.au)
# 
# This file is part of arb finite volume solver, referred to as `arb'.
# 
# arb is a software package designed to solve arbitrary partial
# differential equations on unstructured meshes using the finite volume
# method.  Primarily it consists of fortran source code, perl source
# code and shell scripts.  arb replies on certain third party software
# to run, most notably the computer algebra system maxima
# <http://maxima.sourceforge.net/> which is released under the GNU GPL.
# 
# The copyright of arb is held by Dalton Harvie.
# 
# arb is released under the GNU GPL.  arb is free software: you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License (version 3) as published by the Free Software Foundation.
# You should have received a copy of the GNU General Public License
# along with arb (see file license/gpl.txt after unpacking).  If not,
# see <http://www.gnu.org/licenses/>.
# 
# arb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# For full details of arb's license see the license directory.
# 
# The current homepage for the arb finite volume solver project is
# <http://www.chemeng.unimelb.edu.au/people/staff/daltonh/downloads/arb>.
#
#-------------------------------------------------------------------------
# perl script to create equation f90 module for arb

use strict;
use File::Basename;

my $version="0.25";

# now called from build directory
my $working_dir="..";
my $src_dir="$working_dir/src";
my $build_dir="$working_dir/build";
my $tmp_dir="$working_dir/tmp/setup";
my $tmp_file_number=0;
our $maxima_bin='maxima';

# create/clear out tmp directory
my ($systemcall);
if (! -d $tmp_dir) {
  $systemcall="mkdir -p $tmp_dir";
  (!(system("$systemcall"))) or do {print "could not $systemcall\n"; die;};
}
$systemcall="rm -f $tmp_dir/*";
(!(system("$systemcall"))) or do {print "could not $systemcall\n"; die;};

my $stopfile="$working_dir/stop";
if (-e $stopfile) {
  $systemcall="rm $stopfile";
  (!(system("$systemcall"))) or do {print "could not $systemcall\n"; die;};
}

open(DEBUG, ">$tmp_dir/debug");
print "\nperl solution_setup script to create f90 subroutines for arb\n";
print "dalton harvie, v$version\n\n";
print DEBUG "\nperl solution_setup script to create f90 subroutines for arb\n";
print DEBUG "dalton harvie, v$version\n\n";

my @input_files;
if (scalar(@ARGV)) {
  @input_files = @ARGV;
} else {
  @input_files = ("constants.in", "equations.in");
}
my $string="";
foreach (@input_files) { $string = $string." ".$_; }
print "INFO: input files being used are$string\n";

my %m=(); # is a hash of the maximum number of variables of each type
$m{"constant"}=0;
$m{"transient"}=0;
$m{"unknown"}=0;
$m{"derived"}=0;
$m{"equation"}=0;
$m{"output"}=0;
$m{"someloop"}=0;
$m{"system"}=0;
$m{"empty"}=0;
$m{"compound"}=0;
$m{"general"}=0;

my %variable=(); # is a hash/array/hash of all the variables with associated data of each type and number and datatype

my %sub_string=(); # is a hash of strings that we need to assemble for substitution in the equations.f90 routine

my %basename;
$basename{"constant"}="var";
$basename{"transient"}="var";
$basename{"unknown"}="var";
$basename{"derived"}="var";
$basename{"equation"}="var";
$basename{"output"}="var";
$basename{"compound"}="compound";
$basename{"someloop"}="someloop";
$basename{"system"}="sys";

my $lengthunit="m"; # do this better in the future

#--------------------------------------------------------------
# read through setup files storing all the information

create_system_variables(); # create variables generated by the system

read_input_files(); # read input files, generating user defined variables

check_equation_status(); # checks whether we need to proceed with equation construction

create_compounds(); # create vector and tensor allocations

create_mequations(); # create maxima type equations

create_allocations();  # create allocation statements

create_fortran_equations(); # generate fortran equations

write_sub_strings(); # write out sub_strings and while creating new fortran files

#write_latex(); # produce a latexable summary file

print "SUCCESS: equations_module.f90 has been created\n";
print DEBUG "SUCCESS: equations_module.f90 has been created\n";

# finally move equation data to build directory as a record of the successful run
$systemcall="mv $tmp_dir/setup_equation_data $build_dir/last_setup_equation_data";
(!(system("$systemcall"))) or do {print "could not $systemcall\n"; die;};
print DEBUG "INFO: moving setup_equation_data to build directory\n";

close(DEBUG);

exit;

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
# run through each variable checking whether anything has changed that requires the script to be rerun

sub check_equation_status {

  my ($type, $mvar, $component, $line);
  my $same=1; # flag to indicate whether this run and last are the same

  open(NEW_CHECK, ">$tmp_dir/setup_equation_data");
  open(OLD_CHECK, "<$build_dir/last_setup_equation_data") or $same=0; # if file doesn't exist

  foreach $type (sort(keys %variable)) {
    foreach $mvar ( 1 .. $m{$type} ) {
      foreach $component ("name","centring","units","multiplier","magnitude","region","equation","options","output","deriv") {
        if ($variable{$type}[$mvar]{$component}) {
          print NEW_CHECK $variable{$type}[$mvar]{$component}."\n";
          if ($same) {
            if ($line=<OLD_CHECK>) {
              chompm($line);
              if ($line ne $variable{$type}[$mvar]{$component}) {
                print DEBUG "Changed data is $line and $variable{$type}[$mvar]{$component} for: type = $type: mvar = $mvar: component = $component\n";
                $same = 0;
              }
            } else { $same = 0;}
          }
        }
      }
    }
  }
  close(NEW_CHECK);
  close(OLD_CHECK);
  if ($same) {
    print "INFO: setup_equations data has not changed since last run so exiting immediately\n";
    print DEBUG "INFO: setup_equations data not has changed since last run so exiting immediately\n";
    print "SUCCESS: equations_module.f90 was already up-to-date\n";
    print DEBUG "SUCCESS: equations_module.f90 was already up-to-date\n";
    exit;
  } else {
    print "INFO: setup_equations data has changed since last run: recreating fortran\n";
    print DEBUG "INFO: setup_equations data has changed since last run: recreating fortran\n";
  }

}

#-------------------------------------------------------------------------------
# now write out sub_strings to new fortran files

sub write_sub_strings {

  use strict;
  my @f90_files=glob("$src_dir/*_template.f90");
  my ($keyword, $indent, $line, $string, @strings, $n);
  my $linelength = 100;

  foreach my $in_file (@f90_files) {

    open(INFILE, "<$in_file");
    basename($in_file) =~ /_template\.f90$/;
    my $out_file=$build_dir."/$`.f90";

    open(OUTFILE, ">$out_file"); #overwrite any existing files
    print "INFO: writing from equation src file $in_file to output file $out_file\n";

    while ($line=<INFILE>) { chompm($line);
      if (($keyword) = $line =~ /^\s*?<sub_string:(.+)>/) {
        print "INFO: found sub_string marker in fortran file with keyword = $keyword\n";
        if (!($sub_string{$keyword})) {
          print "NOTE: sub_string of keyword $keyword not set\n";
          $line =~ s/^(\s*)/!$1/; # just comment out marker
        } else {
          ($indent) = $line =~ /^(\s*?)</;
          @strings = split("\n",$sub_string{$keyword});
          $sub_string{$keyword} = "";
          print DEBUG "sub_string{keyword} = $sub_string{$keyword}\n";
          foreach $string (@strings) {
            $string =~ s/^(\s*)//; # remove any space from front of string
#           print DEBUG "string = |$string|\n";
            if (!($string =~ /^!/)) {
              if ($string =~ /^(\s*)end(\s*?)if/ || $string =~ /^(\s*)end(\s*?)do/) {substr($indent,-2,2,"");}  # decrease indent
              if ($string =~ /^(\s*)else/) {substr($indent,-2,2,"");}  # decrease indent
# split string every $linelength characters and put in continuation mark if not a comment
              $n=1;
              while (length($string) > $n*$linelength ) { # $#string should be the number of the last element of $string
                substr($string,$n*$linelength-1,0,"&\n  $indent&");  # cut string and put in &, CR with extra indent
                $n++;
              }
              $string =~ s/^/$indent/;
              if ($string =~ /^(\s*)if (.+?)then/ || $string =~ /^(\s*)do /) {$indent = $indent.'  ';} # increase indent
              if ($string =~ /^(\s*)else/) {$indent = $indent.'  ';} # increase indent
            } 
            $sub_string{$keyword} = $sub_string{$keyword}.$string."\n" # comments are printed with no indent
          }
          
          $line =~ s/^(\s*?)<sub_string:$keyword>/$sub_string{$keyword}/;
        }
      }
      print OUTFILE "$line\n";
    }

    close(INFILE);
    close(OUTFILE);
  }
}

#--------------------------------------------------------------
# read through constants.in and equations.in files storing all info

sub read_input_files {

  use strict;
  my ($file, $oline, $line, $type, $name, $cunits, $units, $multiplier, $expression, $boundary_region, $mvar,
    $mcheck, $typecheck, $tmp, $keyword, $centring, $region, $sign, $magnitude, $region_list, $region_constant);

  foreach $file ( @input_files ) {

    $region_list = 0; # number of regions listed in last region list

    open(FILE, "<$working_dir/$file");
    while ($oline=<FILE>) { chompm($oline);  # remove linefeed from end

      $line = $oline;

      if ($line =~ /^\s*VERSION\s+(\S*)/) {
        if ($1 ne $version) {die "ERROR: version mismatch between $file and setup_equations.pl\n".
                                 "  You can increase the version number in $file (from $1 to $version) on the proviso that ".
                                 "the syntax in this file is still current\n";
        }
      }

# type and name
      if ($line =~ /^(CELL_|FACE_|NONE_|)(CONSTANT|REGION_CONSTANT|TRANSIENT|DERIVED|UNKNOWN|EQUATION|OUTPUT|REGION_LIST|INDEPENDENT|DEPENDENT)/) {
        $line = $';
        $type = "\L$2";
        $centring = "";
        if ($1) {
          $centring = "\L$1";
          ($centring) = $centring =~ /^(\S+?)_/;
        }
        if ($type eq "independent") { print "WARNING: INDEPENDENT type has been depreciated, use UNKNOWN instead\n"; $type = "unknown"; }
        if ($type eq "field") { print "WARNING: FIELD type has been depreciated, use UNKNOWN instead\n"; $type = "unknown"; }
        if ($type eq "dependent") { print "WARNING: DEPENDENT type has been depreciated, use DERIVED instead\n"; $type = "derived"; }
        if (!($centring)) {
          if ( $type eq "unknown" ) {
            $centring = "cell";
          } else {
            $centring = "none";
          }
        }
        if ($type eq "region_list") {
          if (!($file eq $input_files[0])) { die "ERROR: REGION_LIST can only be used in the constants file, here $input_files[0]\n"; }
          $region_list = 0;
          while ($line =~ /^\s*<.+?>\s+/) {
            $region_list++;
            $line = $';
          }
          print "INFO: found $centring region_list with $region_list elements\n";
          print DEBUG "INFO: found $centring region_list with $region_list elements\n";
          next;
        }
        $region_constant = 0;
        if ($type eq "region_constant") {
          $type = "constant";
          if (!($region_list)) {die "ERROR: a $centring region_constant appears before a region_list has been defined in $file:\n$oline\n";}
          $region_constant = 1;
        }
          
        if (!($type)) {die "ERROR: problem reading in type from following line of $file:\n$oline\n";}
        if (!($centring)) {die "ERROR: problem reading in centring from following line of $file:\n$oline\n";}

#       ($name,$line) = $line =~ /^\S+?\s+?(<.+?>)\s*(.*)/;  #\s is a space, \S is a nonspace, . is a single wildcard, + is 1 or more times, ? is non greedy
        $line =~ /^\s*(<.+?>)\s+/;
        $name = $1; $line = $';
        if (!($name)) {die "ERROR: problem reading in name from following line of $file:\n$oline\n";}
        print DEBUG "found in input file: name = $name: type = $type: centring = $centring\n";
# see if this name has already been defined
        $mvar = 0;
        foreach $typecheck (sort(keys %variable)) {
          foreach $mcheck ( 1 .. $m{$typecheck} ) {
            if ($variable{$typecheck}[$mcheck]{"name"} eq $name) {
              print "ERROR: trying to define a $type variable $name which has already been defined as a $typecheck variable\n";
              die;
# TODO: possibility of rereading in same variable with different options?
#             if ($type ne $typecheck) {
#               print "ERROR: trying to define a $type variable $name which has already been defined as a $typecheck variable\n";
#               die;
#             } else {
#               print "CAUTION: reading in additional information about $type $name\n";
#               $mvar = $mcheck;
#               last;
#             }
            }
          }
        }
# otherwise create a new variable
        if (!($mvar)) { $m{$type}++; $mvar=$m{$type}; }
        $variable{$type}[$mvar]{"name"}=$name;
        $variable{$type}[$mvar]{"centring"}=$centring;

      } else { next; }

# units and multiplier (optional)
# first determine whether any specification has been made
      if ($line =~ /^\[.*?\]/) {
# units specification is present to determine what they are and strip them from the front of $line
        ($cunits,$line) = $line =~ /^\[(.*?)\]\s*(.*)$/;  #\s is a space, \S is a nonspace, . is a single wildcard, + is 1 or more times, ? is non greedy
# split units if multiplier is present
        if ($cunits =~ /\*/) { ($multiplier,$units) = $cunits =~ /(.*?)\*(.*)/; }
        else { $multiplier=""; $units=$cunits;}
        $multiplier =~ s/e/d/; # convert single and floats to double precision
        $variable{$type}[$mvar]{"units"}=$units;
        $variable{$type}[$mvar]{"multiplier"}=$multiplier;
      }
      if (!($variable{$type}[$mvar]{"units"})) { $variable{$type}[$mvar]{"units"} = "1"; }
      if (!($variable{$type}[$mvar]{"multiplier"})) { $variable{$type}[$mvar]{"multiplier"} = "1.d0"; }

# magnitude - unknown variables require a magnitude
      if ($type eq 'unknown' ) {
        ($magnitude,$line) = $line =~ /^(\S+)\s*(.*)$/;  #\s is a space, \S is a nonspace, . is a single wildcard, + is 1 or more times, ? is non greedy
        if (!($magnitude)) {die "ERROR: problem reading in $type magnitude from following line of $file:\n$oline\n"; }
        $variable{$type}[$mvar]{"magnitude"} = $magnitude;
      } else {
        $variable{$type}[$mvar]{"magnitude"} = "";
      }

# expression = equation or initial_condition - all strings which are enclosed in double quotes
      if ( $type eq "constant" && $file eq $input_files[0] ) { # only constants listed in equation file have expressions and are quoted
        if ($region_constant) {
          $line =~ /^(\S+\s+){$region_list}/; # remove constant values
          $line = $';
        } else {
          $line =~ /^(\S+\s+)/; # remove constant values
          $line = $';
        }
      } else {
        if ($line =~ /^".+"/) { # see if expression is present
          ($expression,$line) = $line =~ /^"(.+)"\s*(.*)$/;  #\s is a space, \S is a nonspace, . is a single wildcard, + is 1 or more times, ? is non greedy
          if (!($expression)) {die "ERROR: problem reading in expression from following line of $file:\n$oline\n"; }
          $variable{$type}[$mvar]{"equation"} = $expression;
          $variable{$type}[$mvar]{"otype"} = $type; # set original and type and mvar for this expression
          $variable{$type}[$mvar]{"omvar"} = $mvar;
        }
      }

# set region
      $variable{$type}[$mvar]{"region"} = "";
      if ($centring ne "none") {
        if ($line =~ /^ON/) { # var is located on a single region
#           ($variable{$type}[$mvar]{"region"}) = $line =~ /^ON\s+(<.+?>)/;
          ($variable{$type}[$mvar]{"region"},$line) = $line =~ /^ON\s+(<.+?>)\s*(.*)$/;
          if (!($variable{$type}[$mvar]{"region"})) { print "ERROR: problem reading in region name from line $line\n"; }
        } elsif ($centring eq "cell") {
          if ($type eq "equation") {
            $variable{$type}[$mvar]{"region"} = "<domain>"; # default cell region for equations
          } else {
            $variable{$type}[$mvar]{"region"} = "<all cells>"; # default cell region
          }
        } elsif ($centring eq "face") {
          if ($type eq "equation") {
            $variable{$type}[$mvar]{"region"} = "<boundaries>"; # default face region for equations
          } else {
            $variable{$type}[$mvar]{"region"} = "<all faces>"; # default face region
          }
        }
        print "INFO: $type $variable{$type}[$mvar]{name} has $variable{$type}[$mvar]{centring} region = $variable{$type}[$mvar]{region}\n"; 
        print DEBUG "$type $variable{$type}[$mvar]{name} has $variable{$type}[$mvar]{centring} region = $variable{$type}[$mvar]{region}\n"; 
      }

# any options (optional) will come last
# possibilities include:
#  noderivative - for CELL_DERIVED or FACE_DERIVED
#  harmonic - harmonic averaging for spatial interpolation to face centres
#  positive - variable must be positive
#  negative - variable must be negative
#  compoundoutput/nocompoundoutput - for (CELL|FACE)_(DERIVED) or FIELD, if this is a component of a compound variable (vector or tensor), do not output that compound
#  componentoutput/nocomponentoutput - for (CELL|FACE)_(DERIVED) or FIELD, if this is a component of a compound variable (vector or tensor), output this component of the compound separately
      if (!($line =~ /^#/) && $line =~ /^(\S+)/) { # if remainder is not a comment but there is something there
        ($variable{$type}[$mvar]{"options"}, $line) = $line =~ /^(\S+)\s*(.*)$/;  # note, options may not be a list of options at all, may be rubbish, numbers etc
        print "INFO: $type $variable{$type}[$mvar]{name} has loaded (unchecked) options of $variable{$type}[$mvar]{options}\n";
      }

# any comments
      if ($line =~ /^.*?#\s*(.+)/) { $variable{$type}[$mvar]{"comments"} = $1; } else {$variable{$type}[$mvar]{"comments"} = "";}

# set output options for individual components (default is off - compounds are on by default for some variables)
      $variable{$type}[$mvar]{"output"} = 0;
      if ($variable{$type}[$mvar]{"options"} && $variable{$type}[$mvar]{"options"} =~ /(^|\,)componentoutput(\,|$)/) {
        $variable{$type}[$mvar]{"output"} = 1;
      }

# set deriv
      if ($variable{$type}[$mvar]{"options"} && $variable{$type}[$mvar]{"options"} =~ /(^|\,)noderivative(\,|$)/) {
        $variable{$type}[$mvar]{"deriv"} = 0; # no derivative specified as an option
      } elsif ($type eq "derived" || $type eq "equation") {
        $variable{$type}[$mvar]{"deriv"} = 1; # otherwise all derived functions and equations require a derivative calculation
      } else {
        $variable{$type}[$mvar]{"deriv"} = 0;
      }

      print "INFO: read in $type [$mvar]: name = $name: centring = $variable{$type}[$mvar]{centring}: deriv = $variable{$type}[$mvar]{deriv}\n";

    }
    close(FILE);
  }

#-------------------------
# now run through allocating number of general var variables and creating their names

  foreach $type ("constant","transient","unknown","derived","equation","output") {
    foreach $mvar ( 1 .. $m{$type} ) {
      $m{"general"}++;

# create maxima names for the variable and fortran names as used in ze code
      if ($variable{$type}[$mvar]{"centring"} eq "cell") {
        $variable{$type}[$mvar]{"maxima"}=$basename{$type}."[i,$m{general}]";
        $variable{$type}[$mvar]{"fortran"}="var($m{general})%funk(nsvar($m{general},i))%v";
      } elsif ($variable{$type}[$mvar]{"centring"} eq "face") {
        $variable{$type}[$mvar]{"maxima"}=$basename{$type}."[j,$m{general}]";
        $variable{$type}[$mvar]{"fortran"}="var($m{general})%funk(nsvar($m{general},j))%v";
      } else { # none
        $variable{$type}[$mvar]{"maxima"}=$basename{$type}."[$m{general}]";
        $variable{$type}[$mvar]{"fortran"}="var($m{general})%funk(1)%v";
      } 
# create mfortran names for the variable, which are what maxima produces from the f90 package
      $variable{$type}[$mvar]{"mfortran"}=$variable{$type}[$mvar]{"maxima"}; # only differ by braces used on indicies
      $variable{$type}[$mvar]{"mfortran"}=~s/\[/(/;
      $variable{$type}[$mvar]{"mfortran"}=~s/\]/)/;
      $variable{$type}[$mvar]{"fortranns"}="var($m{general})%funk(ns)%v"; # fortran name referenced by ns rather than location
      $variable{$type}[$mvar]{"fortran_number"} = $m{"general"}; # fortran var arrar element number

    }
  }
#-------------------------
}

#-------------------------------------------------------------------------------
# chomp and remove mac linefeads too if present

sub chompm {
  use strict;
  chomp($_[0]);  # remove linefeed from end
  $_[0]=~tr/\r//d; # remove control m from mac files
}

#-------------------------------------------------------------------------------
# here we run through all equation expressions substituting in maxima names and 
#  setting correct variable locations - create mequations

sub create_mequations {

  use strict;
  my ($type, $mvar, $otype, $omvar, $name);

  print DEBUG "in create_mequations\n";

# first perform maxima name substitutions
  print "INFO: doing maxima name substitutions for user-supplied variables to create mequations\n";
  foreach $type ("constant","transient","derived","equation","unknown","output") {
    foreach $mvar ( 1 .. $m{$type} ) {
      $variable{$type}[$mvar]{"mequation"}="";
      if (!($variable{$type}[$mvar]{"equation"})) { next; }
      $variable{$type}[$mvar]{"mequation"}=$variable{$type}[$mvar]{"equation"};
      name_to_maxima($variable{$type}[$mvar]{"mequation"});
    }
  }

# now look for operators, do location interpolation and use maxima to simplify the expressions
  foreach $type ("constant","transient","unknown","derived","equation","output","someloop") {
    $mvar = 1;
    while ($m{$type} && $mvar <= $m{$type}) {
      if (!($variable{$type}[$mvar]{"mequation"})) { $mvar++; next; }
      if ($type eq "someloop") {
        $otype = $variable{$type}[$mvar]{"otype"};
        $omvar = $variable{$type}[$mvar]{"omvar"};
        $name = "originating from $otype variable $variable{$otype}[$omvar]{name}";
      } else {
        $name = "$variable{$type}[$mvar]{name}";
      }
      print "INFO: doing mequation interpolation for $type [$mvar]: $name\n";
# do location-specific changes
      mequation_interpolation($variable{$type}[$mvar]{"mequation"},$variable{$type}[$mvar]{"centring"},$variable{$type}[$mvar]{"deriv"},$variable{$type}[$mvar]{"otype"},$variable{$type}[$mvar]{"omvar"});
# run through maxima performing any simplifications
      run_maxima_simplify($variable{$type}[$mvar]{"mequation"},$variable{$type}[$mvar]{"otype"},$variable{$type}[$mvar]{"omvar"});
      $mvar++;
    }
  }

  print DEBUG "INFO: generated mequations\n";
  foreach $type ("constant","transient","unknown","derived","equation","output","someloop") {
    foreach $mvar ( 1 .. $m{$type} ) {
      if (!($variable{$type}[$mvar]{"mequation"})) { next; }
      print DEBUG "\n$type $mvar $variable{$type}[$mvar]{name} $variable{$type}[$mvar]{centring}\nequation: $variable{$type}[$mvar]{equation}\nmequation: $variable{$type}[$mvar]{mequation}\n";
    }
  }

  print DEBUG "leaving create_mequations\n";
}

#-------------------------------------------------------------------------------
# processes a mequation, given
# -> $_[0] contains mequation to process
# -> $_[1] is the centring that we want the mequation to have
# -> $_[2] is eqderiv (=0,1), which is whether a derivative is required for this mequation
# -> $_[3] is otype, which is the type for the original expresion that this one derived from
# -> $_[4] is omvar, which is the mvar for the original expresion that this one derived from
# <- $_[0] is returns the modified mequation

sub mequation_interpolation {

  use strict;
  use Text::Balanced qw ( extract_bracketed );
  my ($type, $mvar, $operator, $tmp, $n, $l, $pre, $post, $eqcentring, $centring, $options,
    $eqderiv, $someregion, $flux, $default, $limiter, $deriv, $quantity, $condition, $true, $false, $dquantity,
    $bound, $previous, $sum, $product, $otype, $omvar);
  my $nbits=0;
  my @outbit=();
  my @inbit=();

  $eqcentring = $_[1];
  $eqderiv = $_[2];
  $otype = $_[3];
  $omvar = $_[4];
  print DEBUG "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n".
              "in mequation_interpolation before: eqcentring = $eqcentring: _[0] = ",$_[0],"\n";

#-------------------------------------------------
# look for any operators and deal with their contents by creating new someloop variables

# while ($_[0] =~ m/(face|cell|none)(div|grad|divgrad|ave|limiter|if|sum|max|min)(\[(.*?)\]){0,1}\(/) { #
  while ($_[0] =~ m/(face|cell|none)(div|grad|divgrad|ave|limiter|if|sum|max|min|product|delta)(\[(.*?)\]){0,1}\(/) { #
    $operator = $&; # $& holds complete regex match
    $outbit[$nbits] = $`; # $` holds before match
    $_[0] = "(".$'; # $' holds after match
#   if ($1) {$centring = $1;} else {$centring = "none";}
    $centring = $1;
    $type = $2;
    if ($3) {$options = $4;} else {$options = "";}
    $operator =~ s/\($//; # remove trailing brace from operator
    ($inbit[$nbits],$_[0]) = extract_bracketed( $_[0], '()' ); # http://search.cpan.org/~adamk/Text-Balanced-2.02/lib/Text/Balanced.pm
    if (!($inbit[$nbits])) { die "ERROR: the $operator operator does not have matching parenthesis\n"; }
    ($inbit[$nbits]) = $inbit[$nbits] =~ /^\((.*)\)$/; # remove surrounding braces
    print DEBUG "contains a $centring centred $type operator with options $options and contents: $inbit[$nbits]\n";

# check consistency of centring
# for these operators the centring should correspond to eqcentring
#   if ($type =~ /^(grad|div|graddiv|ave)$/ && ($centring eq "none" || $eqcentring eq "none")) {
    if ($type =~ /^(grad|div|graddiv|ave)|delta$/ && ($centring eq "none" || $eqcentring eq "none")) {
      die "ERROR: $eqcentring equation centring and $centring operator centring are inconsistent with $type operator\n";
    }

# for these operators the centring of the operator (if specified) should correspond to the centring of the equation
#   if ($type =~ /^(grad|div|graddiv|ave|limiter|if)$/ && $centring ne "none" && $eqcentring ne "none" && $centring ne $eqcentring) {
    if ($type =~ /^(grad|div|graddiv|ave|limiter|if|delta)$/ && $centring ne "none" && $eqcentring ne "none" && $centring ne $eqcentring) {
# if centring is incorrect, average operator to correct centring and try again
      $_[0] = $eqcentring."ave(".$operator."(".$inbit[$nbits]."))".$outbit[$nbits];
      $inbit[$nbits] = 0;
      $outbit[$nbits] = 0;
      print DEBUG "$centring centring was inconsistent with $eqcentring eqcentring $type operator : averaging and resetting to _[0] = $_[0]\n";
      print "WARNING: averaging $centring$type operator to $eqcentring centring for $type operator (do you really want this?)\n";
      next;
    }
      
# looks for generic option of noderivative in this operator
    $deriv = $eqderiv;
    if ($options && $options =~ /(^|\,)noderivative(\,|$)/) { $deriv = 0; }

# process each type of operator

# div is an operator representing the divergence which acts over surrounding faces of a cell (which must be within the domain)
    if ($type eq "div") {
      if ($centring ne "cell") {die "ERROR: $type operator has incorrect $centring centring (do you mean celldiv?)\n";}
      $inbit[$nbits] = 'divop[i,j]*('.$inbit[$nbits].')';
      create_someloop($inbit[$nbits],"sum","face","<nobcelljfaces>",$deriv,$otype,$omvar);
      
# grad operators
    } elsif ($type eq "grad") {

# cellgrad[l=1|2|3] is an operator representing the gradient in the cell, evaluated using surround cell values
      if ($centring eq "cell") {
        if ($options && $options =~ /(^|\,)\s*l\s*=\s*(\d)(\,|$)/) { $l = $2 } else { $l = 0; }
        if (!( $l >= 1 && $l <= 3)) { die "ERROR: $centring $type operator has an incorrect or unspecified direction l=$l (should be 1|2|3)\n";}
        $inbit[$nbits] = "cellkernel[ilast,$l,ns]*(".$inbit[$nbits].')';
        $someregion = '<cellkernelregion[l='.$l.']>';
        create_someloop($inbit[$nbits],"sum","cell",$someregion,$deriv,$otype,$omvar);

# facegrad is an operator representing the gradient of a quantity taken in coordinate (l=1,2,3) or face orientated (l=4,5,6) directions
      } elsif ($centring eq "face") {
        if ($options && $options =~ /(^|\,)\s*l\s*=\s*(\d)(\,|$)/) { $l = $2 } else { $l = 4; }
        if (!( $l >= 1 && $l <= 6)) { die "ERROR: $centring $type operator has an incorrect direction l=$l (should be 1|2|3|4|5|6)\n";}
        $inbit[$nbits] = 'facekernel[j,'.$l.',ns]*('.$inbit[$nbits].')';
        $someregion = '<facekernelregion[l='.$l.']>';
        create_someloop($inbit[$nbits],"sum","cell",$someregion,$deriv,$otype,$omvar);

      }
 
# celldivgrad[l=1|2|3] is an operator representing the gradient in the cell, evaluated using the divergence of surround face values
    } elsif ($type eq "divgrad") {
      if ($centring ne "cell") {die "ERROR: a $centring$type operator does not exist (do you mean celldivgrad?)\n";}
      if ($options && $options =~ /(^|\,)\s*l\s*=\s*(\d)(\,|$)/) { $l = $2 } else { $l = 0; }
      if (!( $l >= 1 && $l <= 3)) { die "ERROR: $centring $type operator has incorrect or unspecified direction l=$l\n";}
      $inbit[$nbits] = "divop[i,j]*facenorm[j,$l]*(".$inbit[$nbits].')';
      create_someloop($inbit[$nbits],"sum","face","<nobcelljfaces>",$deriv,$otype,$omvar);
 
# ave
    } elsif ($type eq "ave") {

# cellave is an operator that takes face centred quanities and averages to cell centres
      if ($centring eq "cell") {

# lastface averaging - suppress interpolation to current cell centre
        if ($options && $options =~ /(^|\,)lastface(\,|$)/) {
          create_someloop($inbit[$nbits],"sum","face","<noloop>",$deriv,$otype,$omvar);

# normal arithmetic average
        } else {
          $inbit[$nbits] = "cellkernel[i,0,ns]*(".$inbit[$nbits].')';
          create_someloop($inbit[$nbits],"sum","face","<cellkernelregion[l=0]>",$deriv,$otype,$omvar);
        }

# faceave is an operator that takes cell centred quanities and averages to face centres
      } elsif ($centring eq "face") {

# advection averaging - takes a cell centred quantity expression, a face centred flux expression, and optionally a cell centred gradient limiter (0->1) expression
        if ($options && $options =~ /(^|\,)advection(\,|$)/) {

          $quantity = split_off_leading_mequation($inbit[$nbits]);
          if ($quantity eq "") { die "ERROR: quantity expression in faceave advection operator not found\n"; }
          $flux = split_off_leading_mequation($inbit[$nbits]);
          if ($flux eq "") { die "ERROR: flux expression in faceave advection operator not found\n"; }
          mequation_interpolation($flux,"face",0,$otype,$omvar); # convert flux to face centred as will not become part of someloop - deriv not needed
          $limiter = split_off_leading_mequation($inbit[$nbits]);
          if ($limiter eq "") { $limiter = "1"; print "INFO: using default limiter = 1 for advection of $quantity\n"; } # default limiter is 1
          print DEBUG "advection operator components found: quantity = $quantity: flux = $flux: limiter = $limiter\n";

          $inbit[$nbits] = $quantity; # assemble expression
          if ($limiter) {
            foreach $l ( 1 .. 3 ) {
              $inbit[$nbits] = $inbit[$nbits].
                "+".$limiter."*cellsum(cellkernel[ilast,$l,ns]*(".$quantity.'),<cellkernelregion[l='.$l.']>)'.
                "*(cellave[lastface](facex[j,$l])-cellx[i,$l])";
            }
          }
          create_someloop($inbit[$nbits],"sum","cell","<upwindfaceicells>",$deriv,$otype,$omvar);
          $variable{"someloop"}[$m{"someloop"}]{"flux"} = $flux; # save flux for evaluating upwindfaceicells region
          print DEBUG "setting flux = $flux for someloop $m{someloop}\n";

        } elsif ($options && ( $options =~ /(^|\,)harmonic(\,|$)/ || $options =~ /(^|\,)limitedharmonic(\,|$)/ )) {
# harmonic averaging for general diffusivities
          $sum = $inbit[$nbits];
          create_someloop($sum,"sum","cell","<adjacentfaceicells>",$deriv,$otype,$omvar);
          $product = $inbit[$nbits];
          create_someloop($product,"product","cell","<adjacentfaceicells>",$deriv,$otype,$omvar);
          if ($options =~ /(^|\,)harmonic(\,|$)/) {
            $inbit[$nbits] = "2.*$product/$sum";
          } else {
# this needs fixing as it isn't continuous right now as a value goe to zero
            print "WARNING: implementation of limitedharmonic only experimental - needs to be continuous\n";
            $inbit[$nbits] = "faceif($sum-tiny,2.*$product/$sum,$sum)"; # allow for zero values within harmonic averaging but contents must be positive
            create_someloop($inbit[$nbits],"sum","face","<noloop>",$deriv,$otype,$omvar);
          }

# lastcell averaging - suppress interpolation to current face centre
        } elsif ($options && $options =~ /(^|\,)lastcell(\,|$)/) {
          create_someloop($inbit[$nbits],"sum","cell","<noloop>",$deriv,$otype,$omvar);

# normal arithmetic average
        } else {
          $inbit[$nbits] = "facekernel[j,0,ns]*(".$inbit[$nbits].')';
          create_someloop($inbit[$nbits],"sum","cell","<facekernelregion[l=0]>",$deriv,$otype,$omvar);
        }

      }

# celllimiter
    } elsif ($type eq "limiter") {
      if ($centring ne "cell") {die "ERROR: $type operator has incorrect $centring centring (do you mean celllimiter?)\n";}

      $quantity = split_off_leading_mequation($inbit[$nbits]);
      if ($quantity eq "") { die "ERROR: no quantity specified in celllimiter\n"; }

      if ($options && ( $options =~ /(^|\,)upper(\,|$)/ || $options =~ /(^|\,)lower(\,|$)/ )) {
        if ($options =~ /(^|\,)upper(\,|$)/ && $options =~ /(^|\,)lower(\,|$)/) {
          die "ERROR: either upper or lower bound must be specified in celllimiter options, not both\n";
        }
        $bound = split_off_leading_mequation($inbit[$nbits]);
        if ($bound eq "") { die "ERROR: bound specified but none given in celllimiter\n"; }
        if ($options =~ /(^|\,)upper(\,|$)/) {
          $bound = "-(".$bound.")";
          $quantity = "-(".$quantity.")";
        }
      } else { $bound = "0.d0"; }

      $previous = split_off_leading_mequation($inbit[$nbits]);
      if ($previous eq "") { $previous = "1.d0"; }

# form expression for change in quantity between cell centre and face
      $dquantity = "";
      foreach $l ( 1 .. 3 ) {
        $dquantity = $dquantity.
          "+faceave[lastcell](cellsum(cellkernel[ilast,$l,ns]*(".$quantity.'),<cellkernelregion[l='.$l.']>))'.
          "*(facex[j,$l]-faceave[lastcell](cellx[i,$l]))";
      }

      if ($options && $options =~ /(^|\,)linear(\,|$)/) {
# linear function
        $inbit[$nbits] = 'cellmin(cellmin(('.$bound.')-('.$quantity.'),0.d0)/'.
          '(facemin('.$dquantity.',<celljfaces>)-limitertolerance),'.$previous.')';
      } else {
# continuous function
        $inbit[$nbits] = 'cellmin(cellmin(('.$bound.')-('.$quantity.'),0.d0)/'.
          '(facemin('.$dquantity.',<celljfaces>)-limitertolerance),limitercontgrad)';
        $inbit[$nbits] = 'cellmin((-cos('.$inbit[$nbits].'*pi/limitercontgrad)+1.d0)/2.d0,'.$previous.')';
      }

      create_someloop($inbit[$nbits],"sum","cell","<noloop>",$deriv,$otype,$omvar);

# if statement 
    } elsif ($type eq "if") {

      $condition = split_off_leading_mequation($inbit[$nbits]);
      if ($condition eq "") { die "ERROR: condition expression in if operator not found\n"; }
      $true = split_off_leading_mequation($inbit[$nbits]);
      if ($true eq "") { die "ERROR: true expression in if operator not found\n"; }
      $false = split_off_leading_mequation($inbit[$nbits]);
      if ($false eq "") { die "ERROR: false expression in if operator not found\n"; }

      print DEBUG "in if statement found: condition = $condition: true = $true: false = $false\n";

# interpolate true and false conditions to centring and save
      mequation_interpolation($true,$centring,$deriv,$otype,$omvar);
      mequation_interpolation($false,$centring,$deriv,$otype,$omvar);

      $inbit[$nbits] = $condition; # main expression is condition (giving the true value if >=0, else the false value)
      create_someloop($inbit[$nbits],$type,$centring,"<noloop>",0,$otype,$omvar);

# save true and false conditions into recently created someloop
      $variable{"someloop"}[$m{"someloop"}]{"false"} = $false;
      $variable{"someloop"}[$m{"someloop"}]{"true"} = $true;
# save deriv for these values which is different to deriv used for condition (=0)
      $variable{"someloop"}[$m{"someloop"}]{"valuederiv"} = $deriv;

# sum|max|min|product
# centring for these is where the contents needs to be located, rather than where the operator is located
    } elsif ($type =~ /^(sum|max|min|product)$/) {

# first argument is always the equation
# if a min|max:
#   if one argument is present default value and <noloop> is assumed
#   if two arguments are present, second argument is a default value and <noloop> is assumed (ie, max|min of two quantities)
#   if three arguments are present, third is region and second may be left empty for a default value to be used
# if a sum|product
#   if one argument is present then default regions are applied depending on the operator
#   if a second argument is present then it is the region

      $quantity = split_off_leading_mequation($inbit[$nbits]);
      if ($quantity eq "") {
        print "ERROR: quanitity for $operator in $variable{$otype}[$omvar]{name} is empty:\n".
          "  quantity = $quantity\n  remainder = $inbit[$nbits]\n";
        print DEBUG "ERROR: quanitity for $operator in $variable{$otype}[$omvar]{name} is empty\n".
          "  quantity = $quantity\n  remainder = $inbit[$nbits]\n";
        die;
      }
      
      $default = 0;
      if ($type eq "min" || $type eq "max") {
        $default = split_off_leading_mequation($inbit[$nbits]);
        if ($default eq "" && $type eq "min") {
#         $default = "1.d+20";
          $default = "+huge";
        } elsif ($default eq "" && $type eq "max") {
#         $default = "-1.d+20";
          $default = "-huge";
        } else {
          mequation_interpolation($default,$eqcentring,$deriv,$otype,$omvar); # interpolate default to current location
        }
        print DEBUG "default value will be: $default\n";
      }

      $someregion = split_off_leading_mequation($inbit[$nbits]);
      if ( $someregion =~ /^\<.+?\>$/ ) {
        print DEBUG "INFO: location of $someregion has been found to $centring centred $type operator used in $otype variable $variable{$otype}[$omvar]{name}\n";
      } elsif ($someregion eq "") {
        if (($type eq "sum" || $type eq "product") && $eqcentring eq "cell" && $centring eq "face") {
          $someregion = "<celljfaces>"; # default face loop around a cell
        } elsif (($type eq "sum" || $type eq "product") && $eqcentring eq "cell" && $centring eq "cell") {
          $someregion = "<adjacentcellicells>"; # default cell loop around a cell is just adjacent faces
        } elsif (($type eq "sum" || $type eq "product") && $eqcentring eq "face" && $centring eq "cell") {
          $someregion = "<adjacentfaceicells>"; # default cell loop around a face is just the 2 adjacent faces
        } else {
          $someregion = "<noloop>"; # default min/max region is current location
        }
        print "INFO: default location of $someregion has been applied to $operator operator used in $otype variable $variable{$otype}[$omvar]{name}\n";
        print DEBUG "INFO: default location of $someregion has been applied to $operator operator used in $otype variable $variable{$otype}[$omvar]{name}\n";
      } else {
        print "ERROR: region for $operator in $variable{$otype}[$omvar]{name} is not a valid region name: $someregion\n";
        print DEBUG "ERROR: region for $operator in $variable{$otype}[$omvar]{name} is not a valid region name: |someregion| = |$someregion|\n";
        die;
      }

      if ($inbit[$nbits] ne "") {
        print "ERROR: the contents of $operator in $variable{$otype}[$omvar]{name} is not a valid:\n".
          "  quantity = $quantity\n  default = $default\n  someregion = $someregion\n  remainder = $inbit[$nbits]\n";
        print DEBUG "ERROR: the contents of $operator in $variable{$otype}[$omvar]{name} is not a valid:\n".
          "  quantity = $quantity\n  default = $default\n  someregion = $someregion\n  remainder = $inbit[$nbits]\n";
        die;
      }
      
      print DEBUG "INFO: the contents of $operator in $variable{$otype}[$omvar]{name} was found to be:\n".
        "  quantity = $quantity\n  default = $default\n  someregion = $someregion\n";

      $inbit[$nbits] = $quantity;

      create_someloop($inbit[$nbits],$type,$centring,$someregion,$deriv,$otype,$omvar);

      if ($default) {
        $variable{"someloop"}[$m{"someloop"}]{"default"} = $default; # save default
        print DEBUG "setting default = $default for someloop $m{someloop}\n";
      }

# delta
# the delta function is one in the specified region, zero otherwise
    } elsif ($type eq "delta") {

      if ( $inbit[$nbits] !~ /^\s*(<.+?>)\s*$/ ) {
        die "ERROR: format for region incorrect in $centring $operator\n";
      }
      if ( $centring eq 'cell' ) {
        $inbit[$nbits] = 'region_delta(i,"'.$centring.'","'.$1.'")';
      } elsif ( $centring eq 'face' ) {
        $inbit[$nbits] = 'region_delta(j,"'.$centring.'","'.$1.'")';
      } else { die "ERROR: problem with $centring centring of $operator\n"; }

    } else { die "ERROR: problem with $operator in mequation_interpolation\n"; }

    print DEBUG "current parts in mequation_interpolation: operator = $operator: nbits = $nbits: outbit = $outbit[$nbits]: inbit = $inbit[$nbits]: _[0] = $_[0]\n";
    $nbits++;
  }
  $outbit[$nbits] = $_[0];

#-------------------------------------------------
# loop through bits outside of operators checking on validity and doing replacements

  $n = 0;
  while ( $n <= $nbits ) {
    print DEBUG "in mequation_interpolation checking centring of outbit = $outbit[$n]\n";
# as empties are all zero don't need to interpolate those
    foreach $type ("constant","transient","unknown","derived","equation","output","system") {
      foreach $mvar ( 1 .. $m{$type} ) {
        if ($variable{$type}[$mvar]{"centring"} eq "none") { next; } # skip variables that have no centring
        if ($variable{$type}[$mvar]{"centring"} eq $eqcentring ) { next; } # skip variables that have correct centring
        if ($outbit[$n] =~ /\Q$variable{$type}[$mvar]{"maxima"}/) { # look for variable in outbit
          $pre = $`; $post = $';
          if ($eqcentring eq "none") {
            print "WARNING: mequation is none centred but variable $variable{$type}[$mvar]{name} in outbit has $variable{$type}[$mvar]{centring} centring\n";
            next;
          }
          print DEBUG "found match with $variable{$type}[$mvar]{centring} centred $variable{$type}[$mvar]{maxima} in outbit = $outbit[$n]\n";
          print DEBUG "before splicing: nbits = $nbits: outbit = @outbit: inbit = @inbit\n";
# splice in new in and out bits here
          $nbits++;
          splice(@outbit,$n,1,$pre,$post); # remove present outbit and replace with two new elements, pre and post
          $tmp = $variable{$type}[$mvar]{"maxima"};
          if ($eqcentring eq "cell" && $variable{$type}[$mvar]{"centring"} eq "face") {
            $tmp = 'cellkernel[i,0,ns]*('.$tmp.')';
            $someregion = "<cellkernelregion[l=0]>";
          } elsif ($eqcentring eq "face" && $variable{$type}[$mvar]{"centring"} eq "cell") {
            $tmp = 'facekernel[j,0,ns]*('.$tmp.')';
            $someregion = "<facekernelregion[l=0]>";
          } else { die "ERROR: problem with locating $variable{$type}[$mvar]{name} in outbit\n"; }
# looks for noderivative for this variable
          $deriv = $eqderiv;
          if ($variable{$type}[$mvar]{"options"} && $variable{$type}[$mvar]{"options"} =~ /(^|\,)noderivative(\,|$)/) { $deriv = 0; }
          create_someloop($tmp,"sum",$variable{$type}[$mvar]{"centring"},$someregion,$deriv,$otype,$omvar);
          splice(@inbit,$n,0,$tmp); # add a new element at location n-1
          print DEBUG "after splicing: nbits = $nbits: outbit = @outbit: inbit = @inbit\n";
        }
      }
    }
    $n++;
  }

#-------------------------------------------------
# now assemble equation again
  $_[0] = "";
  foreach $n ( 0 .. $nbits-1 ) {
    $_[0] = $_[0].$outbit[$n].$inbit[$n];
  }
  $_[0] = $_[0].$outbit[$nbits];

#-------------------------------------------------
# run through equations again checking on dependencies of variables, and redefining any if possible to someloops

  if ($otype eq "constant" || $otype eq "transient" || $otype eq "unknown" || $otype eq "derived" ||
    $otype eq "equation" || $otype eq "output") {
    foreach $type ("constant","transient","unknown","derived","equation","output") {
      foreach $mvar ( 1 .. $m{$type} ) {
        if ($_[0] =~ /\Q$variable{$type}[$mvar]{"maxima"}/) { # look for variable in the final equation
          $pre = $`; $post = $';
          if (!($variable{$type}[$mvar]{"equation"})) { next; } # an equation for this variable has to be defined for a dependency issue
# first do checks on variables of it's own type
          if ($otype eq $type) {
            if ($omvar == $mvar) {
              print "ERROR: the equation for $otype variable $variable{$otype}[$omvar]{name} contains a circular reference to itself: check the equations\n";
              print DEBUG "ERROR: the equation for $otype variable $variable{$otype}[$omvar]{name} contains a circular reference to itself: check the equations\n";
              print DEBUG "  offending variable is $type variable $variable{$type}[$mvar]{name}\n";
              die;
            } elsif ($mvar > $omvar) {
              print "ERROR: the equation for $otype variable $variable{$otype}[$omvar]{name} depends upon another $otype variable ".
                "$variable{$type}[$mvar]{name} that is defined after it: either the equation ordering is incorrect or there is a circular dependence\n";
              print DEBUG "ERROR: the equation for $otype variable $variable{$otype}[$omvar]{name} depends upon another $otype variable ".
                "$variable{$type}[$mvar]{name} that is defined after it: either the equation ordering is incorrect or there is a circular dependence\n";
              die;
            }
# replace any wayward variables with someloops if possible
          } elsif ($otype eq "unknown" && ($type eq "derived" || $type eq "transient" || $type eq "equation" || $type eq "output")) {
            print "INFO: replacing $type variable $variable{$type}[$mvar]{name} with a someloop in expression for $otype variable $variable{$otype}[$omvar]{name}\n";
            print DEBUG "INFO: replacing $type variable $variable{$type}[$mvar]{name} with a someloop in expression for $otype variable $variable{$otype}[$omvar]{name}\n";
            print DEBUG "  before change mequation was $_[0]\n";
            create_someloop($variable{$type}[$mvar]{"mequation"},"sum",$variable{$type}[$mvar]{"centring"},"<noloop>",$variable{$type}[$mvar]{"deriv"},$otype,$omvar);
            $_[0] = $pre.$variable{"someloop"}[$m{"someloop"}]{"maxima"}.$post;
            print DEBUG "  after change mequation was $_[0]\n";
# find any dependencies that cannot be resolved
          } elsif ($otype eq "constant") {
            print "ERROR: $otype variable $variable{$otype}[$omvar]{name} depends on $type variable $variable{$type}[$mvar]{name}: ".
              "constant variables must only depend on other constant and system variables\n";
            die;
          }
        }
      }
    }
  }

  print DEBUG "in mequation_interpolation: after _[0] = ",$_[0],"\n".
              "------------------------------------------------------------------------\n";
}

#-------------------------------------------------------------------------------
# here we create a new someloop
# enter with
# - $_[0] as the mequation, to be replaced by the someloop maxima name
# - $_[1] as the loop type, (sum|max|min|product)
# - $_[2] as the centring of the region over which the loop is performed
# - $_[3] as the region name over which the loop is performed
# - $_[4] is 0 or 1 dependending on whether a derivative is required
# - $_[5] is otype, which is the type from which the expression originated
# - $_[6] is omvar, which is the mvar from which the expression originated

sub create_someloop {

  use strict;

  $m{"someloop"} ++;
  $variable{"someloop"}[$m{"someloop"}]{"mequation"} = $_[0];
  $variable{"someloop"}[$m{"someloop"}]{"type"} = $_[1];
  $variable{"someloop"}[$m{"someloop"}]{"centring"} = $_[2];
  $variable{"someloop"}[$m{"someloop"}]{"region"} = $_[3];
# $variable{"someloop"}[$m{"someloop"}]{"name"} = "SYSTEM generated $_[1] someloop over $_[2] region $_[3] which is a component of $_[5] $variable{$_[5]}[$_[6]]{name}";
# $variable{"someloop"}[$m{"someloop"}]{"equation"} = "SYSTEM generated $_[1] someloop over $_[2] region $_[3] which is a component of $_[5] $variable{$_[5]}[$_[6]]{name}";
  $variable{"someloop"}[$m{"someloop"}]{"name"} = "$_[1] someloop over $_[2] region $_[3] which is part of $_[5] variable $variable{$_[5]}[$_[6]]{name}";
  $variable{"someloop"}[$m{"someloop"}]{"equation"} = "MEQUATION: $_[0]";
  $variable{"someloop"}[$m{"someloop"}]{"deriv"} = $_[4];
  $variable{"someloop"}[$m{"someloop"}]{"maxima"} = "someloop[$m{someloop}]";
  $variable{"someloop"}[$m{"someloop"}]{"mfortran"} = "someloop($m{someloop})";
  $variable{"someloop"}[$m{"someloop"}]{"fortran"} = "someloop($m{someloop})%v";
  $variable{"someloop"}[$m{"someloop"}]{"fortranns"} = "someloop($m{someloop})%v";
  $variable{"someloop"}[$m{"someloop"}]{"fortran_number"} = $m{"someloop"};
  $variable{"someloop"}[$m{"someloop"}]{"otype"} = $_[5];
  $variable{"someloop"}[$m{"someloop"}]{"omvar"} = $_[6];

# $variable{"someloop"}[$m{"someloop"}]{"default"} = $default;

  $_[0] = $variable{"someloop"}[$m{"someloop"}]{"maxima"}; 

# renew someloop allocation string
  $sub_string{"allocate_someloop"}="allocate(someloop($m{someloop}))\n";

  print DEBUG "+++++++++++++++++++++\n".
    "in create_someloop with: number = $m{someloop}: originated from $variable{someloop}[$m{someloop}]{otype} variable ".
    "$variable{$variable{someloop}[$m{someloop}]{otype}}[$variable{someloop}[$m{someloop}]{omvar}]{name}: maxima = $variable{someloop}[$m{someloop}]{maxima}: ".
    "mequation = $variable{someloop}[$m{someloop}]{mequation}: type = $variable{someloop}[$m{someloop}]{type}: ".
    "centring = $variable{someloop}[$m{someloop}]{centring}: region = $variable{someloop}[$m{someloop}]{region}\n".
    "---------------------\n";
}

#-------------------------------------------------------------------------------

sub split_off_leading_mequation {

  use strict;
  use Text::Balanced qw ( extract_bracketed );
  my ($equation, $braces_match, $in);

  $in = $_[0]; # save for debug output

  $equation = "";
# while ( $_[0] =~ /(<.+?>)|(\,)|$/ ) {
# while ( $_[0] =~ /(\[.+?\])|(\,)|$/ ) { # made for maxima variables
  while ( $_[0] =~ /(\[.+?\])|(\()|(\,)|$/ ) { # made for maxima variables
    $_[0] = $';
    if ($1) {
      $equation = $equation.$`.$1;
    } elsif ($2) { # remove until corresponding brace
      $_[0] = '('.$_[0];
      ($braces_match,$_[0]) = extract_bracketed( $_[0], '()' ); # http://search.cpan.org/~adamk/Text-Balanced-2.02/lib/Text/Balanced.pm
      if (!($braces_match)) { die "ERROR: matching braces not found in $in\n"."  Possibly an error in this equation?\n"; }
      $equation = $equation.$`.$braces_match;
    } else {
      $equation = $equation.$`;
      last;
    }
  }

# remove any leading or trailing spaces
  $equation =~ s/^\s*//; $equation =~ s/\s*$//;
# if ($equation =~ /^\s*(.+?)\s*$/) {
#   $equation = $1;
# } else { 
#   print DEBUG "empty string found in split_off_leading_equation: |equation| = |$equation|\n";
#   $equation = "";
# }

  return $equation;
}

#-------------------------------------------------------------------------------
# runs a comma separated list of mequations through maxima for simplification

sub run_maxima_simplify {

  use strict;
  my ($systemcall, $line, $out, $in, $otype, $omvar, $bit);

  if (-e $stopfile) {die "HALT: found $stopfile in run_maxima_simplify\n";}

  print DEBUG "+++++++++++++++++++++\n".
    "in run_maxima_simpilfy: before _[0] = ",$_[0],"\n";

  $in = $_[0];
  $otype = $_[1];
  $omvar = $_[2];
  $out = 0;

  while ($in) {

    $bit = split_off_leading_mequation($in);

    $tmp_file_number ++;
    open(MAXIMA, ">$tmp_dir/tmp$tmp_file_number.maxima") or die "ERROR: problem opening temporary maxima file $tmp_dir/tmp$tmp_file_number.maxima: something funny is going on: check permissions??\n";
    print MAXIMA "foo:$bit;";
    print MAXIMA "with_stdout (\"$tmp_dir/tmp$tmp_file_number.simp\", grind(foo));";
    close(MAXIMA);
#   $systemcall="maxima -b $tmp_dir/tmp$tmp_file_number.maxima >$tmp_dir/tmp$tmp_file_number.txt";
    $systemcall="$maxima_bin -b $tmp_dir/tmp$tmp_file_number.maxima >$tmp_dir/tmp$tmp_file_number.txt";
    (!(system("$systemcall"))) or do {print "could not $systemcall\n"; die;};
    open(SIMP, "<$tmp_dir/tmp$tmp_file_number.simp") or do {
      print "ERROR: problem in run_maxima_simplify processing the following mequation (maxima equation):\n'$bit'\n".
        "  This equation originated from $otype variable $variable{$otype}[$omvar]{name} and this\n".
        "  error indicates that maxima had problems processing this equation.  All variables that\n".
        "  that appear in this equation should be referred to by their maxima names:  Hence if any\n".
        "  variables appear in the above as their original names (e.g. <my variable>) then\n".
        "  check their equations as there is some problem with their definition (maybe a typo?).\n".
        "  Also check that any mathematical functions and syntax used in the above are compatible with maxima.\n";
      print "  Maxima input lines are contained in the file $tmp_dir/tmp$tmp_file_number.maxima\n";
      print "  Maxima processing output contained in the file $tmp_dir/tmp$tmp_file_number.txt\n";
      print "  Maxima output lines are contained in the file $tmp_dir/tmp$tmp_file_number.simp\n";
      print DEBUG "ERROR: problem in run_maxima_simplify processing the following mequation:\n'$bit'\n".
      die;
    };
    $bit = "";
    while ($line = <SIMP>) {
      chompm($line); 
      $line =~ s/\$$//;
      $line =~ s/^\s*//;
      $bit=$bit.$line;
    }
    close(SIMP);

    print DEBUG "midway through run_maxima_simplify: in = $in: out = $out: bit = $bit\n";
    if ($out) {$out = $out.",$bit";} else {$out = "$bit";}

  }

  $_[0] = $out;

  print DEBUG "in run_maxima_simplify: after _[0] = ",$_[0],"\n".
    "---------------------\n";
}

#-------------------------------------------------------------------------------
# takes a string which contains user defined <> delimited names and replaces with maxima names

sub name_to_maxima {

  use strict;
  use Text::Balanced qw ( extract_bracketed );
  my ($type, $mvar);
  my ($pre, $post, $operator, $midl, $midr, $midrnovar, $n1, $n2);

  print DEBUG "in name_to_maxima: before _[0] = ",$_[0],"\n";

# look for any dot and ddot operators and replace with scalar mathematics

  while ($_[0] =~ /ddot\(|dot\(/) {
    $pre = $`;
    $post = $';
    $operator = $&; $operator =~ s/\($//;
    print "INFO: found operator $operator in $_[0]\n";
    print DEBUG "found operator $operator in $_[0]\n";
    print DEBUG "pre = $pre: post = $post\n";

    $midl = "";
    while ( $post =~ /(<.+?>)|(\,)/ ) {
      $post = $';
      if ($1) {
        $midl = $midl.$`.$1;
      } else {
        $midl = $midl.$`;
        last;
      }
    }
    print DEBUG "left side finalised: midl = $midl: post = $post\n";
        
    $midr = "";
    $midrnovar = "";
    while ( $post =~ /(<.+?>)|(\))/ ) {
      $post = $';
      if ($1) {
        $midr = $midr.$`.$1;
        $midrnovar = $midrnovar.$`;
      } else {
        if (extract_bracketed( "($midrnovar)", '()' )) {last;} # if contents of right without the variables has matching brackets, then we're in business
        $midr = $midr.$`.")";
        $midrnovar = $midrnovar.$`.")";
      }
    }
    print DEBUG "right side finalised: midr = $midr: midrnovar = $midrnovar: post = $post\n";

    my @posl = ();
    while ($midl =~ /<.+?\[l=.{0,2}(:).{0,2}\]>/) {
#     print "midl = $midl: posl = @posl\n";
      push (@posl, $+[1]); # $+[1] is the starting position of the first match
      substr($midl, $+[1]-1, 1, " "); #replace : with blank so that it is not matched
    }
    @posl = sort{ $a <=> $b } @posl; # make sure index positions are in order
    my @posr = ();
    while ($midr =~ /<.+?\[l=.{0,2}(:).{0,2}\]>/) {
      push (@posr, $+[1]); # $+[1] is the starting position of the first match
      substr($midr, $+[1]-1, 1, " "); #replace : with blank so that it is not matched
    }
    @posr = sort{ $a <=> $b } @posr; # make sure index positions are in order
    print DEBUG "after: midl = $midl: posl = @posl\n";
    print DEBUG "after: midr = $midr: posr = @posr\n";

    my $mid = "(";
    if ($operator eq "dot") {
      if (@posl != 1 || @posr != 1) { print "ERROR: wrong number of colons in dot product\n"; die; } # scalar(@posl) is number of elements of posl array
      foreach $n1 ( 1 .. 3 ) {
        substr($midl, $posl[0]-1, 1, $n1); #replace blank with correct index
        substr($midr, $posr[0]-1, 1, $n1); #replace blank with correct index
        $mid = $mid." + ".$midl."*".$midr;
      }
    } else { # double dot product
      if (@posl != 2 || @posr != 2) { print "ERROR: wrong number of colons in ddot product\n"; die; } # scalar(@posl) is number of elements of posl array
      foreach $n1 ( 1 .. 3 ) {
        substr($midl, $posl[0]-1, 1, $n1); #replace blank with correct index
        substr($midr, $posr[1]-1, 1, $n1); #replace blank with correct index
        foreach $n2 ( 1 .. 3 ) {
          substr($midl, $posl[1]-1, 1, $n2); #replace blank with correct index
          substr($midr, $posr[0]-1, 1, $n2); #replace blank with correct index
          $mid = $mid." + ".$midl."*".$midr;
        }
      }
    }
    $mid = $mid." )";
        
    print DEBUG "mid = $mid\n";
      
    $_[0] = $pre.$mid.$post;

    print DEBUG "after $operator operator expansion: _[0] = $_[0]\n";

  }
      
# replace standard variables
  foreach $type ("constant","transient","unknown","derived","output","empty","system") {
    foreach $mvar ( 1 .. $m{$type} ) {
      replace_substrings($_[0],$variable{$type}[$mvar]{"name"},$variable{$type}[$mvar]{"maxima"});
    }
  }

  print DEBUG "in name_to_maxima: after _[0] = ",$_[0],"\n";

}

#-------------------------------------------------------------------------------
# calling sequence is: string, substring, replacement

sub replace_substrings {
  use strict;
  $_[0] =~ s/\Q$_[1]/$_[2]/g; # \Q escapes any funny characters in $_[1]
}

#-------------------------------------------------------------------------------
# assemble compound arrays for each of the variables

sub create_compounds {

  use strict;
  use Text::Balanced qw ( extract_bracketed );
  my ($type, $mvar, $name, $indices, $type2, $mvar2, $index, $mcheck, $rank, $component_list,
    $nrank, $mmvar, $tmp, $l1, $l2, $nvar, $nametmp);

# loop through all vars
  foreach $type ("constant","transient","unknown","derived","equation","output") {
    foreach $mvar ( 1 .. $m{$type} ) {

      $name = $variable{$type}[$mvar]{"name"};
      ($name) = $name =~ /^<(.*)>$/;
      $index = 1;
      $rank = "scalar"; # default is a scalar
      $nrank = 1;
      if ($name =~ /^(.+?)\[(.+?)\]$/) {
        ($nametmp,$indices) = ($1, $2);
        if ($indices =~ /l=([123]),([123])/) {
          $index = ($1-1)*3 + $2; # l = (j-1)*3+i where j = row number, i = col number
          $rank = "tensor";
          $nrank = 9;
          $name = $nametmp;
        }
        elsif ($indices =~ /l=([123])/) {
          $index = $1;
          $rank = "vector";
          $nrank = 3;
          $name = $nametmp;
        }
      }
      $name = "<".$name.">";

# find `base' name of variable
      $mvar2 = 0;
      foreach $mcheck ( 1 .. $m{"compound"} ) {
        if ($variable{"compound"}[$mcheck]{"name"} eq $name) {
          print "INFO: found existing $rank compound with name = $name for component $variable{$type}[$mvar]{name}\n";
          print DEBUG "INFO: found existing $rank compound with name = $name for component $variable{$type}[$mvar]{name}\n";
          if (!($rank eq $variable{"compound"}[$mcheck]{"rank"})) {
            print "ERROR: rank mismatch found between component ($rank) and compound ($variable{compound}[$mcheck]{rank})\n";
            die;
          } elsif (!($variable{$type}[$mvar]{"units"} eq $variable{"compound"}[$mcheck]{"units"})) {
            print "ERROR: units mismatch found between component ($variable{$type}[$mvar]{units}) and compound ($variable{compound}[$mcheck]{units})\n";
            die;
          } elsif (!($variable{$type}[$mvar]{"centring"} eq $variable{"compound"}[$mcheck]{"centring"})) {
            print "ERROR: centring mismatch found between component ($variable{$type}[$mvar]{centring}) and compound ($variable{compound}[$mcheck]{centring})\n";
            die;
          } elsif (!($variable{$type}[$mvar]{"region"} eq $variable{"compound"}[$mcheck]{"region"})) {
            print "ERROR: region mismatch found between component ($variable{$type}[$mvar]{region}) and compound ($variable{compound}[$mcheck]{region})\n";
            die;
          } elsif (!($variable{$type}[$mvar]{"multiplier"} eq $variable{"compound"}[$mcheck]{"multiplier"})) {
            print "ERROR: multiplier mismatch found between component ($variable{$type}[$mvar]{multiplier}) and compound ($variable{compound}[$mcheck]{multiplier})\n";
            die;
          }
          $mvar2 = $mcheck;
          last;
        }
      }

# create new compound variable if it doesn't exist
      if (!($mvar2)) {
        $m{"compound"}++;
        $mvar2=$m{"compound"};
        $variable{"compound"}[$mvar2]{"name"} = $name;
        $variable{"compound"}[$mvar2]{"rank"} = $rank;
        $variable{"compound"}[$mvar2]{"nrank"} = $nrank;
        $variable{"compound"}[$mvar2]{"units"} = $variable{$type}[$mvar]{"units"};
        $variable{"compound"}[$mvar2]{"multiplier"} = $variable{$type}[$mvar]{"multiplier"};
        $variable{"compound"}[$mvar2]{"centring"} = $variable{$type}[$mvar]{"centring"};
        $variable{"compound"}[$mvar2]{"type"} = $type;
        $variable{"compound"}[$mvar2]{"fortran_number"} = $mvar2;
        $variable{"compound"}[$mvar2]{"region"} = $variable{$type}[$mvar]{"region"};
# set default compoundoutput - if var is an unknown or a output or a cell centred derived then on by default
        if (($type eq "derived" && $variable{$type}[$mvar]{"centring"} eq "cell") || $type eq "unknown" || $type eq "output") {
          $variable{"compound"}[$mvar2]{"output"} = 1; 
        } else {
          $variable{"compound"}[$mvar2]{"output"} = 0; 
        }
        print "INFO: creating new $rank compound with name = $name and centring = $variable{$type}[$mvar]{centring} for component $variable{$type}[$mvar]{name}\n";
        print DEBUG "INFO: creating new $rank compound with name = $name and centring = $variable{$type}[$mvar]{centring} for component $variable{$type}[$mvar]{name}\n";
        foreach $mmvar ( 1 .. $nrank ) {
          $variable{"compound"}[$mvar2]{"component"}[$mmvar] = 0;
          $variable{"compound"}[$mvar2]{"fortran_component"}[$mmvar] = 0;
        }
      }

# finally allocate component after checking that it is not already allocated
      if ($variable{"compound"}[$mvar2]{"component"}[$index]) {
        print "ERROR: component $index of $rank $name has already been allocated\n";
        die;
      }
      $variable{"compound"}[$mvar2]{"component"}[$index] = $mvar;
      $variable{"compound"}[$mvar2]{"fortran_component"}[$index] = $variable{$type}[$mvar]{"fortran_number"};

# check if compoundoutput it has been turned on or off via this component
# conflicting options will favour the last read component
      if ($variable{$type}[$mvar]{"options"} && $variable{$type}[$mvar]{"options"} =~ /(^|\,)nocompoundoutput(\,|$)/) {
        $variable{"compound"}[$mvar2]{"output"} = 0;
      }
      if ($variable{$type}[$mvar]{"options"} && $variable{$type}[$mvar]{"options"} =~ /(^|\,)compoundoutput(\,|$)/) {
        $variable{"compound"}[$mvar2]{"output"} = 1;
      }

    } # all compound variables for this type have been constructed

  }

#-----------------
# run through each compound variable now disseminating the info to the components and
#  creating new empty scalar components that aren't otherwise defined

  foreach $mvar2 ( 1 .. $m{"compound"} ) {
    $type = $variable{"compound"}[$mvar2]{"type"};
    $component_list = "[";
    foreach $nvar ( 1 .. $variable{"compound"}[$mvar2]{"nrank"} ) {
      $component_list = $component_list." $variable{compound}[$mvar2]{fortran_component}[$nvar] ,";
    }
    $component_list =~ s/,$/]/;
    print "INFO: compound $mvar2 $variable{compound}[$mvar2]{name} has components: $component_list\n";
    $variable{"compound"}[$mvar2]{"component_list"} = $component_list; # a string suitable for fortran

    $l1 = 1;
    $l2 = 1;
    foreach $nvar ( 1 .. $variable{"compound"}[$mvar2]{"nrank"} ) {
      $mvar = $variable{"compound"}[$mvar2]{"component"}[$nvar];
      print DEBUG "l1 = $l1: l2 = $l2: mvar = $mvar\n";
      if (!($mvar)) {
# create new zero scalar
        $m{"empty"}++ ;
        $tmp = $variable{"compound"}[$mvar2]{"name"};
        $tmp =~ s/>$//;
        if ($variable{"compound"}[$mvar2]{"nrank"} eq 3) {
          $variable{"empty"}[$m{"empty"}]{"name"} = $tmp."[l=".$l1."]>";
        } elsif ($variable{"compound"}[$mvar2]{"nrank"} eq 9) {
          $variable{"empty"}[$m{"empty"}]{"name"} = $tmp."[l=".$l2.",".$l1."]>";
        } else {
          $variable{"empty"}[$m{"empty"}]{"name"} = $tmp.">";
        }
        print DEBUG "INFO: creating new empty $type scalar component: $variable{empty}[$m{empty}]{name}\n";
        print "INFO: creating new empty $type scalar component: $variable{empty}[$m{empty}]{name}\n";
        $variable{"empty"}[$m{"empty"}]{"maxima"} = "0.";
        $variable{"empty"}[$m{"empty"}]{"fortran"} = "0.d0";
        $variable{"empty"}[$m{"empty"}]{"fortranns"} = "0.d0";
        $variable{"empty"}[$m{"empty"}]{"rank"} = $variable{"compound"}[$mvar2]{"rank"};
        $variable{"empty"}[$m{"empty"}]{"nrank"} = $variable{"compound"}[$mvar2]{"nrank"};
        $variable{"empty"}[$m{"empty"}]{"units"} = $variable{"compound"}[$mvar2]{"units"};
        $variable{"empty"}[$m{"empty"}]{"multiplier"} = $variable{"compound"}[$mvar2]{"multiplier"};
        $variable{"empty"}[$m{"empty"}]{"centring"} = $variable{"compound"}[$mvar2]{"centring"};
        $variable{"empty"}[$m{"empty"}]{"compound_name"} = $variable{"compound"}[$mvar2]{"name"};
        $variable{"empty"}[$m{"empty"}]{"component_list"} = $component_list;
        $variable{"empty"}[$m{"empty"}]{"deriv"} = 0;
        $variable{"empty"}[$m{"empty"}]{"fortran_number"} = $m{"empty"};
        $variable{"empty"}[$m{"empty"}]{"type"} = $type;
      } else {
        $variable{$type}[$mvar]{"rank"} = $variable{"compound"}[$mvar2]{"rank"};
        $variable{$type}[$mvar]{"nrank"} = $variable{"compound"}[$mvar2]{"nrank"};
        $variable{$type}[$mvar]{"compound_name"} = $variable{"compound"}[$mvar2]{"name"};
        $variable{$type}[$mvar]{"component_list"} = $component_list;
      }
      $l1++;
      if ($l1 > 3) { $l1 = $l1 - 3; $l2++; }
    }
  }
      
# @{$variable{"compound"}[$mvar2]{component}}

}

#-------------------------------------------------------------------------------
# assemble metadata allocation statements

sub create_allocations {

  use strict;
  my ($type, $mvar, $tmp, $comparator, $index1, $index2);

  $sub_string{allocate_meta_arrays}="";

# allocate var and compound
  $sub_string{allocate_meta_arrays} = $sub_string{allocate_meta_arrays}.
    "\n! allocate var and compound arrays\n".
    "allocate(var($m{general}))\n".
    "allocate(compound($m{compound}))\n";

  foreach $type ("constant","transient","unknown","derived","equation","output","compound") {
    print "INFO: creating meta statements for type = ",$type,"\n";

    if (!($m{$type})) { next; }

    foreach $mvar ( 1 .. $m{$type} ) {

# write out initialisation statements for meta data
      if ( $type ne "compound" ) {
        $sub_string{allocate_meta_arrays}=$sub_string{allocate_meta_arrays}.
          "\n!$type: $variable{$type}[$mvar]{name} -> $variable{$type}[$mvar]{maxima} ".
          "-> $variable{$type}[$mvar]{mfortran} -> $variable{$type}[$mvar]{fortran}\n".
          "$basename{$type}($variable{$type}[$mvar]{fortran_number})%name=\'$variable{$type}[$mvar]{name}\'\n".
          "$basename{$type}($variable{$type}[$mvar]{fortran_number})%multiplier=$variable{$type}[$mvar]{multiplier}\n".
          "$basename{$type}($variable{$type}[$mvar]{fortran_number})%units=\'$variable{$type}[$mvar]{units}\'\n".
          "$basename{$type}($variable{$type}[$mvar]{fortran_number})%centring=\'$variable{$type}[$mvar]{centring}\'\n".
          "$basename{$type}($variable{$type}[$mvar]{fortran_number})%type=\'$type\'\n".
          "$basename{$type}($variable{$type}[$mvar]{fortran_number})%rank=\'$variable{$type}[$mvar]{rank}\'\n".
          "$basename{$type}($variable{$type}[$mvar]{fortran_number})%region=\'$variable{$type}[$mvar]{region}\'\n".
          "$basename{$type}($variable{$type}[$mvar]{fortran_number})%compound_name=\'$variable{$type}[$mvar]{compound_name}\'\n".
          "allocate($basename{$type}($variable{$type}[$mvar]{fortran_number})%component($variable{$type}[$mvar]{nrank}))\n".
          "$basename{$type}($variable{$type}[$mvar]{fortran_number})%component=$variable{$type}[$mvar]{component_list}\n".
          "$basename{$type}($variable{$type}[$mvar]{fortran_number})%output=".fortran_logical_string($variable{$type}[$mvar]{"output"})."\n";
        if ($variable{$type}[$mvar]{"magnitude"}) {
          $sub_string{allocate_meta_arrays}=$sub_string{allocate_meta_arrays}.
            "$basename{$type}($variable{$type}[$mvar]{fortran_number})%magnitude=$variable{$type}[$mvar]{magnitude}\n";
        }
      } else {
        $sub_string{allocate_meta_arrays}=$sub_string{allocate_meta_arrays}.
          "\n!$type: $variable{$type}[$mvar]{name} of $variable{$type}[$mvar]{rank} rank\n".
          "$basename{$type}($variable{$type}[$mvar]{fortran_number})%name=\'$variable{$type}[$mvar]{name}\'\n".
          "$basename{$type}($variable{$type}[$mvar]{fortran_number})%units=\'$variable{$type}[$mvar]{units}\'\n".
          "$basename{$type}($variable{$type}[$mvar]{fortran_number})%centring=\'$variable{$type}[$mvar]{centring}\'\n".
          "$basename{$type}($variable{$type}[$mvar]{fortran_number})%type=\'$variable{$type}[$mvar]{type}\'\n".
          "$basename{$type}($variable{$type}[$mvar]{fortran_number})%rank=\'$variable{$type}[$mvar]{rank}\'\n".
          "$basename{$type}($variable{$type}[$mvar]{fortran_number})%region=\'$variable{$type}[$mvar]{region}\'\n".
          "allocate($basename{$type}($variable{$type}[$mvar]{fortran_number})%component($variable{$type}[$mvar]{nrank}))\n".
          "$basename{$type}($variable{$type}[$mvar]{fortran_number})%component=$variable{$type}[$mvar]{component_list}\n".
          "$basename{$type}($variable{$type}[$mvar]{fortran_number})%output=".fortran_logical_string($variable{$type}[$mvar]{"output"})."\n";
      }
    }
  }

# intitialise constants_file names
  $sub_string{allocate_meta_arrays}=$sub_string{allocate_meta_arrays}.
    "\nconstants_file = \"$input_files[0]\" ! filename that was used by setup_equations\n";

# run through variables creating any check functions
  $sub_string{"variable_checker"}="";
  foreach $type ("unknown","derived") {
    foreach $mvar ( 1 .. $m{$type} ) {
      $comparator = 0;
      if ($variable{$type}[$mvar]{"options"} && $variable{$type}[$mvar]{"options"} =~ /(^|\,)positive(\,|$)/) { $comparator = '<'; }
      if ($variable{$type}[$mvar]{"options"} && $variable{$type}[$mvar]{"options"} =~ /(^|\,)negative(\,|$)/) { $comparator = '>'; }
      if ($comparator) {
        print "INFO: creating check statements for $type $variable{$type}[$mvar]{name} with options $variable{$type}[$mvar]{options}\n";
        $sub_string{"variable_checker"}=$sub_string{"variable_checker"}.
          "\n!$type: $variable{$type}[$mvar]{name} has options $variable{$type}[$mvar]{options}\n".
          "error = 2\n".
          "do ns = 1, ubound(var($variable{$type}[$mvar]{fortran_number})%funk,1)\n".
          "if ($variable{$type}[$mvar]{fortranns} $comparator 0.d0) return\n".
          "end do\n".
          "error = 0\n";
      }
    }
  }

}

#-------------------------------------------------------------------------------
# this runs a maxima formated equation through the f90 package of maxima

sub run_maxima_fortran {

  use strict;
  my ($line, $systemcall, $bit, $otype, $omvar);

  if (-e $stopfile) {die "HALT: found $stopfile in run_maxima_fortran";}

  print DEBUG "+++++++++++++++++++++\n".
    "in run_maxima_fortran: before _[0] = ",$_[0],"\n";
  $bit = $_[0];
  $otype = $_[1];
  $omvar = $_[2];

  $tmp_file_number ++;
  open(MAXIMA, ">$tmp_dir/tmp$tmp_file_number.maxima") or die "ERROR: problem opening temporary maxima file $tmp_dir/tmp$tmp_file_number.maxima: something funny is going on: check permissions??\n";
  print MAXIMA "load(f90);";
  print MAXIMA "gradef(heaviside(x),0);"; # define derivative of heaviside function to be zero
  print MAXIMA "foo:$bit;";
  print MAXIMA "with_stdout (\"$tmp_dir/tmp$tmp_file_number.fortran\", f90(foo));";
  close(MAXIMA);
# $systemcall="maxima -b $tmp_dir/tmp$tmp_file_number.maxima >$tmp_dir/tmp$tmp_file_number.txt";
  $systemcall="$maxima_bin -b $tmp_dir/tmp$tmp_file_number.maxima >$tmp_dir/tmp$tmp_file_number.txt";
  (!(system("$systemcall"))) or do {print "could not $systemcall\n"; die;};
# open(FORTRAN, "<$tmp_dir/tmp$tmp_file_number.fortran") or die "ERROR: problem in run_maxima_fortran with: $bit\n";
  open(FORTRAN, "<$tmp_dir/tmp$tmp_file_number.fortran") or do {
    print "ERROR: problem in run_maxima_fortran processing the following mequation (maxima equation):\n'$bit'\n".
      "  This equation originated from $otype variable $variable{$otype}[$omvar]{name} and this\n".
      "  error indicates that maxima had problems processing this equation.  All variables that\n".
      "  that appear in this equation should be referred to by their maxima names:  Hence if any\n".
      "  variables appear in the above as their original names (e.g. <my variable>) then\n".
      "  check their equations as there is some problem with their definition (maybe a typo?).\n".
      "  Also check that any mathematical functions and syntax used in the above are compatible with\n".
      "  maxima and the f90 package.  Finally, if a diff (derivative) is requested, check that the\n".
      "  derivative exists and that maxima is actually able to calculate it.\n";
    print "  Maxima input lines are contained in the file $tmp_dir/tmp$tmp_file_number.maxima\n";
    print "  Maxima processing output contained in the file $tmp_dir/tmp$tmp_file_number.txt\n";
    print "  Maxima output lines are contained in the file $tmp_dir/tmp$tmp_file_number.fortran\n";
    print DEBUG "ERROR: problem in run_maxima_fortran processing the following mequation:\n'$bit'\n".
    die;
  };
  $_[0] = "";
  while ($line = <FORTRAN>) {
    chompm($line); 
    $line =~ s/&$//;
    $_[0]=$_[0].$line;
  }
  close(FORTRAN);

  print DEBUG "in run_maxima_fortran: after _[0] = ",$_[0],"\n".
    "---------------------\n";
}

#-------------------------------------------------------------------------------

sub fortran_logical_string {

  use strict;
  my $string;

  if ($_[0]) {
    $string = ".true.";
  } else {
    $string = ".false.";
  }
  return $string;

}

#-------------------------------------------------------------------------------
# this takes a single maxima line and converts it into a single fortran line with
#  references suitable for the final code

sub maxima_to_fortran {

  use strict;
  use Text::Balanced qw ( extract_bracketed );
  my ($type, $mvar, $ii, $ji, $pre, $post, $search, $replace, $n, $tmp1, $tmp2);
  my ($ns, $nsnext, $nlast, $nnext); 
  my @iilist = ();
  my $otype = $_[1];
  my $omvar = $_[2];

  print DEBUG "in maxima_to_fortran: before _[0] = ",$_[0],"\n";

  run_maxima_fortran($_[0],$otype,$omvar);

# create two lists of words which should be swapped in the string
  my @searches = ();
  my @replaces = ();

  foreach $type ("constant","transient","unknown","derived","equation","output","system","someloop") {

    foreach $mvar ( 1 .. $m{$type}) {

      push(@searches,$variable{$type}[$mvar]{"mfortran"});
      push(@replaces,$variable{$type}[$mvar]{"fortran"});

    }
  }

# print DEBUG "doing search replacements of the following pairs\n";
# foreach $ns ( 0 .. $#searches ) {
#   print DEBUG "ns = $ns: searches[ns] = $searches[$ns]: replaces[ns] = $replaces[$ns]\n";
# }
  
# replace strings by looping along each character from the start, to avoid double matches
  print DEBUG "about to do replacements on string _[0] = $_[0]\n";
  $nlast = 0;
  $nnext = 0;
  while ($nnext < length($_[0])) {

    $nnext = length($_[0]); # find position of next match
    foreach $ns ( 0 .. $#searches ) { # $#array is the index of the last element of the array
      $n = index($_[0],$searches[$ns],$nlast); # find occurance of search string which is closest to previous
      if ($n >= 0 && $n < $nnext) { $nnext = $n, $nsnext = $ns; }
    }

    if ($nnext < length($_[0])) {
      print DEBUG "found match with searches = $searches[$nsnext]: replaces = $replaces[$nsnext]: at nnext = $nnext\n".
        "on string _[0] = $_[0]\n";
      substr($_[0],$nnext,length($searches[$nsnext]),$replaces[$nsnext]);
      $nlast = $nnext + length($replaces[$nsnext]);
    }

  }

  print DEBUG "in maxima_to_fortran: after _[0] = ",$_[0],"\n";

}

#-------------------------------------------------------------------------------
# here we setup equations

sub create_fortran_equations {

  use strict;
  use Text::Balanced qw ( extract_bracketed );
  my ($type, $mvar, $mequation, $tmp, $m2var, $deriv, $inequality, $first, $fequation);

  foreach $type ("constant","transient","unknown","derived","equation","output","someloop") {
    
    $sub_string{$type}="";
    $first = 1;

    foreach $mvar ( 1 .. $m{$type} ) {

      if ($type eq "constant" && !($variable{$type}[$mvar]{"mequation"})) {next;}

      $deriv = $variable{$type}[$mvar]{"deriv"};
      $mequation = $variable{$type}[$mvar]{"mequation"};

      print "INFO: generating fortran equations for $type [$mvar]: $variable{$type}[$mvar]{name}: deriv = $deriv\n";

# start writing fortran

      if (!($first)) { $sub_string{$type}=$sub_string{$type}."else "; }
      $first = 0;
      $sub_string{$type}=$sub_string{$type}."if (m == $variable{$type}[$mvar]{fortran_number}) then\n";

      $sub_string{$type}=$sub_string{$type}.
        "! $type $variable{$type}[$mvar]{name}\n".
        "! $variable{$type}[$mvar]{equation}\n".
        "! $variable{$type}[$mvar]{mequation}\n\n";

# pre loop statements - deal with variables required prior to loop commencing
# set default value for max/min
      if ($type eq "someloop" && $variable{$type}[$mvar]{"default"} ) {
        $tmp = $variable{$type}[$mvar]{"default"}; 
        maxima_to_fortran($tmp,$variable{$type}[$mvar]{"otype"},$variable{$type}[$mvar]{"omvar"});
        $sub_string{$type}=$sub_string{$type}.
          "! update default expression outside main loop\n".
          "! ".$variable{$type}[$mvar]{"default"}."\n". 
          update_someloop_fortran($variable{$type}[$mvar]{"default"})."\n".
          $variable{$type}[$mvar]{"fortran"}." = ".$tmp."\n";
        if ($deriv) {
          $tmp = create_deriv_fortran($variable{$type}[$mvar]{"fortran"},$variable{$type}[$mvar]{"default"},$variable{$type}[$mvar]{"otype"},$variable{$type}[$mvar]{"omvar"});
          $sub_string{$type}=$sub_string{$type}."$tmp\n";
        }
      }
# set flux value for upwind/downwind cells regions
      if ($type eq "someloop" && $variable{$type}[$mvar]{"flux"} ) {
        $tmp = $variable{$type}[$mvar]{"flux"}; 
        maxima_to_fortran($tmp,$variable{$type}[$mvar]{"otype"},$variable{$type}[$mvar]{"omvar"});
        $sub_string{$type}=$sub_string{$type}.
          "! update flux expression outside main loop\n".
          "! ".$variable{$type}[$mvar]{"flux"}."\n". 
          update_someloop_fortran($variable{$type}[$mvar]{"flux"})."\n".
          "flux_direction = nint(sign(1.d0,".$tmp."))\n";
      }
      if ($type eq "someloop" && $variable{$type}[$mvar]{"type"} eq "product" ) {
# set initial value for product loop
        $sub_string{$type}=$sub_string{$type}.
          "! setting product initially to 1.\n".
          "$variable{$type}[$mvar]{fortranns} = 1.d0\n\n";
      }

# start loop
# all var variables
      if ( $type ne "someloop" ) {
        if ($variable{$type}[$mvar]{"centring"} eq "face") {
          $sub_string{$type}=$sub_string{$type}.
            "do ns = 1, ubound(region(var(m)%region_number)%ij,1)\n".
            "j = region(var(m)%region_number)%ij(ns)\n";
        } elsif ($variable{$type}[$mvar]{"centring"} eq "cell") {
          $sub_string{$type}=$sub_string{$type}.
            "do ns = 1, ubound(region(var(m)%region_number)%ij,1)\n".
            "i = region(var(m)%region_number)%ij(ns)\n";
        } else {
          $sub_string{$type}=$sub_string{$type}.
            "do ns = 1, 1\n";
        }
        $sub_string{$type}=$sub_string{$type}.
          "call reset_funk(var(m)%funk(ns))\n";
# someloop
      } else {

# set up someloops that are relative to current location
        if ($variable{$type}[$mvar]{"region"} eq '<celljfaces>') {
          $sub_string{$type}=$sub_string{$type}.
            "! loop is conducted around region $variable{$type}[$mvar]{region}\n".
            "do ns = 1, ubound(cell(i)%jface,1)\n".
            "j = cell(i)%jface(ns)\n";
        } elsif ($variable{$type}[$mvar]{"region"} eq '<nobcelljfaces>') {
          $sub_string{$type}=$sub_string{$type}.
            "! loop is conducted around region $variable{$type}[$mvar]{region}\n".
            "! move loop to domain cell if cell is a boundary cell\n".
            "if (cell(i)%type == 2) i = face(cell(i)%jface(1))%icell(1)\n".
            "do ns = 1, ubound(cell(i)%jface,1)\n".
            "j = cell(i)%jface(ns)\n";
        } elsif ($variable{$type}[$mvar]{"region"} eq '<cellicells>') {
          $sub_string{$type}=$sub_string{$type}.
            "! loop is conducted around region $variable{$type}[$mvar]{region}\n".
            "do ns = 1, ubound(cell(ilast)%icell,1)\n".
            "i = cell(ilast)%icell(ns)\n";
        } elsif ($variable{$type}[$mvar]{"region"} eq '<faceicells>') {
          $sub_string{$type}=$sub_string{$type}.
            "! loop is conducted around region $variable{$type}[$mvar]{region}\n".
            "do ns = 1, ubound(face(j)%icell,1)\n".
            "i = face(j)%icell(ns)\n";
        } elsif ($variable{$type}[$mvar]{"region"} eq '<adjacentcellicells>') {
          $sub_string{$type}=$sub_string{$type}.
            "! loop is conducted around region $variable{$type}[$mvar]{region}\n".
            "do ns = 1, ubound(cell(ilast)%jface,1)\n".
            "i = cell(ilast)%icell(ns+1)\n";
        } elsif ($variable{$type}[$mvar]{"region"} eq '<adjacentfaceicells>') {
          $sub_string{$type}=$sub_string{$type}.
            "! loop is conducted around region $variable{$type}[$mvar]{region}\n".
            "do ns = 1, 2\n".
            "i = face(j)%icell(ns)\n";
        } elsif ($variable{$type}[$mvar]{"region"} eq '<upwindfaceicells>') {
          $sub_string{$type}=$sub_string{$type}.
            "! loop is conducted around region $variable{$type}[$mvar]{region}\n".
            "do ns = (3-flux_direction)/2, (3-flux_direction)/2\n".
            "i = face(j)%icell(ns)\n";
        } elsif ($variable{$type}[$mvar]{"region"} eq '<downwindfaceicells>') {
          $sub_string{$type}=$sub_string{$type}.
            "! loop is conducted around region $variable{$type}[$mvar]{region}\n".
            "do ns = (3+flux_direction)/2, (3+flux_direction)/2\n".
            "i = face(j)%icell(ns)\n";
        } elsif ($variable{$type}[$mvar]{"region"} eq '<cellkernelregion[l=0]>') {
          $sub_string{$type}=$sub_string{$type}.
            "! loop is conducted around region $variable{$type}[$mvar]{region}\n".
            "do ns = 1, ubound(cell(i)%kernel(0)%ijk,1)\n".
            "j = cell(i)%kernel(0)%ijk(ns)\n";
        } elsif ($variable{$type}[$mvar]{"region"} =~ /<cellkernelregion\[l=([123])\]>/) {
          $sub_string{$type}=$sub_string{$type}.
            "! loop is conducted around region $variable{$type}[$mvar]{region}\n".
            "do ns = 1, ubound(cell(ilast)%kernel($1)%ijk,1)\n".
            "i = cell(ilast)%kernel($1)%ijk(ns)\n";
        } elsif ($variable{$type}[$mvar]{"region"} =~ /<facekernelregion\[l=([0123456])\]>/) {
          $sub_string{$type}=$sub_string{$type}.
            "! loop is conducted around region $variable{$type}[$mvar]{region}\n".
            "do ns = 1, ubound(face(j)%kernel($1)%ijk,1)\n".
            "i = face(j)%kernel($1)%ijk(ns)\n";
        } elsif ($variable{$type}[$mvar]{"region"} =~ /<noloop>/ ||
                 $variable{$type}[$mvar]{"centring"} eq "none") {
          $sub_string{$type}=$sub_string{$type}.
            "! region is $variable{$type}[$mvar]{region} with centring $variable{$type}[$mvar]{centring} ".
            "so dummy loop is performed\n".
            "do ns = 1, 1\n";
        
# set up someloops that are absolute locations
        } elsif ($variable{$type}[$mvar]{"region"}) {
          $sub_string{$type}=$sub_string{$type}.
            "region_number = region_number_from_name(name='$variable{$type}[$mvar]{region}',centring='$variable{$type}[$mvar]{centring}')\n".
            "if (region_number == 0) stop 'ERROR: problem finding $variable{$type}[$mvar]{centring} region $variable{$type}[$mvar]{region} in update_someloops'\n\n".
            "do ns = 1, ubound(region(region_number)%ij,1)\n";
          if ($variable{$type}[$mvar]{"centring"} eq "cell") {
            $sub_string{$type}=$sub_string{$type}.
              "i = region(region_number)%ij(ns)\n";
          } elsif ($variable{$type}[$mvar]{"centring"} eq "face") {
            $sub_string{$type}=$sub_string{$type}.
              "j = region(region_number)%ij(ns)\n";
          }
        } else {
          die "ERROR: someloop has no region specified in create_fortran_equations\n";
        }


      }
  
# do inner loop operation
      if ($type eq "someloop" && ( $variable{$type}[$mvar]{"type"} eq "min" ||
        $variable{$type}[$mvar]{"type"} eq "max" )) {
# min or max

        $tmp = $mequation;
        maxima_to_fortran($tmp,$variable{$type}[$mvar]{"otype"},$variable{$type}[$mvar]{"omvar"});
        if ($variable{$type}[$mvar]{"type"} eq "min") {
          $inequality = "<";
        } else {
          $inequality = ">";
        }
        $sub_string{$type}=$sub_string{$type}.
          update_someloop_fortran($mequation)."\n".
          "if (".$tmp.$inequality.$variable{$type}[$mvar]{"fortran"}.") then\n".
          "call reset_funk(someloop(m))\n".
          $variable{$type}[$mvar]{"fortran"}." = ".$tmp."\n";
        if ($deriv) {
          $tmp = create_deriv_fortran($variable{$type}[$mvar]{"fortran"},$mequation,$variable{$type}[$mvar]{"otype"},$variable{$type}[$mvar]{"omvar"});
          $sub_string{$type}=$sub_string{$type}."$tmp\n";
        }
        $sub_string{$type}=$sub_string{$type}."end if\n";

      } elsif ($type eq "someloop" && $variable{$type}[$mvar]{"type"} eq "if" ) {
# if

        $deriv = $variable{$type}[$mvar]{"valuederiv"}; # deriv for value is separate from condition deriv (=0)
        $tmp = $mequation; # mequation is condition, so no derivative required 
        maxima_to_fortran($tmp,$variable{$type}[$mvar]{"otype"},$variable{$type}[$mvar]{"omvar"});
        $sub_string{$type}=$sub_string{$type}.
          update_someloop_fortran($mequation)."\n".
#         "if (".$tmp." >= 0.d0) then\n".
          "if (".$tmp." > 0.d0) then\n".
          "! value for positive condition (true)\n";
        $tmp = $variable{$type}[$mvar]{"true"}; 
        maxima_to_fortran($tmp,$variable{$type}[$mvar]{"otype"},$variable{$type}[$mvar]{"omvar"});
        $sub_string{$type}=$sub_string{$type}.
          update_someloop_fortran($variable{$type}[$mvar]{"true"})."\n".
          $variable{$type}[$mvar]{"fortran"}." = ".$tmp."\n";
        if ($deriv) {
          $tmp = create_deriv_fortran($variable{$type}[$mvar]{"fortran"},$variable{$type}[$mvar]{"true"},$variable{$type}[$mvar]{"otype"},$variable{$type}[$mvar]{"omvar"});
          $sub_string{$type}=$sub_string{$type}."$tmp\n";
        }
        $sub_string{$type}=$sub_string{$type}.
          "else\n".
          "! value for negative condition (false)\n";
        $tmp = $variable{$type}[$mvar]{"false"}; 
        maxima_to_fortran($tmp,$variable{$type}[$mvar]{"otype"},$variable{$type}[$mvar]{"omvar"});
        $sub_string{$type}=$sub_string{$type}.
          update_someloop_fortran($variable{$type}[$mvar]{"false"})."\n".
          $variable{$type}[$mvar]{"fortran"}." = ".$tmp."\n";
        if ($deriv) {
          $tmp = create_deriv_fortran($variable{$type}[$mvar]{"fortran"},$variable{$type}[$mvar]{"false"},$variable{$type}[$mvar]{"otype"},$variable{$type}[$mvar]{"omvar"});
          $sub_string{$type}=$sub_string{$type}."$tmp\n";
        }
        $sub_string{$type}=$sub_string{$type}."end if\n";

      } elsif ($type eq "someloop" && $variable{$type}[$mvar]{"type"} eq "product" ) {
# product

        $fequation = "$mequation";
        maxima_to_fortran($fequation,$variable{$type}[$mvar]{"otype"},$variable{$type}[$mvar]{"omvar"});
        $tmp = "$variable{$type}[$mvar]{fortranns} = $variable{$type}[$mvar]{fortranns}*(".$fequation.")";
        $sub_string{$type}=$sub_string{$type}.
          update_someloop_fortran($mequation)."\n".
          "$tmp\n\n".
          "! if any element is small then exit\n".
          "if (abs($fequation)<eps_dv) exit\n\n";

        if ($deriv) {
          $tmp = create_deriv_fortran($variable{$type}[$mvar]{"fortranns"},"log($mequation)",$variable{$type}[$mvar]{"otype"},$variable{$type}[$mvar]{"omvar"});
          $sub_string{$type}=$sub_string{$type}.
            "! using the product rule we first calculate the derivative of the log($mequation) for each element".
            "$tmp\n";
        }

      } else {
# sum

        $tmp = "$mequation";
        maxima_to_fortran($tmp,$variable{$type}[$mvar]{"otype"},$variable{$type}[$mvar]{"omvar"});
        $tmp = "$variable{$type}[$mvar]{fortranns} = $variable{$type}[$mvar]{fortranns} + (".$tmp.")";
        $sub_string{$type}=$sub_string{$type}.
          update_someloop_fortran($mequation)."\n".
          "$tmp\n";

        if ($deriv) {
          $tmp = create_deriv_fortran($variable{$type}[$mvar]{"fortranns"},$mequation,$variable{$type}[$mvar]{"otype"},$variable{$type}[$mvar]{"omvar"});
          $sub_string{$type}=$sub_string{$type}."$tmp\n";
        }

      }

# close loop
      $sub_string{$type}=$sub_string{$type}."end do\n\n";

# post loop operations
# alter derivative for product
      if ($type eq "someloop" && $variable{$type}[$mvar]{"type"} eq "product" && $deriv) {
        $tmp = $variable{$type}[$mvar]{"fortranns"}; # name as will be used in add_to_dv (fortran)
        $tmp =~ s/%v$//;
        $sub_string{$type}=$sub_string{$type}.
          "! and then multiply all derivatives by the resulting product of all elements\n".
          "call multiply_dv($tmp,$variable{$type}[$mvar]{fortranns})\n\n";
      }
        
      $mvar++;

    }

    if ($sub_string{$type}) {
      if ($type eq "constant") {
        $sub_string{$type}=$sub_string{$type}."\nend if\n";
      } else {
        $sub_string{$type}=$sub_string{$type}.
          "\nelse\nwrite(*,*) \"ERROR: the index \",m,\" was not found when updating $type\"\nstop\nend if\n";
      }
    }
  }

}

#-------------------------------------------------------------------------------
# here we scan a string for any someloops, and write corresponding fortran
#  to call update_someloop
# enter with $_[0] as the mequation (unchanged), exit with fortran

sub update_someloop_fortran {

  use strict;
  my ($fortran, $mvar);

  $fortran = "";

  $mvar = 1;
  while ($m{"someloop"} && $mvar <= $m{"someloop"}) {
    if ($_[0] =~ /\Q$variable{"someloop"}[$mvar]{"maxima"}/) {
      $fortran = $fortran."call update_someloop($mvar,i,j)\n";
    }
    $mvar++;
  }

  return $fortran;
}

#-------------------------------------------------------------------------------
# takes an expression and performs all possible derivatives on that expression, 
#  returning fortran code to evaluate those derivatives
# called with arguments $mname and $mequation - returns fortran string 

sub create_deriv_fortran {

  use strict;
  my ($typewrt, $mvarwrt, $tmp, $tmpwrt, $dfortran, $ii, $itmp);

  my $fortran = $_[0]; # name as will be used in add_to_dv (fortran)
  $fortran =~ s/%v$//;
  my $string = "";
  my $mequation = $_[1]; # maxima type equation as passed to routine
  my $otype = $_[2];
  my $omvar = $_[3];

  print DEBUG "in create_deriv_fortran with fortran = $fortran and mequation = $mequation\n";

# maxima_to_fortran($name); #convert to fortran format
# $name =~ s/%v//; # but lose subelement reference

# cycle through looking for unknowns, deriveds and someloops - anything that the equations ultimately depend on
  foreach $typewrt ("unknown","derived","someloop") {
    foreach $mvarwrt ( 1 .. $m{$typewrt} ) {
      $tmpwrt = $variable{$typewrt}[$mvarwrt]{"maxima"};
      if ($mequation !~ /\Q$tmpwrt/) {
#       print DEBUG "variable not present in expression so skipping\n";
        next;
      }
      print DEBUG "in create_deriv_fortran doing relative diff of: mequation = $mequation: wrt: tmpwrt = $tmpwrt\n";
      $tmp = "diff($mequation,$tmpwrt)";
      maxima_to_fortran($tmp,$otype,$omvar);
      if ($tmp) {
        $dfortran = $variable{$typewrt}[$mvarwrt]{"fortran"};
        $dfortran =~ s/%v//;
        $string=$string."\nderiv = $tmp\ncall add_to_dv($fortran,deriv,$dfortran)\n";
      }

    }
  }

  print DEBUG "leaving create_deriv_fortran with string = $string\n";

  return $string;
}

#-------------------------------------------------------------------------------
# creates details regarding the system (code generated) variables
# also some default regions

sub create_system_variables {

  use strict;
  my $mvar;
  my $type = "system";

# generally index order should be the same between maxima and fortran and be: i,j,ji,k,l,m,n,p
# i = cell (or ilast being the last cell reference)
# j = face
# k = node
# l = dimension, m = number (of variable), n = time

  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<cellx[l=1]>"; # cell centred location vector
  $variable{"system"}[$m{"system"}]{"centring"} = "cell";
  $variable{"system"}[$m{"system"}]{"maxima"} = "cellx[i,1]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "cell(i)%x(1)";
  $variable{"system"}[$m{"system"}]{"units"} = $lengthunit;
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<cellx[l=2]>"; # cell centred location vector
  $variable{"system"}[$m{"system"}]{"centring"} = "cell";
  $variable{"system"}[$m{"system"}]{"maxima"} = "cellx[i,2]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "cell(i)%x(2)";
  $variable{"system"}[$m{"system"}]{"units"} = $lengthunit;
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<cellx[l=3]>"; # cell centred location vector
  $variable{"system"}[$m{"system"}]{"centring"} = "cell";
  $variable{"system"}[$m{"system"}]{"maxima"} = "cellx[i,3]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "cell(i)%x(3)";
  $variable{"system"}[$m{"system"}]{"units"} = $lengthunit;
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facex[l=1]>"; # face centred location vector
  $variable{"system"}[$m{"system"}]{"centring"} = "face";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facex[j,1]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%x(1)";
  $variable{"system"}[$m{"system"}]{"units"} = $lengthunit;
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facex[l=2]>"; # face centred location vector
  $variable{"system"}[$m{"system"}]{"centring"} = "face";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facex[j,2]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%x(2)";
  $variable{"system"}[$m{"system"}]{"units"} = $lengthunit;
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facex[l=3]>"; # face centred location vector
  $variable{"system"}[$m{"system"}]{"centring"} = "face";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facex[j,3]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%x(3)";
  $variable{"system"}[$m{"system"}]{"units"} = $lengthunit;
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facedx>"; # face centred distance between cell centres
  $variable{"system"}[$m{"system"}]{"centring"} = "face";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facedx[j]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%dx";
  $variable{"system"}[$m{"system"}]{"units"} = $lengthunit;
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facearea>"; # face centred face area
  $variable{"system"}[$m{"system"}]{"centring"} = "face";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facearea[j]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%area";
  $variable{"system"}[$m{"system"}]{"units"} = "$lengthunit^2";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<cellvol>"; # cell centred cell volume
  $variable{"system"}[$m{"system"}]{"centring"} = "cell";
  $variable{"system"}[$m{"system"}]{"maxima"} = "cellvol[i]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "cell(i)%vol";
  $variable{"system"}[$m{"system"}]{"units"} = "$lengthunit^3";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facenorm[l=1]>"; # face centred unit normal vector
  $variable{"system"}[$m{"system"}]{"centring"} = "face";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facenorm[j,1]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%norm(1,1)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facenorm[l=2]>"; # face centred unit normal vector
  $variable{"system"}[$m{"system"}]{"centring"} = "face";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facenorm[j,2]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%norm(2,1)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facenorm[l=3]>"; # face centred unit normal vector
  $variable{"system"}[$m{"system"}]{"centring"} = "face";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facenorm[j,3]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%norm(3,1)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facetang1[l=1]>"; # face centred unit tangent vector
  $variable{"system"}[$m{"system"}]{"centring"} = "face";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facetang1[j,1]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%norm(1,2)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facetang1[l=2]>"; # face centred unit tangent vector
  $variable{"system"}[$m{"system"}]{"centring"} = "face";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facetang1[j,2]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%norm(2,2)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facetang1[l=3]>"; # face centred unit tangent vector
  $variable{"system"}[$m{"system"}]{"centring"} = "face";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facetang1[j,3]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%norm(3,2)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facetang2[l=1]>"; # face centred unit tangent vector
  $variable{"system"}[$m{"system"}]{"centring"} = "face";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facetang2[j,1]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%norm(1,3)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facetang2[l=2]>"; # face centred unit tangent vector
  $variable{"system"}[$m{"system"}]{"centring"} = "face";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facetang2[j,2]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%norm(2,3)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facetang2[l=3]>"; # face centred unit tangent vector
  $variable{"system"}[$m{"system"}]{"centring"} = "face";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facetang2[j,3]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%norm(3,3)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facekernel[l=0]>"; # kernel 0 associated with face which gives average from surrounding cells
  $variable{"system"}[$m{"system"}]{"centring"} = "cell";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facekernel[j,0,ns]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%kernel(0)%v(ns)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facekernel[l=1]>"; # kernel 1 associated with face which gives derivative in lth coordinate direction from surrounding cells
  $variable{"system"}[$m{"system"}]{"centring"} = "cell";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facekernel[j,1,ns]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%kernel(1)%v(ns)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facekernel[l=2]>"; # kernel 2 associated with face which gives derivative in lth coordinate direction from surrounding cells
  $variable{"system"}[$m{"system"}]{"centring"} = "cell";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facekernel[j,2,ns]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%kernel(2)%v(ns)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facekernel[l=3]>"; # kernel 3 associated with face which gives derivative in lth coordinate direction from surrounding cells
  $variable{"system"}[$m{"system"}]{"centring"} = "cell";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facekernel[j,3,ns]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%kernel(3)%v(ns)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facekernel[l=4]>"; # kernel 4 associated with face which gives derivative in face normal direction from surrounding cells
  $variable{"system"}[$m{"system"}]{"centring"} = "cell";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facekernel[j,4,ns]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%kernel(4)%v(ns)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facekernel[l=5]>"; # kernel 5 associated with face which gives derivative in face tang1 direction from surrounding cells
  $variable{"system"}[$m{"system"}]{"centring"} = "cell";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facekernel[j,5,ns]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%kernel(5)%v(ns)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<facekernel[l=6]>"; # kernel 6 associated with face which gives derivative in face tang2 direction from surrounding cells
  $variable{"system"}[$m{"system"}]{"centring"} = "cell";
  $variable{"system"}[$m{"system"}]{"maxima"} = "facekernel[j,6,ns]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "face(j)%kernel(6)%v(ns)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<cellkernel[l=0]>"; # kernel 0 associated with cell which gives average from surrounding faces
  $variable{"system"}[$m{"system"}]{"centring"} = "face";
  $variable{"system"}[$m{"system"}]{"maxima"} = "cellkernel[i,0,ns]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "cell(i)%kernel(0)%v(ns)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<cellkernel[l=1]>"; # kernel 1 associated with cell which gives derivative in lth coordinate direction from surrounding cells
  $variable{"system"}[$m{"system"}]{"centring"} = "cell";
  $variable{"system"}[$m{"system"}]{"maxima"} = "cellkernel[ilast,1,ns]"; # in all instances i will be used for kernel cell, and ilast will refer to central cell
  $variable{"system"}[$m{"system"}]{"fortran"} = "cell(ilast)%kernel(1)%v(ns)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<cellkernel[l=2]>"; # kernel 2 associated with cell which gives derivative in lth coordinate direction from surrounding cells
  $variable{"system"}[$m{"system"}]{"centring"} = "cell";
  $variable{"system"}[$m{"system"}]{"maxima"} = "cellkernel[ilast,2,ns]"; # in all instances i will be used for kernel cell, and ilast will refer to central cell
  $variable{"system"}[$m{"system"}]{"fortran"} = "cell(ilast)%kernel(2)%v(ns)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<cellkernel[l=3]>"; # kernel 3 associated with cell which gives derivative in lth coordinate direction from surrounding cells
  $variable{"system"}[$m{"system"}]{"centring"} = "cell";
  $variable{"system"}[$m{"system"}]{"maxima"} = "cellkernel[ilast,3,ns]"; # in all instances i will be used for kernel cell, and ilast will refer to central cell
  $variable{"system"}[$m{"system"}]{"fortran"} = "cell(ilast)%kernel(3)%v(ns)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<cellkernel[l=4]>"; # kernel 4 associated with cell which gives average from surrounding nodes
  $variable{"system"}[$m{"system"}]{"centring"} = "node";
  $variable{"system"}[$m{"system"}]{"maxima"} = "cellkernel[i,4,ns]";
  $variable{"system"}[$m{"system"}]{"fortran"} = "cell(i)%kernel(4)%v(ns)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<delta[l=1,1]>"; # coordinate direction delta function
  $variable{"system"}[$m{"system"}]{"maxima"} = "1.d0";
  $variable{"system"}[$m{"system"}]{"fortran"} = "1.d0";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<delta[l=1,2]>"; # coordinate direction delta function
  $variable{"system"}[$m{"system"}]{"maxima"} = "0.d0";
  $variable{"system"}[$m{"system"}]{"fortran"} = "0.d0";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<delta[l=1,3]>"; # coordinate direction delta function
  $variable{"system"}[$m{"system"}]{"maxima"} = "0.d0";
  $variable{"system"}[$m{"system"}]{"fortran"} = "0.d0";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<delta[l=2,1]>"; # coordinate direction delta function
  $variable{"system"}[$m{"system"}]{"maxima"} = "0.d0";
  $variable{"system"}[$m{"system"}]{"fortran"} = "0.d0";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<delta[l=2,2]>"; # coordinate direction delta function
  $variable{"system"}[$m{"system"}]{"maxima"} = "1.d0";
  $variable{"system"}[$m{"system"}]{"fortran"} = "1.d0";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<delta[l=2,3]>"; # coordinate direction delta function
  $variable{"system"}[$m{"system"}]{"maxima"} = "0.d0";
  $variable{"system"}[$m{"system"}]{"fortran"} = "0.d0";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<delta[l=3,1]>"; # coordinate direction delta function
  $variable{"system"}[$m{"system"}]{"maxima"} = "0.d0";
  $variable{"system"}[$m{"system"}]{"fortran"} = "0.d0";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<delta[l=3,2]>"; # coordinate direction delta function
  $variable{"system"}[$m{"system"}]{"maxima"} = "0.d0";
  $variable{"system"}[$m{"system"}]{"fortran"} = "0.d0";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<delta[l=3,3]>"; # coordinate direction delta function
  $variable{"system"}[$m{"system"}]{"maxima"} = "1.d0";
  $variable{"system"}[$m{"system"}]{"fortran"} = "1.d0";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<limitertolerance>"; # coordinate direction delta function
  $variable{"system"}[$m{"system"}]{"maxima"} = "limitertolerance";
  $variable{"system"}[$m{"system"}]{"fortran"} = "limitertolerance";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<limitercontgrad>"; # coordinate direction delta function
  $variable{"system"}[$m{"system"}]{"maxima"} = "limitercontgrad";
  $variable{"system"}[$m{"system"}]{"fortran"} = "limitercontgrad";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<pi>"; # pi
  $variable{"system"}[$m{"system"}]{"maxima"} = "pi";
  $variable{"system"}[$m{"system"}]{"fortran"} = "pi";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<huge>"; # largest representable number
  $variable{"system"}[$m{"system"}]{"maxima"} = "huge";
  $variable{"system"}[$m{"system"}]{"fortran"} = "huge(0.d0)";
  $m{"system"}++;
  $variable{"system"}[$m{"system"}]{"name"} = "<tiny>"; # smallest representable number
  $variable{"system"}[$m{"system"}]{"maxima"} = "tiny";
  $variable{"system"}[$m{"system"}]{"fortran"} = "tiny(0.d0)";

  foreach $mvar ( 1 .. $m{"system"} ) {
# create mfortran names for the variable, which are what maxima produces from the f90 package
    $variable{$type}[$mvar]{"mfortran"}=$variable{$type}[$mvar]{"maxima"}; # only differ by braces used on indicies
    $variable{$type}[$mvar]{"mfortran"}=~s/\[/(/g;
    $variable{$type}[$mvar]{"mfortran"}=~s/\]/)/g;
    if (!($variable{"system"}[$mvar]{"units"})) {$variable{"system"}[$mvar]{"units"} = "1";}
    if (!($variable{"system"}[$mvar]{"multiplier"})) {$variable{"system"}[$mvar]{"multiplier"} = "1.d0";}
    if (!($variable{"system"}[$mvar]{"centring"})) {$variable{"system"}[$mvar]{"centring"} = "none";}
    if (!($variable{"system"}[$mvar]{"region"})) {
      if ($variable{"system"}[$mvar]{"centring"} eq "cell") {
        $variable{"system"}[$mvar]{"region"} = "<all cells>";
      } elsif ($variable{"system"}[$mvar]{"centring"} eq "face") {
        $variable{"system"}[$mvar]{"region"} = "<all faces>";
      } else {
        $variable{"system"}[$mvar]{"region"} = "";
      }
    }
    if (!($variable{"system"}[$mvar]{"fortranns"})) { $variable{"system"}[$mvar]{"fortranns"} = $variable{"system"}[$mvar]{"fortran"}; }
    $variable{"system"}[$mvar]{"fortran_number"} = $mvar;
  }

}

#-------------------------------------------------------------------------------
# finds mvar number for system variable with name $_[0]
# if not found dies

sub get_system_mvar {

  use strict;
  my ($mvar);


  foreach $mvar ( 1 .. $m{"system"}+1 ) {
    if ($variable{"system"}[$mvar]{"name"} eq $_[0]) {exit;}
  }
  if ($mvar > $m{"system"}) {die "ERROR: system variable $_[0] not found in get_system_mvar\n";}

  return $mvar;

}

#-------------------------------------------------------------------------------
# takes a string which contains maxima names and replaces with user defined names
# this subroutine possibly not used

sub maxima_to_name {

  my $type;
  my $mvar;

  print DEBUG "in maxima_to_name: before _[0] = ",$_[0],"\n";

# substitute user names with maxima names
  foreach $type ("constant","transient","unknown","derived","output","system","empty") {
    foreach $mvar ( 1 .. $m{$type} ) {
      replace_substrings($_[0],$variable{$type}[$mvar]{"maxima"},$variable{$type}[$mvar]{"name"});
    }
  }

  print DEBUG "in maxima_to_name: after _[0] = ",$_[0],"\n";
  return $_[0];

}

#-------------------------------------------------------------------------------
# creates latexable summary files

sub write_latex {

  my ($type, $mvar);
  my ($type2, $mvar2);
  my ($tmp, $name, $sub, $sup, $part, $text, $texcodename, $texmultiplier, $textype);
  my ($pre, $mid, $post, $operator);

  print DEBUG "in write_latex";

  open(LATEX, ">$tmp_dir/variables.tex");

# create latex names for variables and write a tabular list to tmp.tex
  foreach $type ("constant","transient","unknown","derived","output") {
    print LATEX "%Variable definition for variable type = $type\n";
    foreach $mvar ( 1 .. $m{$type} ) {
      $tmp = $variable{$type}[$mvar]{"name"};
      $tmp =~ s/^<//; $tmp =~ s/>$//;
      $variable{$type}[$mvar]{"latex"} = "";
#     print LATEX "tmp = $tmp\n";
      while ($tmp) {
        $tmp =~ /^(.+?)(\s+|$)(.*)/;
        $part = $1;
        $tmp = $3;
        if ($part =~ /\(.+\)/) {
          $variable{$type}[$mvar]{"latex"} = $variable{$type}[$mvar]{"latex"}.'\text{'.$part.'}';
        } else {
          $part =~ /^(.+?)((_|\^|$).*)/;
          $name = $1;
          $part = $2;
          $sub = "";
          $sup = "";
          while ($part) {
            if ($part =~ /^_(.*?)((_|\^|$).*)/) {
              if ($sub) {$sub = $sub.',';}
              $sub = $sub.$1;
              $part = $2;
            } elsif ($part =~ /^\^(.*?)((_|\^|$).*)/) {
              if ($sup) {$sup = $sup.',';}
              $sup = $sup.$1;
              $part = $2;
            }
          }
# hard code in a few exceptions
          if ($name eq 'tau') {$name = '\tau';}
          if ($name eq 'mu') {$name = '\mu';}
          if ($name eq 'nu') {$name = '\nu';}
          if ($name eq 'alpha') {$name = '\alpha';}
          if ($name eq 'beta') {$name = '\beta';}
          if ($name eq 'gamma') {$name = '\gamma';}
          if ($name eq 'delta') {$name = '\delta';}
          if ($name eq 'epsilon') {$name = '\epsilon';}
          if ($name eq 'phi') {$name = '\phi';}
          if ($name eq 'theta') {$name = '\theta';}
          if ($name eq 'sigma') {$name = '\sigma';}
          if ($name eq 'rho') {$name = '\rho';}
          if ($sub) {
            $variable{$type}[$mvar]{"latex"} = $variable{$type}[$mvar]{"latex"}.'\scal['.$sub.']{'.$name.'}';
          } else {
            $variable{$type}[$mvar]{"latex"} = $variable{$type}[$mvar]{"latex"}.'\scal{'.$name.'}';
          }
          if ($sup) { $variable{$type}[$mvar]{"latex"} = $variable{$type}[$mvar]{"latex"}.'^\text{'.$sup.'}'; }
        }
#       print LATEX "tmp = $tmp: part = $part\n";
      }
      $texcodename = $variable{$type}[$mvar]{"name"};
      $texcodename =~ s/_/\\_/g;
      $texcodename =~ s/\^/\\^/g;
      $texcodename =~ s/\{/\\{/g;
      $texcodename =~ s/\}/\\}/g;
      $texmultiplier = $variable{$type}[$mvar]{"multiplier"};
      $texmultiplier =~ s/(\d+)\.?d((\+|\-|\d)(\d*))/$1\\ee{$2}/g; # replace double precision floats with exponential
      $texmultiplier =~ s/(\d\.?)\\ee\{0\}/$1/g;
      if ($variable{$type}[$mvar]{"units"}) {
        print LATEX '$'.$variable{$type}[$mvar]{"latex"}.'$ & \units{'.$variable{$type}[$mvar]{"units"}.'} & '.
          $texmultiplier.' & \code{'.$texcodename."} & ".$variable{$type}[$mvar]{"comments"}." \\\\\n";
      } else {
        print LATEX '$'.$variable{$type}[$mvar]{"latex"}.'$ & & '.
          $texmultiplier.' &  \code{'.$texcodename."} & ".$variable{$type}[$mvar]{"comments"}." \\\\\n";
      }
    }
  }
  close(LATEX);

# create and write latex equations
  open(LATEX, ">$tmp_dir/equations.tex");

  foreach $type ("equation","derived","output") {
    $textype = $type;
    $textype =~ s/_/ /g;
    $textype =~ s/$/s/g; # plural
    $textype = "\U$textype";
    print LATEX "\n\\equationheading{$textype}\n";
    print LATEX "%Equation type = $type\n";
    foreach $mvar ( 1 .. $m{$type} ) {
      print LATEX "%Equation name = $variable{$type}[$mvar]{name}\n";
      print LATEX "%Equation = $variable{$type}[$mvar]{equation}\n";
      $tmp = $variable{$type}[$mvar]{"equation"};
      if ($type =~ /function$/) { $tmp = "$variable{$type}[$mvar]{name} = $tmp"; }
      else { $tmp = "$tmp = 0"; }
# swap div and grad names
      while ($tmp =~ /div\(|grad\(|\^\(|min\(|max\(|log\(/) {
        $pre = $`;
        $post = '('.$';
        $operator = $&;
        $operator =~ s/\($//;
        ($mid,$post) = extract_bracketed( $post, '()' ); # http://search.cpan.org/~adamk/Text-Balanced-2.02/lib/Text/Balanced.pm
        if ($operator eq '^') {
          $tmp = $pre.'^{'.$mid.'}'.$post;
        } elsif ($operator eq 'div') {
          $mid =~ s/^\(//; $mid =~ s/\)$//;
          $tmp = $pre.'\vect{\nabla} \cdot \left ('.$mid.' \right )'.$post;
#         $tmp = $pre.'\frac{d}{dx}\left ('.$mid.' \right )'.$post;
        } elsif ($operator eq 'grad') {
          $mid =~ s/^\(//; $mid =~ s/\)$//;
          $tmp = $pre.'\vect{\nabla} \left ('.$mid.' \right )'.$post;
#         $tmp = $pre.'\frac{d}{dx}\left ('.$mid.' \right )'.$post;
        } else {
          $mid =~ s/^\(//; $mid =~ s/\)$//;
          $tmp = $pre.'\\'.$operator.'\left ('.$mid.' \right )'.$post;
        }
      }
# remove *
      $tmp =~ s/\*\*/\^/g;
#     $tmp =~ s/>\*</></g; 
#     $tmp =~ s/\*/\\times/g; 
      $tmp =~ s/\*//g; 
      $tmp =~ s/(\d+)\.?d((\+|\-|\d)(\d*))/$1\\ee{$2}/g; # replace double precision floats with exponential
      $tmp =~ s/(\d\.?)\\ee\{0\}/$1/g;

      foreach $type2 ("constant","transient","unknown","derived") {
        foreach $mvar2 ( 1 .. $m{$type2} ) {
          $tmp =~ s/\Q$variable{$type2}[$mvar2]{"name"}/$variable{$type2}[$mvar2]{"latex"}/g;
        }
      }
      print LATEX "\\begin{equation}\n$tmp\n\\end{equation}\n";
    }
  }

  close(LATEX);

}
#-------------------------------------------------------------------------------
