! file src/setup_module.f90
!
! Copyright 2009-2011 Dalton Harvie (daltonh@unimelb.edu.au)
! 
! This file is part of arb finite volume solver, referred to as `arb'.
! 
! arb is a software package designed to solve arbitrary partial
! differential equations on unstructured meshes using the finite volume
! method.  Primarily it consists of fortran source code, perl source
! code and shell scripts.  arb replies on certain third party software
! to run, most notably the computer algebra system maxima
! <http://maxima.sourceforge.net/> which is released under the GNU GPL.
! 
! The copyright of arb is held by Dalton Harvie.
! 
! arb is released under the GNU GPL.  arb is free software: you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License (version 3) as published by the Free Software Foundation.
! You should have received a copy of the GNU General Public License
! along with arb (see file license/gpl.txt after unpacking).  If not,
! see <http://www.gnu.org/licenses/>.
! 
! arb is distributed in the hope that it will be useful, but WITHOUT
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
! FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
! for more details.
! 
! For full details of arb's license see the license directory.
! 
! The current homepage for the arb finite volume solver project is
! <http://www.chemeng.unimelb.edu.au/people/staff/daltonh/downloads/arb>.
!
!-------------------------------------------------------------------------
module setup_module

!-----------------------------------------------------------------
contains

!-----------------------------------------------------------------

subroutine setup

! this handles everything prior to a simulation starting

use general_module
use equations_module
use gmsh_module
use kernel_module
logical, parameter :: debug = .true.

if (debug) write(*,'(80(1h+)/a)') 'subroutine setup'

if (debug) write(*,*) 'calling setup_gmsh_element'

call setup_gmsh_element ! create reference gmsh_element array (used for output as well as possible input)

if (debug) write(*,*) 'calling allocate_meta_arrays'

call allocate_meta_arrays ! allocate arrays that have meta data about variables

if (debug) write(*,*) 'calling read_constants_file'

call read_constants_file ! read constants file for numerical data, some region data and the mesh

if (debug) write(*,*) 'calling setup_mesh'

call setup_mesh ! read in mesh and create associated arrays

if (debug) write(*,*) 'calling setup_kernels'

call setup_kernels ! create averaging and differencing kernels

if (debug) write(*,*) 'calling setup_regions'

call setup_regions ! assign ij indices to each region not setup by mesh

if (debug) write(*,*) 'calling setup_vars'

call setup_vars ! do setup for each var variable

if (debug) write(*,'(a/80(1h-))') 'subroutine setup'

end subroutine setup

!-----------------------------------------------------------------

subroutine read_constants_file

! here we read in all sorts of simulation data from the constants.in file, but
! not the actual constants values (which is done in read_constants)

use general_module
integer, parameter :: fconstants = 1
integer :: error, cut, n, l
character(len=1000) :: textline, keyword, name, formatline, region_name, location
character(len=4) :: centring
real :: versiontmp
logical, parameter :: debug = .true.
                  
if (debug) write(*,'(80(1h+)/a)') 'subroutine read_constants_file'

dimensions = 0 ! set to default incase it is not set in constants file

! read in constants values from constants.in file

write(*,'(a)') "INFO: reading simulation information from constants file "//trim(constants_file)
open(unit=fconstants,file=trim(constants_file),status='old',iostat=error)
if (error /= 0) stop 'ERROR: problem opening constants file'

fileloop: do
  read(fconstants,'(a)',iostat=error) textline
  if (error /= 0) exit fileloop ! reached end of file
  call remove_comments(textline) ! remove comments from line
  cut=scan(textline,' ') ! split input line at first space
  if (cut<=1) cycle fileloop ! if there is space at first character or string is empty, cycle
  keyword=textline(1:cut-1)
  textline=adjustl(textline(cut:len(textline)))
! if (debug) write(*,'(a)') 'keyword = '//trim(keyword)//': textline = '//trim(textline)//''
! if (debug) write(*,'(a)') 'keyword = '//trim(keyword)

!---------------
! name
  if (trim(keyword) == 'FACE_REGION'.or.trim(keyword) == 'CELL_REGION') then ! read in variable names
    cut=scan(textline,'>') ! find end of name string
    if (textline(1:1) /= '<'.or.cut<2) then
      write(*,*) 'ERROR: name in '//trim(keyword)//' in constants file incorrectly specified on line:'
      write(*,*) trim(textline)
      stop
    end if
!   name = textline(2:cut-1)
    name = textline(1:cut) ! name now includes <>
    textline=adjustl(textline(cut+1:len(textline)))
  end if

!---------------
! location

  if (trim(keyword) == 'FACE_REGION' .or. trim(keyword) == 'CELL_REGION') then
    read(textline,*,iostat=error) location ! with no format specification string in constants file should be quotted
    if (error /= 0) then
      write(*,*) 'ERROR: problem reading in location for region '//name
      stop
    end if
    if (trim(keyword) == 'FACE_REGION') then
      centring = 'face'
    else
      centring = 'cell'
    end if
    m = region_number_from_name(name=name,location=location,centring=centring,creatable=.true.)
    if (m == 0) then
      write(*,*) 'ERROR: allocation of region = '//trim(region_name)// &
        ' failed, most likely because centring was inconsistent with previously set centring'
      stop
    end if
    if (debug) write(*,'(a,i3,a,a,a,a)') '(re)set region number ',m,' and name ',trim(region(m)%name), &
      ' to location ',trim(region(m)%location)
  end if

!---------------
! mesh info

  if (trim(keyword) == 'READ_GMSH') then
    read(textline,*,iostat=error) gmsh_file
    if (error /= 0) then
      write(*,*) 'ERROR: problem reading in name of mesh file from line '//textline
      stop
    end if
    if (debug) write(*,'(a,a)') 'INFO: gmsh_file = ',trim(gmsh_file)
  end if

!---------------
! dimensions

  if (trim(keyword) == 'DIMENSIONS') then
    read(textline,*,iostat=error) dimensions
    if (error /= 0) then
      write(*,*) 'ERROR: problem reading in number of dimensions from line '//textline
      stop
    end if
    if (debug) write(*,'(a,i1)') 'INFO: dimensions = ',dimensions
  end if

!---------------
! newtrestol

  if (trim(keyword) == 'NEWTRESTOL') then
    read(textline,*,iostat=error) newtrestol
    if (error /= 0) then
      write(*,*) 'ERROR: problem reading in newton loop tolerance from line '//textline
      stop
    end if
    if (debug) write(*,'(a,g14.6)') 'INFO: newtrestol = ',newtrestol
  end if

!---------------
! newtstepmax

  if (trim(keyword) == 'NEWTSTEPMAX') then
    read(textline,*,iostat=error) newtstepmax
    if (error /= 0) then
      write(*,*) 'ERROR: problem reading in maximum number of newton steps from line '//textline
      stop
    end if
    if (debug) write(*,'(a,i5)') 'INFO: newtstepmax = ',newtstepmax
  end if

!---------------
! version

  if (trim(keyword) == 'VERSION') then
    read(textline,*,iostat=error) versiontmp
    if (version /= versiontmp) then
      write(*,'(a,f5.2,a,f5.2,a)') 'ERROR: version mismatch between constants.in (',versiontmp,') and arb (',version,')'
      stop
    end if
  end if

!---------------
! linear solver type

  if (trim(keyword) == 'LINEAR_SOLVER') then
    read(textline,*,iostat=error) linear_solver
    if (error /= 0) then
      write(*,*) 'ERROR: problem reading in linear solver type from line '//textline
      stop
    end if
    if (debug) write(*,'(a,a)') 'INFO: linear_solver = ',trim(linear_solver)
  end if

!---------------

end do fileloop

close(fconstants)

if (debug) write(*,'(a/80(1h-))') 'subroutine read_constants_file'

end subroutine read_constants_file

!-----------------------------------------------------------------

subroutine setup_mesh

! here we do all sorts of mesh setting
! set types for all cells, faces and nodes
! check on icell associations for faces
! set i,j,k boundary and domain and numbers
! set x for faces and nodes

use general_module
use gmsh_module
integer, parameter :: fdetail = 1
integer :: i, ii, j, jj, k, kk, l, jj2, ji, ij, ji2, inew, jimax, ii2, ncells
type(cell_type) :: default_cell
double precision, dimension(totaldimensions) :: tangc, normc
character(len=1000) :: formatline, filename
logical, parameter :: debug = .true., mesh_details_file = .false.

if (debug) write(*,'(80(1h+)/a)') 'subroutine setup_mesh'

!-------------------------------------
! create the basic mesh

if (gmsh_file /= '') then
  if (debug) write(*,*) 'calling read_gmsh_mesh'
  call read_gmsh_mesh ! read in gmsh
else
  stop 'ERROR: no other mesh types done yet'
end if

if (debug) write(*,*) 'doing generic mesh setup'

!-------------------------------------
! run through cells checking icells and possibly setting icell(2) indicies
do i = 1, itotal
  cell(i)%type = 1 ! domain cell
  jj_loop: do jj = 1, ubound(cell(i)%jface,1)
    j = cell(i)%jface(jj)
    if (face(j)%icell(1) == 0) then
      write(*,*) 'ERROR: a face that is adjacent to a cell has no icell=1 attached: j = ',j,': face = ',trim(print_face(face(j)))
      stop
    end if
    if (face(j)%icell(1) == i) cycle jj_loop
    if (ubound(face(j)%icell,1) < 2) call push_integer_array(array=face(j)%icell,new_element=0)
    if (face(j)%icell(2) == i) cycle jj_loop
    if (face(j)%icell(2) /= 0) then
      write(*,*) 'ERROR: a face has more than 2 cells attached to it: j = ',j,': face = ',trim(print_face(face(j)))
      stop
    end if
    face(j)%icell(2) = i
  end do jj_loop
end do
      
!-------------------------------------
! run through nodes setting type to default
do k = 1, ktotal
  node(k)%type = 1
end do

!-------------------------------------
! now assign face type and create any boundary cells
idomain = itotal
jboundary = 0
jdomain = 0

if (debug) write(*,*) 'creating boundary cells'

do j = 1, jtotal
  if (face(j)%icell(1) == 0) then
    write(*,*) 'ERROR: a face has no icell=1 attached to it: j = ',j,': face = ',trim(print_face(face(j)))
    stop
  end if
  if (ubound(face(j)%icell,1) < 2) call push_integer_array(array=face(j)%icell,new_element=0)
  if (face(j)%icell(2) == 0) then
    face(j)%type = 2 ! boundary face
    jboundary = jboundary + 1

! create new boundary cell, storing in default_cell
    call reset_cell(default_cell)
    default_cell%type = 2
    default_cell%dimensions = face(j)%dimensions
    default_cell%gmsh_element = face(j)%gmsh_element ! adopt gmsh element number of face - may not be set
    default_cell%gmsh_type = face(j)%gmsh_type ! adopt gmsh element type - must be set
    call resize_integer_array(keep_data=.false.,array=default_cell%jface,new_size=1)
    default_cell%jface(1) = j
    call copy_integer_array(original=face(j)%knode,copy=default_cell%knode)
! boundary cells adopt gmsh regions from adjacent domain cell
    call copy_integer_array(original=cell(face(j)%icell(1))%region_list,copy=default_cell%region_list)
    call copy_integer_array(original=cell(face(j)%icell(1))%gmsh_region_list,copy=default_cell%gmsh_region_list)

    itotal = itotal + 1
    if (ubound(cell,1) < itotal) call resize_cell_array(change=int(ktotal/10+1))
    call set_cell(cell_to_set=cell(itotal),new_value=default_cell)
    face(j)%icell(2) = itotal

!   if (debug) write(*,*) 'creating new boundary cell: i = ',itotal,': jface = ',default_cell%jface

! set node icell references for the new cell
    call add_icell_to_nodes(itotal,default_cell%knode)

! nodes associated with boundary faces are also boundary faces
    do kk = 1, ubound(face(j)%knode,1)
      k = face(j)%knode(kk)
      node(k)%type = 2
    end do

  else
    face(j)%type = 1 ! domain face
    jdomain = jdomain + 1
  end if

end do

iboundary = itotal - idomain

! resize cell arrays
call resize_cell_array(new_size=itotal)

!-------------------------------------
! run through nodes again counting types

kboundary = 0
do k = 1, ktotal
  if (node(k)%type == 2) kboundary = kboundary + 1
end do
kdomain = ktotal - kboundary

!-------------------------------------
! run through faces increasing icell

do j = 1, jtotal
! first two elements (the immediate cell neighbours) are already present

  if (.true.) then
! second elements are from the surrounding nodes
    kk_loop: do kk = 1, ubound(face(j)%knode,1)
      k = face(j)%knode(kk)
      ii_loop: do ii = 1, ubound(node(k)%icell,1)
        if (location_in_list(array=face(j)%icell,element=node(k)%icell(ii)) == 0) &
          call push_integer_array(array=face(j)%icell,new_element=node(k)%icell(ii))
      end do ii_loop
    end do kk_loop
  else
! new method based solely on expansion through number of degrees of separation
    do n = 2, 6 ! loop through number of expansion stages
      ncells = ubound(face(j)%icell,1) ! save list of cells that are included at the start
      do ii = 1, ncells
        i = face(j)%icell(ii)
        do jj = 1, ubound(cell(i)%jface,1)
          if (location_in_list(array=face(j)%icell,element=face(cell(i)%jface(jj))%icell(1)) == 0) then
            call push_integer_array(array=face(j)%icell,new_element=face(cell(i)%jface(jj))%icell(1))
          else if (location_in_list(array=face(j)%icell,element=face(cell(i)%jface(jj))%icell(2)) == 0) then
            call push_integer_array(array=face(j)%icell,new_element=face(cell(i)%jface(jj))%icell(2))
          end if
        end do
      end do
    end do
  end if

end do

!-------------------------------------
! run through cells setting icell (surrounding cells)
! this list includes all cells that share a common node with the central cell,
! included in the order of:
! 1 = central cell: 2-ubound(cell%jface,1)+1 = cells sharing a face: rest are the
! cells that only share a node

do i = 1, itotal

! first element is central cell
  call resize_integer_array(keep_data=.false.,array=cell(i)%icell,new_size=1)
  cell(i)%icell(1) = i

! second elements are from the surrounding faces
  do jj = 1, ubound(cell(i)%jface,1)
    j = cell(i)%jface(jj)
    do ii = 1, 2
      if (face(j)%icell(ii) /= 0 .and. face(j)%icell(ii) /= i) then
        call push_integer_array(array=cell(i)%icell,new_element=face(j)%icell(ii))
      end if
    end do
  end do

! third elements are from the surrounding nodes
  do kk = 1, ubound(cell(i)%knode,1)
    k = cell(i)%knode(kk)
    ii2_loop: do ii = 1, ubound(node(k)%icell,1)
      if (location_in_list(array=cell(i)%icell,element=node(k)%icell(ii)) == 0) &
        call push_integer_array(array=cell(i)%icell,new_element=node(k)%icell(ii))
    end do ii2_loop
  end do

end do

!-------------------------------------
! run through faces and cells checking that number of components is consistent with dimensions

if (debug) write(*,*) 'checking mesh consistency'

do i = 1, itotal
  if (cell(i)%type == 2) then
    if (ubound(cell(i)%jface,1) /= 1 ) then
      write(*,*) 'ERROR: boundary cell does not have 1 face'
      write(*,'(a,i6,a)') 'cell: i = ',i,': ',trim(print_cell(cell(i)))
      stop
    end if
    if (cell(i)%dimensions == 2) then
      if (ubound(cell(i)%knode,1) < 3 ) then
        write(*,*) 'ERROR: 2d boundary cell has less than 3 nodes'
        write(*,'(a,i6,a)') 'cell: i = ',i,': ',trim(print_cell(cell(i)))
        stop
      end if
    else if (cell(i)%dimensions == 1) then
      if (ubound(cell(i)%knode,1) /= 2 ) then
        write(*,*) 'ERROR: 1d boundary cell does not have 2 nodes'
        write(*,'(a,i6,a)') 'cell: i = ',i,': ',trim(print_cell(cell(i)))
        stop
      end if
    else if (cell(i)%dimensions == 0) then
      if (ubound(cell(i)%knode,1) /= 1 ) then
        write(*,*) 'ERROR: 0d boundary cell does not have 1 node'
        write(*,'(a,i6,a)') 'cell: i = ',i,': ',trim(print_cell(cell(i)))
        stop
      end if
    else
      write(*,*) 'ERROR: boundary cell has incorrect dimensions/nodes'
      write(*,'(a,i6,a)') 'cell: i = ',i,': ',trim(print_cell(cell(i)))
      stop
    end if
  else
    if (cell(i)%dimensions == 3) then
      if (ubound(cell(i)%jface,1) < 4 .or. ubound(cell(i)%knode,1) < 4) then
        write(*,*) 'ERROR: 3d domain cell has less than 4 faces/nodes'
        write(*,'(a,i6,a)') 'cell: i = ',i,': ',trim(print_cell(cell(i)))
        stop
      end if
    else if (cell(i)%dimensions == 2) then
      if (ubound(cell(i)%jface,1) < 3 .or. ubound(cell(i)%knode,1) < 3) then
        write(*,*) 'ERROR: 2d domain cell has less than 3 faces/nodes'
        write(*,'(a,i6,a)') 'cell: i = ',i,': ',trim(print_cell(cell(i)))
        stop
      end if
    else if (cell(i)%dimensions == 1) then
      if (ubound(cell(i)%jface,1) /= 2  .or. ubound(cell(i)%knode,1) /= 2) then
        write(*,*) 'ERROR: 1d domain cell does not have 2 faces/nodes'
        write(*,'(a,i6,a)') 'cell: i = ',i,': ',trim(print_cell(cell(i)))
        stop
      end if
    else
      write(*,*) 'ERROR: domain cell has incorrect dimensions'
      write(*,'(a,i6,a)') 'cell: i = ',i,': ',trim(print_cell(cell(i)))
      stop
    end if
    do jj = 1, ubound(cell(i)%jface,1)
      j = cell(i)%jface(jj)
      if (face(j)%dimensions /= cell(i)%dimensions - 1) then
        write(*,*) 'ERROR: face has dimensions which are inconsistent with adjacent cell'
        write(*,'(a,i6,a)') 'cell: i = ',i,': ',trim(print_cell(cell(i)))
        write(*,'(a,i6,a)') 'face: j = ',j,': ',trim(print_face(face(j)))
        stop
      end if
    end do
  end if
end do

do j = 1, jtotal
  if (face(j)%dimensions == 2) then
    if (ubound(face(j)%knode,1) < 3 ) then
      write(*,*) 'ERROR: 2d face has less than 3 nodes'
      write(*,'(a,i6,a)') 'face: j = ',j,': ',trim(print_face(face(j)))
      stop
    end if
  else if (face(j)%dimensions == 1) then
    if (ubound(face(j)%knode,1) /= 2 ) then
      write(*,*) 'ERROR: 1d face does not have 2 nodes'
      write(*,'(a,i6,a)') 'face: j = ',j,': ',trim(print_face(face(j)))
      stop
    end if
  else if (face(j)%dimensions == 0) then
    if (ubound(face(j)%knode,1) /= 1 ) then
      write(*,*) 'ERROR: 0d face does not have 1 node'
      write(*,'(a,i6,a)') 'face: j = ',j,': ',trim(print_face(face(j)))
      stop
    end if
  else
    write(*,*) 'ERROR: face has incorrect dimensions/nodes'
    write(*,'(a,i6,a)') 'face: j = ',j,': ',trim(print_face(face(j)))
    stop
  end if
end do

!-------------------------------------
! run through cells calculating vol and x (centre) 

if (debug) write(*,*) 'calculating mesh cell geometries'

do i = 1, itotal
  if (cell(i)%dimensions == 3) then
    call find_3d_geometry(jface=cell(i)%jface,volume=cell(i)%vol,centre=cell(i)%x)
  else if (cell(i)%dimensions == 2) then
    call find_2d_geometry(knode=cell(i)%knode,area=cell(i)%vol,centre=cell(i)%x)
  else if (cell(i)%dimensions == 1) then
    cell(i)%vol = distance( node(cell(i)%knode(1))%x , node(cell(i)%knode(2))%x )
    cell(i)%x = ( node(cell(i)%knode(1))%x + node(cell(i)%knode(2))%x )/2.d0
  else if (cell(i)%dimensions == 0) then
    cell(i)%vol = 1.d0
    cell(i)%x = node(cell(i)%knode(1))%x
  end if
end do

!-------------------------------------
! run through faces calculating area, x (centre) and norm

if (debug) write(*,*) 'calculating mesh face geometries'

do j = 1, jtotal
  if (face(j)%dimensions == 2) then
    call find_2d_geometry(knode=face(j)%knode,area=face(j)%area,norm=face(j)%norm,centre=face(j)%x)
  else if (face(j)%dimensions == 1) then
    face(j)%area = distance( node(face(j)%knode(1))%x , node(face(j)%knode(2))%x )
    tangc = cell(face(j)%icell(2))%x - cell(face(j)%icell(1))%x ! vector that is in same plane as the adjacent cells
    normc = cross_product( tangc , node(face(j)%knode(2))%x - node(face(j)%knode(1))%x )
    face(j)%norm(:,1) = cross_product( node(face(j)%knode(2))%x - node(face(j)%knode(1))%x , normc )
    call normalise_vector(face(j)%norm(:,1))
    face(j)%norm(:,2) = node(face(j)%knode(2))%x - node(face(j)%knode(1))%x ! vector that is tangent to line
    call normalise_vector(face(j)%norm(:,2))
    face(j)%norm(:,3) = cross_product( face(j)%norm(:,1) , face(j)%norm(:,2) ) ! third vector which points out of solution plane
    call normalise_vector(face(j)%norm(:,3))
    face(j)%x = ( node(face(j)%knode(1))%x + node(face(j)%knode(2))%x )/2.d0
  else if (face(j)%dimensions == 0) then
    face(j)%area = 1.d0
    face(j)%norm(:,1) = cell(face(j)%icell(2))%x - cell(face(j)%icell(1))%x
    call normalise_vector(face(j)%norm(:,1))
! from orthogonal vectors select one which has smallest dot product with first
    face(j)%norm(:,2) = [1.d0, 0.d0, 0.d0]
    face(j)%norm(:,3) = [0.d0, 1.d0, 0.d0]
    if (abs(dot_product(face(j)%norm(:,2),face(j)%norm(:,1))) < abs(dot_product(face(j)%norm(:,3),face(j)%norm(:,1))) ) then
      face(j)%norm(:,3) = cross_product( face(j)%norm(:,1) , face(j)%norm(:,2) )
      face(j)%norm(:,2) = cross_product( face(j)%norm(:,1) , face(j)%norm(:,3) )
    else
      face(j)%norm(:,2) = cross_product( face(j)%norm(:,1) , face(j)%norm(:,3) )
      face(j)%norm(:,3) = cross_product( face(j)%norm(:,1) , face(j)%norm(:,2) )
    end if
    call normalise_vector(face(j)%norm(:,2))
    call normalise_vector(face(j)%norm(:,3))
    face(j)%x = node(face(j)%knode(1))%x
  end if
  face(j)%dx = distance( cell(face(j)%icell(1))%x, cell(face(j)%icell(2))%x ) 
end do

!------------------------------------------
! write out mesh details if requested

write(*,'(a)') 'INFO: number of mesh elements::'
formatline = '(3(a,'//trim(indexformat)//'))'
write(*,fmt=formatline) ' NODES: ktotal = ',ktotal,': kdomain = ',kdomain,': kboundary = ',kboundary
formatline = '(3(a,'//trim(indexformat)//'))'
write(*,fmt=formatline) ' FACES: jtotal = ',jtotal,': jdomain = ',jdomain,': jboundary = ',jboundary
formatline = '(3(a,'//trim(indexformat)//'))'
write(*,fmt=formatline) ' CELLS: itotal = ',itotal,': idomain = ',idomain,': iboundary = ',iboundary

if (mesh_details_file) then
  if (debug) write(*,*) 'writing mesh details to mesh_details.txt file'

  filename = "output/mesh_details.txt"
  open(fdetail,file=trim(filename),status='replace',iostat=ierror)
  if (ierror /= 0) then
    write(*,*) 'ERROR: problem opening file ',filename
    stop
  end if

  write(fdetail,'(a)') 'MESH DETAILS:'
  formatline = '(3(a,'//trim(indexformat)//'))'
  write(fdetail,fmt=formatline) 'NODES: ktotal = ',ktotal,': kdomain = ',kdomain,': kboundary = ',kboundary
  formatline = '(a,'//trim(indexformat)//',a,a)'
  do k = 1,ktotal
    write(fdetail,fmt=formatline) 'node: k = ',k,': ',trim(print_node(node(k)))
  end do
  formatline = '(3(a,'//trim(indexformat)//'))'
  write(fdetail,fmt=formatline) 'FACES: jtotal = ',jtotal,': jdomain = ',jdomain,': jboundary = ',jboundary
  formatline = '(a,'//trim(indexformat)//',a,a)'
  do j = 1,jtotal
    write(fdetail,fmt=formatline) 'face: j = ',j,': ',trim(print_face(face(j)))
  end do
  formatline = '(3(a,'//trim(indexformat)//'))'
  write(fdetail,fmt=formatline) 'CELLS: itotal = ',itotal,': idomain = ',idomain,': iboundary = ',iboundary
  formatline = '(a,'//trim(indexformat)//',a,a)'
  do i = 1,itotal
    write(fdetail,fmt=formatline) 'cell: i = ',i,': ',trim(print_cell(cell(i)))
  end do

  close(fdetail)
end if

if (debug) write(*,'(a/80(1h-))') 'subroutine setup_mesh'

end subroutine setup_mesh

!-----------------------------------------------------------------

subroutine setup_regions

! here we setup the regions by finding the i or j indices for each

use general_module
use gmsh_module
integer, parameter :: fdetail = 1
integer :: error, m, i, j, n, ineighbour, cut, nregion, ijregion, nsregion, ns, nscompound
double precision :: tmp, tmpmax
double precision, dimension(totaldimensions) :: x ! a single location
character(len=1000) :: textline, keyword, name, formatline, location, aregion, region_list, filename
character(len=4) :: centring
character(len=1) :: rsign
logical :: existing, one_region
logical, parameter :: debug = .true., region_details_file = .false.
                  
if (debug) write(*,'(80(1h+)/a)') 'subroutine setup_regions'

if (debug) write(*,*) 'REGIONS:'

! create system generated regions

location='SYSTEM'
centring='cell'

m = region_number_from_name(name='<all cells>',location=location,centring=centring,existing=existing,creatable=.true.)
if (existing) then
  write(*,*) "ERROR: the system region = "//trim(name)//" has already been set:- this name should not be used user-defined regions"
  stop
end if
allocate(region(m)%ij(itotal))
do n = 1, itotal
  region(m)%ij(n) = n
end do
  
m = region_number_from_name(name='<domain>',location=location,centring=centring,existing=existing,creatable=.true.)
if (existing) then
  write(*,*) "ERROR: the system region = "//trim(name)//" has already been set:- this name should not be used user-defined regions"
  stop
end if
allocate(region(m)%ij(idomain))
n = 0
do i = 1, itotal
  if (cell(i)%type == 1) then
    n = n + 1
    region(m)%ij(n) = i
  end if
end do
  
m = region_number_from_name(name='<boundary cells>',location=location,centring=centring,existing=existing,creatable=.true.)
if (existing) then
  write(*,*) "ERROR: the system region = "//trim(name)//" has already been set:- this name should not be used user-defined regions"
  stop
end if
allocate(region(m)%ij(iboundary))
n = 0
do i = 1, itotal
  if (cell(i)%type == 2) then
    n = n + 1
    region(m)%ij(n) = i
  end if
end do
  
centring='face'

m = region_number_from_name(name='<all faces>',location=location,centring=centring,existing=existing,creatable=.true.)
if (existing) then
  write(*,*) "ERROR: the system region = "//trim(name)//" has already been set:- this name should not be used user-defined regions"
  stop
end if
allocate(region(m)%ij(jtotal))
do n = 1, jtotal
  region(m)%ij(n) = n
end do
  
m = region_number_from_name(name='<domain faces>',location=location,centring=centring,existing=existing,creatable=.true.)
if (existing) then
  write(*,*) "ERROR: the system region = "//trim(name)//" has already been set:- this name should not be used user-defined regions"
  stop
end if
allocate(region(m)%ij(jdomain))
n = 0
do j = 1, jtotal
  if (face(j)%type == 1) then
    n = n + 1
    region(m)%ij(n) = j
  end if
end do
  
m = region_number_from_name(name='<boundaries>',location=location,centring=centring,existing=existing,creatable=.true.)
if (existing) then
  write(*,*) "ERROR: the system region = "//trim(name)//" has already been set:- this name should not be used user-defined regions"
  stop
end if
allocate(region(m)%ij(jboundary))
n = 0
do j = 1, jtotal
  if (face(j)%type == 2) then
    n = n + 1
    region(m)%ij(n) = j
  end if
end do
  
! now run through remaining regions doing any user-defined locating

do m=1,ubound(region,1)

! save user defined regions in gmsh file for analysis
  if (.not.allocated(region(m)%ij)) region(m)%gmsh_region = .true.

!---------------------

  if (region(m)%location(1:2) == "AT") then ! a user defined region from the equations file that is a single point

    read(region(m)%location(3:1000),*,iostat=error) x
    if (error /= 0) then
      write(*,*) 'ERROR: AT location for region ',trim(region(m)%name),' is not understood as a single point'
      stop
    end if

! if region is already defined then it is overwritten
    if (allocated(region(m)%ij)) then
      write(*,'(a/a)') &
        "NOTE: an AT region operator is acting on region "//trim(region(m)%name)// &
        " that already contains an element:", &
        " the previous element will be overwritten with the new"
      region(m)%ij(1) = 0
    else
      allocate(region(m)%ij(1))
    end if

    tmpmax = 1.d+20
    if (region(m)%centring == "cell") then
      do i=1,itotal
        tmp = distance(x,cell(i)%x)
        if (tmp < tmpmax) then
          region(m)%ij(1) = i
          tmpmax = tmp
        end if
      end do
    else 
      do j=1,jtotal
        tmp = distance(x,face(j)%x)
        if (tmp < tmpmax) then
          region(m)%ij(1) = j
          tmpmax = tmp
        end if
      end do
    end if

!---------------------

  else if (region(m)%location(1:11) == "BOUNDARY OF") then ! a new region composed of the boundary to another region

! note: this may mean that the region is already defined but defining twice won't hurt if the definition is the same
    if (allocated(region(m)%ij)) then
      write(*,'(a/a)') &
        "NOTE: a BOUNDARY OF region operator is acting on region "//trim(region(m)%name)// &
        " that already contains some elements:", &
        " the existing elements will be overwritten"
      deallocate(region(m)%ij)
    end if

! check centring of requested region
    if (region(m)%centring /= 'face'.and.region(m)%centring /= 'cell') then
      write(*,*) 'ERROR: incorrect centring for requested BOUNDARY OF region '//trim(region(m)%name)
      stop
    end if

! find constitutent region around which to find boundary
    aregion = adjustl(trim(region(m)%location(12:1000)))
    
    nregion = region_number_from_name(name=aregion,existing=existing,creatable=.false.)
! check that region exists and that centring is consistent
    if (.not.existing) then
      write(*,*) "ERROR: region ",trim(aregion)," which is specified in BOUNDARY OF operator for ", &
        trim(region(m)%name)," is not found"
      stop
    end if
    if (nregion == 0) then
      write(*,*) "ERROR: problem with region ",trim(aregion)," in BOUNDARY OF region ",trim(region(m)%name)
      stop
    end if
    if (region(nregion)%centring /= 'face'.and.region(nregion)%centring /= 'cell') then
      write(*,*) 'ERROR: incorrect centring for BOUNDARY OF consitutent region '//trim(region(nregion)%name)
      stop
    end if

    do nsregion = 1,ubound(region(nregion)%ij,1)  ! loop through all ij indices in constituent region
      ijregion = region(nregion)%ij(nsregion) ! NB, ijregion and region(m)%ij may actually be i or j values depending on centring
      if (region(nregion)%centring == 'cell') then
        if (region(m)%centring == 'cell') then
! cell boundary around cell region
          do ii = 1, ubound(cell(ijregion)%jface,1)+1
            i = cell(ijregion)%icell(ii)
            if (cell(i)%type /= 2) cycle
            if (location_in_list(region(m)%ij,i) == 0) call push_array(array=region(m)%ij,new_element=i)
          end do
        else
! face boundary around cell region
          do jj = 1, ubound(cell(ijregion)%jface,1)
            j = cell(ijregion)%jface(jj)
            if (face(j)%type /= 2) cycle
            if (location_in_list(region(m)%ij,j) == 0) call push_array(array=region(m)%ij,new_element=j)
          end do
        end if
      else
! cell boundary around face region - this will pick out boundary cells coincident with boundary faces
        if (region(m)%centring == 'cell') then
          i = face(ijregion)%icell(2)
          if (cell(i)%type == 2.and.location_in_list(region(m)%ij,i) == 0) &
            call push_array(array=region(m)%ij,new_element=i)
        else
! face boundary around face region - this will pick out faces that are on the boundary
          j = ijregion
          if (face(j)%type == 2.and.location_in_list(region(m)%ij,j) == 0) &
            call push_array(array=region(m)%ij,new_element=j)
        end if
      end if
    end do

!---------------------

  else if (region(m)%location(1:8) == "COMPOUND") then ! a new region composed of a compound list of other regions

! right now save compound regions in gmsh file for analysis
    region(m)%gmsh_region = .true.

! note: this may mean that the region is already defined but defining twice won't hurt if the definition is the same
    if (allocated(region(m)%ij)) write(*,'(a/a)') &
      "NOTE: a COMPOUND region operator is acting on region "//trim(region(m)%name)// &
      " that already contains some elements:", &
      " new elements will be added to/subtracted from the existing list"

    region_list = trim(region(m)%location(9:1000))

! loop through all regions
    do while (len_trim(region_list) /= 0) 

      region_list = adjustl(region_list(1:len(region_list))) ! remove leading spaces

      if (region_list(1:1) == "+" .or. region_list(1:1) == "-") then
        rsign = region_list(1:1)
        region_list = adjustl(region_list(2:len(region_list)))
      else
        rsign = "+"
      end if

      cut = scan(region_list,">")
      if (region_list(1:1) /= "<" .or. cut <=1 ) then
        write(*,*) "ERROR: format for equation region incorrect in COMPOUND operator list ",trim(region_list)
        stop
      end if

      aregion = region_list(1:cut)
      region_list = adjustl(region_list(cut+1:len(region_list)))

      nregion = region_number_from_name(name=aregion,centring=region(m)%centring,existing=existing,creatable=.false.)
! check that region exists and that centring is consistent
      if (.not.existing) then
        write(*,*) "ERROR: region ",trim(aregion)," which is part of COMPOUND region ", &
          trim(region(m)%name)," is not found"
        stop
      end if
      if (nregion == 0) then
        write(*,*) "ERROR: problem with region ",trim(aregion)," in COMPOUND region ",trim(region(m)%name), &
          ":- regions most likely have different centrings"
        stop
      end if

! loop through all existing ij indices in region
! - if indice exists and we are adding, ignore, otherwise do
! - if indice exists and we are subtracting, do, otherwise ignore

      if (.not.allocated(region(nregion)%ij)) then
        write(*,*) "ERROR: region ij indices not allocated in COMPOUND operator for region ",trim(region(nregion)%name)
        stop
      end if
        
      do nsregion = 1,ubound(region(nregion)%ij,1)  ! loop through all ij indices in constituent region
        ijregion = region(nregion)%ij(nsregion) ! NB, ijregion and region(m)%ij may actually be i or j values depending on centring
        
        nscompound = 0 ! this is the position of iregion in the compound region's ij indices
        if (allocated(region(m)%ij)) then
          do ns = 1,ubound(region(m)%ij,1)
            if (region(m)%ij(ns) == ijregion) then
              nscompound = ns ! region indice has been found in compound region's ij indices
              exit
            end if
          end do
        end if

        if (rsign == "+" .and. nscompound == 0) then
          call push_integer_array(array=region(m)%ij,new_element=ijregion) ! add region location to equation ij indices
        else if (rsign == "-" .and. nscompound /= 0) then
          if (nscompound /= ubound(region(m)%ij,1)) then ! unless it is the last element of indicies
            region(m)%ij(nscompound:ubound(region(m)%ij,1)-1) = & ! shift indicies one space to the left to remove reference
              region(m)%ij(nscompound+1:ubound(region(m)%ij,1))
          end if
          call resize_integer_array(array=region(m)%ij,change=-1) ! reduce ij array by 1
        end if

      end do

    end do

!---------------------

  else if (.not.allocated(region(m)%ij)) then

! TODO: other fancier region specifications

    write(*,*) 'ERROR: location for region ',trim(region(m)%name),' not understood'
    stop

  end if

! find ns indicies which give the data number corresponding to location i or j

  if (region(m)%centring == "cell") then
    allocate(region(m)%ns(itotal))
  else if (region(m)%centring == "face") then
    allocate(region(m)%ns(jtotal))
  else
    stop 'ERROR: a region has neither cell or face centring'
  end if

  region(m)%ns = 0 ! a zero indicates that this region does not include this i or j index
  do ns = 1, ubound(region(m)%ij,1)
    region(m)%ns(region(m)%ij(ns)) = ns
  end do
    
  if (debug.and..false.) then
    write(82,*) '# region = '//trim(region(m)%name)//': centring = '//region(m)%centring
    if (region(m)%centring == "cell") then
      write(82,*) '# i, ns'
    else
      write(82,*) '# j, ns'
    end if
    do ij = 1, ubound(region(m)%ns,1)
      write(82,*) ij, region(m)%ns(ij)
    end do
  end if

  if (allocatable_size(region(m)%ij) == 0) then
    write(*,*) 'WARNING: region ',trim(region(m)%name),' contains no elements'
    if (debug) write(*,*) 'm = ',m,': name = ',trim(region(m)%name),': location = ', &
      trim(region(m)%location),': centring = ',region(m)%centring,': gmsh_region = ',region(m)%gmsh_region, &
      ': contains no elements'
  else
    if (debug) write(*,*) 'm = ',m,': name = ',trim(region(m)%name),': location = ', &
      trim(region(m)%location),': centring = ',region(m)%centring,': gmsh_region = ',region(m)%gmsh_region, &
      ': ij(1) = ',region(m)%ij(1),': ij(',ubound(region(m)%ij,1),') = ', &
      region(m)%ij(ubound(region(m)%ij,1))
  end if

end do

! now do reverse indicies - ie, list of regions which each cell is a member of

! cells
do i = 1, itotal
  do m = 1, ubound(region,1)
    if (region(m)%centring /= 'cell') cycle
    if (location_in_list(region(m)%ij,i) /= 0) call push_array(array=cell(i)%region_list,new_element=m)
  end do
end do

! faces
do j = 1, jtotal
  do m = 1, ubound(region,1)
    if (region(m)%centring /= 'face') cycle
    if (location_in_list(region(m)%ij,j) /= 0) call push_array(array=face(j)%region_list,new_element=m)
  end do
end do

! for both cells and faces create gmsh_region_list which specifies what gmsh region an element belongs too
n_gmsh_elements = 0
do i = 1, idomain
  one_region = .false.
  do m = 1, ubound(cell(i)%region_list,1)
    if (region(cell(i)%region_list(m))%gmsh_region) then
      one_region = .true.
      call push_array(array=cell(i)%gmsh_region_list,new_element=cell(i)%region_list(m))
      n_gmsh_elements = n_gmsh_elements + 1
    end if
  end do
  if (.not.one_region) then
    call push_array(array=cell(i)%gmsh_region_list,new_element=0)
    n_gmsh_elements = n_gmsh_elements + 1
  end if
end do
do j = 1, jtotal
  one_region = .false.
  do m = 1, ubound(face(j)%region_list,1)
    if (region(face(j)%region_list(m))%gmsh_region) then
      one_region = .true.
      call push_array(array=face(j)%gmsh_region_list,new_element=face(j)%region_list(m))
      n_gmsh_elements = n_gmsh_elements + 1
    end if
  end do
  if (.not.one_region) then
    call push_array(array=face(j)%gmsh_region_list,new_element=0)
    n_gmsh_elements = n_gmsh_elements + 1
  end if
end do

if (region_details_file) then
  if (debug) write(*,*) 'writing region details to region_details.txt file'

  filename = "output/region_details.txt"
  open(fdetail,file=trim(filename),status='replace',iostat=ierror)
  if (ierror /= 0) then
    write(*,*) 'ERROR: problem opening file ',filename
    stop
  end if

  write(fdetail,'(a)') 'MESH DETAILS:'
  formatline = '(3(a,'//trim(indexformat)//'))'
  write(fdetail,fmt=formatline) 'NODES: ktotal = ',ktotal,': kdomain = ',kdomain,': kboundary = ',kboundary
  formatline = '(a,'//trim(indexformat)//',a,a)'
  do k = 1,ktotal
    write(fdetail,fmt=formatline) 'node: k = ',k,': ',trim(print_node(node(k)))
  end do
  formatline = '(3(a,'//trim(indexformat)//'))'
  write(fdetail,fmt=formatline) 'FACES: jtotal = ',jtotal,': jdomain = ',jdomain,': jboundary = ',jboundary
  formatline = '(a,'//trim(indexformat)//',a,a)'
  do j = 1,jtotal
    write(fdetail,fmt=formatline) 'face: j = ',j,': ',trim(print_face(face(j)))
  end do
  formatline = '(3(a,'//trim(indexformat)//'))'
  write(fdetail,fmt=formatline) 'CELLS: itotal = ',itotal,': idomain = ',idomain,': iboundary = ',iboundary
  formatline = '(a,'//trim(indexformat)//',a,a)'
  do i = 1,itotal
    write(fdetail,fmt=formatline) 'cell: i = ',i,': ',trim(print_cell(cell(i)))
  end do

  close(fdetail)
end if

if (debug) write(*,'(a/80(1h-))') 'subroutine setup_regions'

end subroutine setup_regions

!-----------------------------------------------------------------

subroutine read_constants

! here we read in values for the constants given in constants.in

use general_module
integer, parameter :: fconstants = 1
integer :: error, cut, n, l, nn, ns
character(len=1000) :: textline, keyword, name, formatline, region_name, location
character(len=4) :: centring
integer, dimension(:), allocatable :: region_list ! contains the last read in region_list by number
integer, dimension(:), allocatable :: region_number ! array of regions to use when reading in region_constants
double precision, dimension(:), allocatable :: region_value
logical :: existing
logical, parameter :: debug = .true.
                  
if (debug) write(*,'(80(1h+)/a)') 'subroutine read_constants'

write(*,'(a)') "INFO: reading constant values from file "//trim(constants_file)
open(unit=fconstants,file=trim(constants_file),status='old',iostat=error)
if (error /= 0) stop 'ERROR: problem opening constants file'

fileloop: do
  read(fconstants,'(a)',iostat=error) textline
  if (error /= 0) exit fileloop ! reached end of file
  call remove_comments(textline) ! remove comments from line
  cut=scan(textline,' ') ! split input line at first space
  if (cut<=1) cycle fileloop ! if there is space at first character or string is empty, cycle
  keyword=textline(1:cut-1)
  textline=adjustl(textline(cut:len(textline)))
  if (keyword(1:5) == "CELL_".or.keyword(1:5) == "FACE_".or.keyword(1:5) == "NONE_") then
    if (keyword(1:5) == "CELL_") then
      centring = "cell"
    else if (keyword(1:5) == "FACE_") then
      centring = "face"
    else if (keyword(1:5) == "NONE_") then
      centring = "none"
    end if
    keyword=adjustl(keyword(6:len(keyword)))
  else
    centring = "none"
  end if
! if (debug) write(*,'(a)') 'INFO: in constants file found keyword = '//trim(keyword)//': with centring = '//centring
! if (debug) write(*,'(a)') 'textline = '//trim(textline)

  if (keyword == "REGION_CONSTANT" .and. centring == "none") then
    write(*,*) 'ERROR: problem in constants file: a REGION_CONSTANT must have either CELL or FACE centring'
    stop
  end if

!---------------
  if (trim(keyword) == 'REGION_LIST') then ! read in region names and create a local array (region_list) containing them
    if (allocated(region_list)) deallocate(region_list)
    do while (len_trim(textline)>0)
      cut=scan(textline,'>') ! split input line at region end delimiter
      if (textline(1:1) /= '<'.or.cut<2) then
        write(*,*) 'ERROR: a region name in '//trim(keyword)//' in constants file incorrectly specified on line:'
        write(*,*) trim(textline)
        stop
      end if
      region_name = textline(1:cut)
      textline=adjustl(textline(cut+1:len(textline)))
!write(*,*) 'region_name = '//trim(region_name)//': textline = '//trim(textline)
      m = region_number_from_name(name=region_name,centring=centring,creatable=.false.,existing=existing)
      if (.not.existing .or. m == 0) then
        write(*,*) 'ERROR: the region '//trim(region_name)//' which appears in a REGION_LIST does not exist or has ', &
          'inconsistent centring'
        stop
      end if
      call push_integer_array(array=region_list,new_element=m)
    end do
    if (debug) then
      write(*,'(a,i2,a)') 'INFO: a REGION_LIST has been found that contains ',ubound(region_list,1),' elements'
      do m=1,ubound(region_list,1)
        write(*,*) 'm = ',m,': region_list = ',region_list(m),': region_name = ', &
          trim(region(region_list(m))%name)
      end do
    end if
  end if
!---------------

!---------------
! name
  if (trim(keyword) == 'REGION_CONSTANT'.or.trim(keyword) == 'CONSTANT') then ! read in variable names
    cut=scan(textline,'>') ! find end of name string
    if (textline(1:1) /= '<'.or.cut<2) then
      write(*,*) 'ERROR: name in '//trim(keyword)//' in constants file incorrectly specified on line:'
      write(*,*) trim(textline)
      stop
    end if
    name = textline(1:cut) ! name includes <>
    textline=adjustl(textline(cut+1:len(textline)))

!---------------
! units

    cut=scan(textline,']') ! find end of units string
    if (textline(1:1) /= '['.or.cut<2) then
      write(*,*) 'WARNING: no units assigned to '//trim(keyword)//' '//trim(name)//' in constants file on line:'
      write(*,*) trim(textline)
    else
      textline=adjustl(textline(cut+1:len(textline)))
    end if

!---------------
! find constant var number

    m = var_number_from_name(name)
    if (m == 0.or.trim(var(m)%type) /= "constant") then
      write(*,*) 'ERROR: constant variable '//trim(name)//' was not found when reading constants file'
      stop
    end if
    if (trim(var(m)%centring) /= trim(centring)) then
      write(*,*) 'ERROR: constant variable '//trim(name)//' seems inconsistent when reading constants file'
      stop
    end if

!---------------

    if (allocated(region_value)) deallocate(region_value)
    if (allocated(region_number)) deallocate(region_number)

    if (trim(keyword) == 'REGION_CONSTANT') then
      if (.not.allocated(region_list)) then
        write(*,*) 'ERROR: a REGION_LIST statement must precede a REGION_CONSTANT statement in the constants file'
        stop
      end if
      allocate(region_number(ubound(region_list,1)))
      region_number = region_list
      allocate(region_value(ubound(region_list,1)))
    else if (trim(keyword) == 'CONSTANT') then
      allocate(region_value(1))
      allocate(region_number(1))
    end if

    read(textline,*,iostat=error) region_value
    if (error /= 0) then
      write(*,*) 'ERROR: problem reading in constant '//name
      stop
    end if
    region_value = region_value*var(m)%multiplier ! convert to specified units

!---------------
! location statements for constant

    if (trim(keyword) == 'CONSTANT' .and. trim(centring) /= "none") then
      cut=scanstring(textline,'ON ') ! find location sequence
      if (cut > 0) textline=adjustl(trim(textline(cut+2:len(textline))))
      cut=scan(textline,'>') ! find end of region name
      if (textline(1:1) == '<'.and.cut>1) then
        region_name = textline(1:cut)
      else if (centring == "cell") then
        region_name = "<all cells>"
      else if (centring == "face") then
        region_name = "<all faces>"
      end if
      region_number(1) = region_number_from_name(name=region_name,creatable=.false.,existing=existing)
      if (.not.existing .or. region_number(1) == 0) then
        write(*,*) 'ERROR: the region '//trim(region_name)//' which appears in a CONSTANT statement does not exist'
        stop
      end if
    end if

!---------------
! finally assign values to correct regions

    if (trim(centring) == "none") then
      var(m)%funk(1)%v = region_value(1)
    else
      do n = 1, ubound(region_number,1)
        if (trim(region(region_number(n))%centring) /= trim(var(m)%centring)) then
          write(*,*) 'ERROR: when reading the constants file, the centring of constant '//trim(var(m)%name)// &
            ' does not match that of region '//trim(region(region_number(n))%name)
          stop
        end if
        do nn = 1, ubound(region(region_number(n))%ij,1)
          ij = region(region_number(n))%ij(nn)
          ns = nsvar(m,ij)
          if (ns /= 0) var(m)%funk(ns)%v = region_value(n)
        end do
      end do
    end if

    if (debug) then
      write(*,'(a)') 'INFO: read in '//var(m)%centring//' '//trim(var(m)%type)//' '//trim(var(m)%name)
      if (centring == "none") then
        write(*,*) '  none: value = ',var(m)%funk(1)%v
      else
        do n = 1, ubound(region_number,1)
          write(*,*) '  '//trim(region(region_number(n))%name)//': value = ',region_value(n)
        end do
      end if
    end if
    
  end if
    
!---------------

end do fileloop

close(fconstants)

if (debug) write(*,'(a/80(1h-))') 'subroutine read_constants'

end subroutine read_constants

!----------------------------------------------------------------------------

subroutine setup_vars

! here we allocate array elements for the fields and functions and initialise fields

use general_module
use equations_module
integer :: m, i, j, ns, n, mc, pptotal, mtype
!character(len=100), dimension(6), save :: local_type = [ "constant   ", "transient  ", "unknown", "derived  ", "equation   ", "output     " ]
character(len=100), dimension(6) :: local_type
logical :: existing
logical, parameter :: debug = .true.
                  
if (debug) write(*,'(80(1h+)/a)') 'subroutine setup_vars'

local_type = [ "constant   ", "transient  ", "unknown    ", "derived    ", "equation   ", "output     " ]

! allocate var funks, zero these funks and set region numbers
do m = 1, ubound(var,1)
  if (var(m)%centring /= "none") then
    var(m)%region_number = region_number_from_name(name=var(m)%region,centring=var(m)%centring,existing=existing)
! check that region exists and that the variable and region centring are consistent
    if (.not.existing) then
      write(*,*) 'ERROR: there is a problem with region '//trim(var(m)%region)//' which is associated with '// &
        trim(var(m)%type)//' '//trim(var(m)%name)//': the region does not exist'
      stop
    end if
    if (var(m)%region_number == 0) then
      write(*,*) 'ERROR: there is a problem with region '//trim(var(m)%region)//' which is associated with '// &
        trim(var(m)%type)//' '//trim(var(m)%name)//': most probably the centrings are inconsistent'
      stop
    end if
    allocate(var(m)%funk(ubound(region(var(m)%region_number)%ij,1)))
  else
    var(m)%region_number = 0 ! dummy region number for none centred variables
    allocate(var(m)%funk(1))
  end if
  do ns = 1, ubound(var(m)%funk,1)
    call reset_funk(var(m)%funk(ns))
  end do
end do

! copy region numbers set in components to compound
do mc = 1, ubound(compound,1)
  do n = 1, ubound(compound(mc)%component,1)
    m = compound(mc)%component(n)
    if (m /= 0) then
      compound(mc)%region_number = var(m)%region_number
      exit
    end if
  end do
end do

! now setup var_lists
allocate(var_list(var_list_number(type="all",centring="all")))
do mtype = 1, ubound(local_type,1)
  do m = 1, ubound(var,1)
    if (trim(var(m)%type) /= trim(local_type(mtype))) cycle ! lists are created in order
    call push_array(array=var_list(var_list_number(type="all",centring="all"))%list,new_element=m)
    call push_array(array=var_list(var_list_number(type=var(m)%type,centring="all"))%list,new_element=m)
    call push_array(array=var_list(var_list_number(type="all",centring=var(m)%centring))%list,new_element=m)
    call push_array(array=var_list(var_list_number(type=var(m)%type,centring=var(m)%centring))%list,new_element=m)
  end do
end do
if (debug.and..false.) then
  write(82,*) 'var_list details:'
  do m = 1, var_list_number(type="all",centring="all")
    write(82,*) 'var_list number = ',m,': allocatable_size = ',allocatable_size(var_list(m)%list)
    if (allocatable_size(var_list(m)%list) > 0) then
      do n = 1, allocatable_size(var_list(m)%list)
        write(82,*) '  contains '//trim(var(var_list(m)%list(n))%name)//' of centring = '// &
          trim(var(var_list(m)%list(n))%centring)//' and type = '//trim(var(var_list(m)%list(n))%type)
      end do
    end if
  end do
end if

! run through constants reading in file values and setting any equation determined ones
call read_constants
call update_constants
  
! run through unknowns setting initial values
call update_unknowns
  
! run through unknowns setting derivatives and calculating ptotal
ptotal = 0 ! this is the number of unknown variables (should equal pptotal)
do n = 1, allocatable_size(var_list(var_list_number(centring="all",type="unknown"))%list)
  m = var_list(var_list_number(centring="all",type="unknown"))%list(n)
  do ns = 1, ubound(var(m)%funk,1)
    ptotal = ptotal + 1
    call push_array(var(m)%funk(ns)%pp,ptotal)
    call push_array(var(m)%funk(ns)%dv,1.d0)
    var(m)%funk(ns)%ndv = 1
  end do
end do
  
! run through equations calculating pptotal
pptotal = 0 ! this is the number of equations (should equal ptotal)
do n = 1, allocatable_size(var_list(var_list_number(centring="all",type="equation"))%list)
  m = var_list(var_list_number(centring="all",type="equation"))%list(n)
  do ns = 1, ubound(var(m)%funk,1)
    pptotal = pptotal + 1
  end do
end do

! spray out some info about all the var variables
if (debug) then
  do mtype = 1, ubound(local_type,1)
    write(*,'(a/a)') '------------------------------------','INFO: '//trim(local_type(mtype))
    do n = 1, allocatable_size(var_list(var_list_number(centring="all",type=trim(local_type(mtype))))%list)
      m = var_list(var_list_number(centring="all",type=trim(local_type(mtype))))%list(n)
      if (var(m)%centring == 'none') then
        write(*,*) 'm = ',m, &
          ': name = ',trim(var(m)%name), &
          ': type = ',trim(var(m)%type), &
          ': centring = ',var(m)%centring, &
          ': rank = ',trim(var(m)%rank)
      else
        write(*,*) 'm = ',m, &
          ': name = ',trim(var(m)%name), &
          ': type = ',trim(var(m)%type), &
          ': centring = ',var(m)%centring, &
          ': rank = ',trim(var(m)%rank), &
          ': region = ',trim(var(m)%region), &
          ': region_number = ',var(m)%region_number, &
          ': ij(1) = ',region(var(m)%region_number)%ij(1), &
          ': ij(',ubound(region(var(m)%region_number)%ij,1),') = ', &
          region(var(m)%region_number)%ij(ubound(region(var(m)%region_number)%ij,1))
      end if
    end do
  end do
  write(*,'(a)') '------------------------------------'
end if

if (pptotal /= ptotal) then
  write(*,*) 'ERROR: the total number of equations does not match the total number of unknown variables'
  write(*,*) '  pptotal (number of equations) = ',pptotal
  write(*,*) '  ptotal (number of unknown variables) = ',ptotal
  stop
end if

! allocate funkmag which is used in residual calculations
allocate(funkmag(ptotal))

if (debug) write(*,'(a/80(1h-))') 'subroutine setup_vars'

end subroutine setup_vars

!-----------------------------------------------------------------

end module setup_module

!-----------------------------------------------------------------
